//////////////////////////////////////////////////////////////////////
// Mainz Rolling Brains RoboCup agents developed for the Simulator 
// League of the RoboCup 2000 competition
//
// Copyright (C) 1998, 1999, 2000 Team Members 
// of the Mainz Rolling Brains RoboCup Team 
//
// This program under the terms of the GNU General Public License as
// published by the Free Software Foundation; version 2 of the
// License. See the README file in the root directory of our archive.
//////////////////////////////////////////////////////////////////////


//-*- c++ -*-/////////////////////////////////////////////////////////
//
//  $Id: asyncinfo.C,v 1.10 2000/10/19 15:27:42 goetz Exp $
// 
//  asyncinfo.C - storing of asynchronously updated agent information
// 
//////////////////////////////////////////////////////////////////////


#include "asyncinfo.h"

// [HACK] repeated sending of delayed/lost commands, need Comm* to send them
#include "comm.h"


#undef CERR
// error trace, abbreviate error messages by commenting out #define
//#define CERR(output) cerr << output << endl;




#undef TRACE
#undef SENSETRACE

#define TRACE(cmd)
#define SENSETRACE(cmd)


//////////////////////////////////////////////////////////////////////


AsyncCommInfo::AsyncCommInfo(const char* _team, Comm* _comm) {

  TRACE(cout << "AsyncInfo: entering constructor" << endl);

  team = new char[strlen(_team)+1];
  strcpy(team, _team);

  number        = -1;
  client        = CLIENT_UNKNOWN;
  neckTwisted   = false;

  serverTime    = 0;
  playMode      = PLAY_MODE_UNKNOWN;
  side          = SIDE_UNKNOWN;
  ourScore      = 0;
  otherScore    = 0;
  secondHalf    = false;

  kickCount     = 0;
  dashCount     = 0;
  turnCount     = 0;
  sayCount      = 0;
  turnNeckCount = 0;

  clearAction(serverTime);
  processedAction = sentAction;

  noSenseBodyYet = true;
  nextTimestep   = false;

  // [HACK]
  comm = _comm;

  TRACE(cout << "AsyncInfo: leaving constructor" << endl);

}


AsyncCommInfo::~AsyncCommInfo() {
  TRACE(cout << "AsyncInfo: entering destructor" << endl);
  delete[] team;
  TRACE(cout << "AsyncInfo: leaving destructor" << endl);
}


void
AsyncCommInfo::setStates(client_type type, int _number, 
			 int _time, play_mode _mode, 
			 playing_side _side, bool _secondHalf) {
  
  number     = _number;
  serverTime = _time;
  playMode   = _mode;
  side       = _side;
  secondHalf = _secondHalf;

  client     = type;

#ifndef CERR
#define CERR(output) cerr << "A";
#warning AsyncInfo error trace is abbreviated to A
  cerr << "*** Warning: AsyncInfo(" 
       << team << " No." << number 
       << "): error trace is abbreviated to A, cf. source" << endl;
#endif
  


}


void 
AsyncCommInfo::setSentAction(CycleCommand cmd, int time, 
				double param1, double param2) {
  TRACE(cout << "AsyncInfo: set current action = ";
	switch (cmd) {
	case NO_COMMAND   : cout << "none" << endl; break;
	case KICK_COMMAND : cout << "kick " << param1 << " " 
				 << param2 << endl; break;
	case DASH_COMMAND : cout << "dash " << param1 << endl; break;
	case TURN_COMMAND : cout << "turn " << param1 << endl; break;
	case CATCH_COMMAND: cout << "catch " << param1 << endl; break;
	case MOVE_COMMAND : cout << "move " << param1 << " "
				 << param2 << endl; break;
	});

  sentAction.command = cmd;
  sentAction.time = time;
  sentAction.param1 = param1;
  sentAction.param2 = param2;
  actionNeeded = false;
}


void
AsyncCommInfo::clearAction(int time) {
  
  setSentAction (NO_COMMAND, time, 0, 0);
  TRACE(cout << "           marking action as unprepared" << endl);
  actionNeeded = true;
}


bool
AsyncCommInfo::actNow() {
  if (isCoach()) {
    if (nextTimestep) {
      nextTimestep = false;
      return true;
    } else {
      return false;
    }
  } else {
    if (actionNeeded && !noSenseBodyYet && nextTimestep) {
      nextTimestep = false;
      return true;
    } else {
      return false;
    }
  }
}


void
AsyncCommInfo::getLastAction(CycleAction &action) {
  
  TRACE(cout << "AsyncInfo: returning last action (of time " 
	<< processedAction.time << ") = ";
	switch (processedAction.command) {
	case NO_COMMAND   : cout << "none" 
				 << endl; break;
	case KICK_COMMAND : cout << "kick " 
				 << processedAction.param1 << " " 
				 << processedAction.param2 << endl; break;
	case DASH_COMMAND : cout << "dash " 
				 << processedAction.param1 << endl; break;
	case TURN_COMMAND : cout << "turn " 
				 << processedAction.param1 << endl; break;
	case CATCH_COMMAND: cout << "catch " 
				 << processedAction.param1 << endl; break;
	case MOVE_COMMAND : cout << "move " 
				 << processedAction.param1 << " "
				 << processedAction.param2 << endl; break;
	});
  
  action = processedAction;
}

void 
AsyncCommInfo::newTimestep() {

  // set flag, so that checking of command counters will take place after
  // next sense_body, and new command will be requested from agent.
  noSenseBodyYet = true;

  // assume sending went ok until checking of counters can be performed
  processedAction = sentAction; 

  nextTimestep = true;

}


void 
AsyncCommInfo::checkSyncCommandCount(SensorData data) {

  TRACE (cout << "AsyncInfo: checking for lost commands:" << endl
	 << "           old (kick, dash, turn) = " 
	 << kickCount << " " 
	 << dashCount << " " 
	 << turnCount << " " 
	 << "           current action = ";
	 switch (sentAction.command) {
	 case NO_COMMAND   : cout << "none"; break;
	 case KICK_COMMAND : cout << "kick"; break;
	 case DASH_COMMAND : cout << "dash"; break;
	 case TURN_COMMAND : cout << "turn"; break;
	 case CATCH_COMMAND: cout << "catch"; break;
	 case MOVE_COMMAND : cout << "move"; break;
	 }
	 cout << endl << "           rcv (kick, dash, turn) = "
	 << data.kick_count << " " 
	 << data.dash_count << " " 
	 << data.turn_count << endl);
  
  // check for missed kick, dash, turn
  switch (sentAction.command) {
  case (KICK_COMMAND):
    if (data.kick_count != kickCount + 1) {
      if (data.kick_count == kickCount) { // missed the last kick
	CERR( serverTime 
	     << " AsyncInfo (" 
	     << getTeam() << " No." << getNumber()
	     << "): kick was lost, or delayed one step");
      } else {
	CERR( serverTime 
	     << " AsyncInfo (" 
	     << getTeam() << " No." << getNumber()
	     << "): missing several kick commands"  );
      }
      processedAction.command = NO_COMMAND; // no action was processed

      /*
      clearSentAction(serverTime);          // new action is needed
      // clear sentAction, so that agent will be forced to compute a new 
      // action and send it. If the action was just delayed, the delayed 
      // action will be processed by the server, the newly computed action will
      // not be effective at all, and be prompted as "lost" in its turn.

      // This feature will confuse the worldmodel simuluation, as DELAYED
      // commands will be executed, but not simulated properly by the 
      // worldmodel. With LOST commands it will be working ok, though.

      // [HERE] solution: store sentAction in lostAction, check next step 
      // if newly sentAction or lostAction was executed, 
      // and set processedAction correctly.
      */


      comm->sendKick(sentAction.param1, sentAction.param2);
      sentAction.time = serverTime;
      // Do not clear sentAction, as it will be executed by server in 
      // next step is the command was delayed. If the command was actually 
      // lost, we will resend the old command, just to make sure it arrives.
      // [HERE] No new action will be needed, but the mechanism 
      // could be improved (see above) !


    } else {
      clearAction(serverTime);
    }
    break;

  case (DASH_COMMAND):
    if (data.dash_count != dashCount + 1) {
      if (data.dash_count == dashCount) { // missed the last dash
	CERR( serverTime 
	     << " AsyncInfo (" 
	     << getTeam() << " No." << getNumber()
	     << "): dash was lost, or delayed one step"  );
      } else {
	CERR( serverTime 
	     << " AsyncInfo (" 
	     << getTeam() << " No." << getNumber()
	     << "): missing several dash commands"  );
      }
      processedAction.command = NO_COMMAND;

      /*
      clearSentAction(serverTime);          // new action is needed
      // clear sentAction, so that agent will be forced to compute a new 
      // action and send it. If the action was just delayed, the delayed 
      // action will be processed by the server, the newly computed action will
      // not be effective at all, and be prompted as "lost" in its turn.

      // This feature will confuse the worldmodel simuluation, as DELAYED
      // commands will be executed, but not simulated properly by the 
      // worldmodel. With LOST commands it will be working ok, though.

      // [HERE] solution: store sentAction in lostAction, check next step 
      // if newly sentAction or lostAction was executed, 
      // and set processedAction correctly.
      */

      comm->sendDash(sentAction.param1);
      sentAction.time = serverTime;
      // Do not clear sentAction, as it will be executed by server in 
      // next step. No new action will be needed !

    } else {
      clearAction(serverTime);
    }
    break;

  case (TURN_COMMAND):
    if (data.turn_count != turnCount + 1) {
      if (data.turn_count == turnCount) { // missed the last turn
	CERR( serverTime 
	     << " AsyncInfo (" 
	     << getTeam() << " No." << getNumber()
	     << "): turn was lost, or delayed one step"  );
      } else {
	CERR( serverTime 
	     << " AsyncInfo (" 
	     << getTeam() << " No." << getNumber()
	     << "): missing several turn commands"  );
      }
      processedAction.command = NO_COMMAND;

      /*
      clearSentAction(serverTime);          // new action is needed
      // clear sentAction, so that agent will be forced to compute a new 
      // action and send it. If the action was just delayed, the delayed 
      // action will be processed by the server, the newly computed action will
      // not be effective at all, and be prompted as "lost" in its turn.

      // This feature will confuse the worldmodel simuluation, as DELAYED
      // commands will be executed, but not simulated properly by the 
      // worldmodel. With LOST commands it will be working ok, though.

      // [HERE] solution: store sentAction in lostAction, check next step 
      // if newly sentAction or lostAction was executed, 
      // and set processedAction correctly.
      */

      comm->sendTurn(sentAction.param1);
      sentAction.time = serverTime;
      // Do not clear sentAction, as it will be executed by server in 
      // next step. No new action will be needed !

    } else {
      clearAction(serverTime);
    }
    break;

  default: clearAction(serverTime); break;

  } //switch

  // after printing error messages: update counters
  kickCount = data.kick_count;
  dashCount = data.dash_count;
  turnCount = data.turn_count;
  
}

void 
AsyncCommInfo::checkAsyncCommandCount(SensorData data) {

  TRACE (cout << "AsyncInfo: checking for lost commands:" << endl
	 << "           old (say,turn_neck) = " 
	 << sayCount << " " 
	 << turnNeckCount << endl;
	 cout << "           rcv (say,turn_neck) = "
	 << data.say_count << " " 
	 << data.turn_neck_count << endl);
  
  // check for missed say, turn_neck
  if (sayCount != data.say_count) {
    CERR( serverTime 
	 << " AsyncInfo (" << getTeam() << " No." << getNumber()
	 << "): missing say commands "
	 << "(sent " << sayCount 
	 << ", was told " << data.say_count << ")"  );
    sayCount = data.say_count;
  }
  if (turnNeckCount != data.turn_neck_count) {
    CERR( serverTime 
	 << " AsyncInfo (" << getTeam() << " No." << getNumber()
	 << "): missing turn_neck commands "
	 << "(sent " << turnNeckCount 
	 << ", was told " << data.turn_neck_count << ")"  );
    turnNeckCount = data.turn_neck_count;
  }
  
}

bool
AsyncCommInfo::adjustTime(SensorData data) {
  
  TRACE (cout << "AsyncInfo: current time = " << serverTime 
	 << ", adjusting to data.time = " << data.time << endl);

  if (data.time > serverTime) {  // new data
    
    if (data.time > serverTime + 1) { // data MUCH TOO new

      CERR( data.time
	   << " AsyncInfo (" << getTeam() << " No." << getNumber()
	   << "): completely missed server timesteps from "
	   << serverTime+1 << " to " << data.time-1 << " (no commands sent)" 
	    );

    } else if ( actionNeeded && 
		!isCoach() &&
		isTimeRunning() ) {

      // isTimeRunning() is appropriate here for the lost command check:

      // if game was stopped, the play mode will still reflect this right now, 
      // as it is updated AFTER adjustTime() in adjust(). So: prompt error
      // only if game was already running -- to avoid error prompts when
      // game is about to be restarted.

      CERR( serverTime
	   << " AsyncInfo (" << getTeam() << " No." << getNumber()
	   << "): agent did not send command"  );

    }

    SENSETRACE(
	       switch(data.type) {
	       case SENSOR_TYPE_SEE:
		 cout << "AsyncInfo: see was first message !" << endl; break;
	       case SENSOR_TYPE_HEAR:
		 cout << "AsyncInfo: hear was first message !" << endl; break;
	       case SENSOR_TYPE_SENSE_BODY:
		 break; // was ok
	       default:
		 cout << "AsyncInfo: " << (int)data.type 
		      << "was first message !" << endl; break;
	       }
	       );

    serverTime = data.time;
    newTimestep();
    return true;
    
  } else {
    return false;
  }

}

void 
AsyncCommInfo::twistNeck() {
  TRACE (cout << "AsyncInfo: twisting neck" <<endl);
  neckTwisted = true;
}

void 
AsyncCommInfo::incSayCount() {
  TRACE (cout << "AsyncInfo: increasing sayCount to " << sayCount <<endl);
  sayCount++;
}

void  
AsyncCommInfo::incTurnNeckCount() {
  TRACE (cout << "AsyncInfo: increasing turnNeckCount to " << turnNeckCount 
	 << endl);
  turnNeckCount++;
}

bool 
AsyncCommInfo::isTimeRunning() {
  return ! ( getPlayMode() == PLAY_MODE_BEFORE_KICK_OFF ||
	     getPlayMode() == PLAY_MODE_TIME_OVER ||
	     getPlayMode() == PLAY_MODE_OFFSIDE_US ||
	     getPlayMode() == PLAY_MODE_OFFSIDE_THEM ||
	     getPlayMode() == PLAY_MODE_GOAL_US ||
	     getPlayMode() == PLAY_MODE_GOAL_THEM );
}

bool 
AsyncCommInfo::adjust(SensorData data) {

  bool firstSenseBody = false;


  // 1. adjust server time according to data
  adjustTime(data);
  

  // 2. check for new timestep for coach when clock is off
  if (data.type == SENSOR_TYPE_EXACT && 
      !isTimeRunning()) {
    newTimestep();
  }

  // 3a. adjust headAngle, 
  // 3b. force new timestep for player when clock is off,
  // 3c. check and update command counters
  if (data.type == SENSOR_TYPE_SENSE_BODY) {

    // we KNOW our own head_angle after receiving a sense_body
    neckTwisted = false;

    
    if (!isTimeRunning()) { // force new timestep when clock is stopped.
      TRACE (cout << "AsyncInfo: game is stopped, "
	     << "had a sense_body -> \"next\" timestep !" << endl);
      newTimestep();
      // This might confuse the action counters temporarily (for one step)
      // when additional sense_body
      // messages are received while clock is stopped, e.g. through turn_necks
      // etc. Just avoid receiving additional (sense_body *) in this case.
    }

    
    if (noSenseBodyYet) { 
      // upon first sense_body received in a timestep:
      
      // clear flag
      noSenseBodyYet = false;
      // check turn, dash, kick (only once per step)
      checkSyncCommandCount(data); 
      // set return value
      firstSenseBody = true;
      
    } 
  
    // 4. check say, turn_neck for every sense_body
    checkAsyncCommandCount(data); 

  }

  // adjust playMode according to referee message
  if ( (data.type == SENSOR_TYPE_HEAR) ) {
    if (data.sender == SENDER_REFEREE) {
      playMode = data.mode;
      if (playMode == PLAY_MODE_GOAL_US) ourScore = data.score;
      if (playMode == PLAY_MODE_GOAL_THEM) otherScore = data.score;
      if (playMode == PLAY_MODE_HALF_TIME) secondHalf = true;
    }
  }

  return firstSenseBody;

}

