//////////////////////////////////////////////////////////////////////
// Mainz Rolling Brains RoboCup agents developed for the Simulator 
// League of the RoboCup 2000 competition
//
// Copyright (C) 1998, 1999, 2000 Team Members 
// of the Mainz Rolling Brains RoboCup Team 
//
// This program under the terms of the GNU General Public License as
// published by the Free Software Foundation; version 2 of the
// License. See the README file in the root directory of our archive.
//////////////////////////////////////////////////////////////////////

//-*- c++ -*-/////////////////////////////////////////////////////////
//
//  $Id: asyncinfo.h,v 1.5 2000/10/19 15:27:43 goetz Exp $
// 
//  asyncinfo.h - storing of player related asynchronous information
// 
//////////////////////////////////////////////////////////////////////


#ifndef _ASYNCINFO_H_
#define _ASYNCINFO_H_

#include <iostream.h>
#include "sensor.h"


/////////////////////////////////////////////////////////////////////////////

// [HACK]
class Comm;

class AsyncCommInfo {
  
 public:

  // [HACK]
  AsyncCommInfo(const char* _team, Comm* _comm);
  
  ~AsyncCommInfo();
  
  const char*  getTeam()       {return team;}
  int          getNumber()     {return number;}
  bool         isOnlineCoach() {return client==CLIENT_ONLINECOACH;}
  bool         isTrainCoach()  {return client==CLIENT_TRAINCOACH;}
  bool         isCoach()       {return isOnlineCoach()||isTrainCoach();}
  bool         isGoalie()      {return client==CLIENT_GOALIE;}
  bool         isNeckTwisted() {return neckTwisted;}

  int          getServerTime() {return serverTime;}
  play_mode    getPlayMode()   {return playMode;}
  playing_side getSide()       {return side;}
  int          getOurScore()   {return ourScore;}
  int          getOtherScore() {return otherScore;}
  bool         inSecondHalf()  {return secondHalf;}

  bool         isTimeRunning();

  void         twistNeck();
  void         incSayCount();
  void         incTurnNeckCount();
  
  void         setSentAction(CycleCommand cmd, int time, 
			     double param1, double param2);
  
  bool         actNow();
  void         getLastAction(CycleAction &action);

  // use Sensor Data to adjust time, playMode and related states,
  // twisted neck, and action counters; and check action count.
  // returns true, if data was the first sense_body data of a timestep
  bool         adjust(SensorData data);

  // for access to states when setting up the agent only !!
  void         setStates(client_type type, int _number, 
			 int _time, play_mode _mode, 
			 playing_side _side, bool _secondHalf);

private:

  char*        team;
  int          number;
  client_type  client;
  bool         neckTwisted;

  int          serverTime;
  play_mode    playMode;
  playing_side side;
  int          ourScore;
  int          otherScore;
  bool         secondHalf;
  
  int kickCount;
  int dashCount;
  int turnCount;
  int sayCount;
  int turnNeckCount;
  
  bool noSenseBodyYet;        // true: no sense body received in current step
  bool actionNeeded;          // true: no action sent yet in current step
  bool nextTimestep;          // true: transition was made to new step

  CycleAction sentAction;     // action that was just sent, unconfirmed
  CycleAction processedAction;// action processed by server from the last step
  
  void newTimestep();
  void clearAction(int time);
  bool adjustTime(SensorData data);

  void checkSyncCommandCount(SensorData data);   // turn, dash, kick
  void checkAsyncCommandCount(SensorData data);  // say, turn_neck

  // [HACK]
  Comm* comm;
  
};


#endif
