//////////////////////////////////////////////////////////////////////
// Mainz Rolling Brains RoboCup agents developed for the Simulator 
// League of the RoboCup 2000 competition
//
// Copyright (C) 1998, 1999, 2000 Team Members 
// of the Mainz Rolling Brains RoboCup Team 
//
// This program under the terms of the GNU General Public License as
// published by the Free Software Foundation; version 2 of the
// License. See the README file in the root directory of our archive.
//////////////////////////////////////////////////////////////////////

//////////////////////////////////////////////////////////////////////
//	$Id: constants.C,v 1.10 2000/10/19 15:27:45 goetz Exp $	
//	$Manager: goetz$
//////////////////////////////////////////////////////////////////////

/*
 *  constants.C - Definition of constants
 *
 *  Author: Goetz Schwandtner <goetz@informatik.uni-mainz.de> )
 *          (Christian Bauer <cbauer@iphcip1.physik.uni-mainz.de> )
 *
 *  RoboCup 98, 99, 2000 - "Mainz Rolling Brains"
 */

#include <stdio.h>
#include <string.h>

#include "constants.h"

// Class with static instance for initializing the constants

class ConstantsInit
{
public:
  ConstantsInit();
};

static ConstantsInit ci;

// Constants from server.conf
// Default values

double GOAL_WIDTH           = 14.02;
double PLAYER_SIZE          = 0.3;
double PLAYER_DECAY         = 0.4;
double PLAYER_RAND          = 0.1;
double PLAYER_WEIGHT        = 60.0;
double PLAYER_SPEED_MAX     = 1.0;
double STAMINA_MAX          = 3500;
double STAMINA_INC_MAX      = 35.0;
double RECOVER_DEC_THR      = 0.3;
double RECOVER_DEC          = 0.002;
double RECOVER_MIN          = 0.5;
double EFFORT_DEC_THR       = 0.3;
double EFFORT_DEC           = 0.005;
double EFFORT_INC_THR       = 0.6;
double EFFORT_INC           = 0.01;
double EFFORT_MIN           = 0.6;
double HEAR_MAX             = 2;
double HEAR_INC             = 1;
double HEAR_DECAY           = 2;
double INERTIA_MOMENT       = 5.0;
int    SENSE_BODY_STEP      = 100;
double CATCHABLE_AREA_L     = 2.0;
double CATCHABLE_AREA_W     = 1.0;
double CATCH_PROBABILITY    = 1.0;
int    CATCH_BAN_CYCLE      = 5;
int    GOALIE_MAX_MOVES     = 2;
double BALL_SIZE            = 0.085;
double BALL_DECAY           = 0.94;
double BALL_RAND            = 0.05;
double BALL_WEIGHT          = 0.2;
double BALL_SPEED_MAX       = 2.7;
double WIND_FORCE           = 0.0;
double WIND_DIR             = 0.0;
double WIND_RAND            = 0.0;
double KICKABLE_MARGIN      = 0.7;
double KICKRANGE            = BALL_SIZE + PLAYER_SIZE + KICKABLE_MARGIN;
double CKICK_MARGIN         = 1.0;
double DASH_POWER_RATE      = 0.006;
double KICK_POWER_RATE      = 0.016;
double VISIBLE_ANGLE        = 90.0;
double AUDIO_CUT_DIST       = 50.0;
double QUANTIZE_STEP        = 0.1;
double QUANTIZE_STEP_L      = 0.01;
double MAXPOWER             = 100;
double MINPOWER             = -100;
double MAXMOMENT            = 180;
double MINMOMENT            = -180;
double MAXNECKMOMENT        = 180;
double MINNECKMOMENT         = -180;
double MAXNECKANG           = 90;
double MINNECKANG           = -90;
int    PORT                 = 6000;
int    COACH_PORT           = 6001;
int    OLCOACH_PORT         = 6002;
int    SAY_COACH_CNT_MAX    = 128;
int    SAY_COACH_MSG_SIZE   = 128;
int    SEND_VI_STEP         = 100;
int    SIMULATOR_STEP       = 100;
int    SEND_STEP            = 150;
int    RECV_STEP            = 10;
int    HALF_TIME            = 300;
int    SAY_MSG_SIZE         = 512;
bool   USE_OFFSIDE          = true;
double OFFSIDE_ACTIVE_AREA  = 5;
bool   FORBID_KICK_OFF_OFFSIDE = true;
double OFFSIDE_KICK_MARGIN  = 9.15;
int    SLOW_DOWN_FACTOR     = 1;

int    HALF_TIME_IN_CYCLES  = 3000;

// Flag positions
double flag_x[NUM_FLAGS], flag_y[NUM_FLAGS];

// Goal positions
double goal_x[NUM_GOALS], goal_y[NUM_GOALS];


/*
 *  Initialize constants
 */

ConstantsInit::ConstantsInit()
{
  // Variables for reading "on" and "off" to boolean variables
  char use_offside_str[5] = "off";
  char forbid_kick_off_offside_str[5] = "off";
  
	// Read server.conf file
	FILE *f = fopen("server.conf", "r");
	if (f) {
		char str[512];
		while (!feof(f)) {
			fgets(str, 512, f);
			sscanf(str, "goal_width : %lf", &GOAL_WIDTH);
			sscanf(str, "player_size : %lf", &PLAYER_SIZE);      
			sscanf(str, "player_decay : %lf", &PLAYER_DECAY);
			sscanf(str, "player_rand : %lf", &PLAYER_RAND);   
			sscanf(str, "player_weight : %lf", &PLAYER_WEIGHT);    
			sscanf(str, "player_speed_max : %lf", &PLAYER_SPEED_MAX); 
			sscanf(str, "stamina_max : %lf", &STAMINA_MAX);
			sscanf(str, "stamina_inc_max : %lf",  &STAMINA_INC_MAX);  
			sscanf(str, "recover_dec_thr : %lf", &RECOVER_DEC_THR);
			sscanf(str, "recover_dec : %lf", &RECOVER_DEC);     
			sscanf(str, "recover_min : %lf", &RECOVER_MIN);    
			sscanf(str, "effort_dec_thr : %lf", &EFFORT_DEC_THR);   
			sscanf(str, "effort_dec : %lf", &EFFORT_DEC); 
			sscanf(str, "effort_inc_thr : %lf", &EFFORT_INC_THR);   
			sscanf(str, "effort_inc : %lf", &EFFORT_INC);      
			sscanf(str, "effort_min : %lf", &EFFORT_MIN);       
			sscanf(str, "hear_max : %lf", &HEAR_MAX);         
			sscanf(str, "hear_inc : %lf", &HEAR_INC);         
			sscanf(str, "hear_decay : %lf", &HEAR_DECAY);       
			sscanf(str, "inertia_moment : %lf", &INERTIA_MOMENT);   
			sscanf(str, "sense_body_step : %d", &SENSE_BODY_STEP);  
			sscanf(str, "catchable_area_l : %lf", &CATCHABLE_AREA_L); 
			sscanf(str, "catchable_area_w : %lf", &CATCHABLE_AREA_W); 
			sscanf(str, "catch_probability : %lf", &CATCH_PROBABILITY);
			sscanf(str, "catch_ban_cycle : %d", &CATCH_BAN_CYCLE);
			sscanf(str, "goalie_max_moves : %d", &GOALIE_MAX_MOVES);
			sscanf(str, "ball_size : %lf", &BALL_SIZE);      
			sscanf(str, "ball_decay : %lf", &BALL_DECAY);     
			sscanf(str, "ball_rand : %lf", &BALL_RAND);      
			sscanf(str, "ball_weight : %lf", &BALL_WEIGHT);    
			sscanf(str, "ball_speed_max : %lf", &BALL_SPEED_MAX); 
			sscanf(str, "wind_force : %lf", &WIND_FORCE);     
			sscanf(str, "wind_dir : %lf", &WIND_DIR);       
			sscanf(str, "wind_rand : %lf", &WIND_RAND);      
			sscanf(str, "kickable_margin : %lf", &KICKABLE_MARGIN);
			sscanf(str, "ckick_margin : %lf", &CKICK_MARGIN);   
			sscanf(str, "dash_power_rate : %lf", &DASH_POWER_RATE);
			sscanf(str, "kick_power_rate : %lf", &KICK_POWER_RATE);
			sscanf(str, "visible_angle : %lf", &VISIBLE_ANGLE);  
			sscanf(str, "audio_cut_dist : %lf", &AUDIO_CUT_DIST); 
			sscanf(str, "quantize_step : %lf", &QUANTIZE_STEP);  
			sscanf(str, "quantize_step_l : %lf", &QUANTIZE_STEP_L);
			sscanf(str, "maxpower : %lf", &MAXPOWER);  
			sscanf(str, "minpower : %lf", &MINPOWER);  
			sscanf(str, "maxmoment : %lf", &MAXMOMENT); 
			sscanf(str, "minmoment : %lf", &MINMOMENT); 
			sscanf(str, "maxneckmoment : %lf", &MAXNECKMOMENT);
			sscanf(str, "minneckmoment : %lf", &MINNECKMOMENT);
			sscanf(str, "maxneckang : %lf", &MAXNECKANG);
			sscanf(str, "minneckang : %lf", &MINNECKANG);
			sscanf(str, "port : %d", &PORT);      
			sscanf(str, "coach_port : %d", &COACH_PORT);
			sscanf(str, "olcoach_port : %d", &OLCOACH_PORT);
			sscanf(str, "say_coach_cnt_max : %d", &SAY_COACH_CNT_MAX);   
			sscanf(str, "say_coach_msg_size : %d", &SAY_COACH_MSG_SIZE); 
			sscanf(str, "send_vi_step : %d", &SEND_VI_STEP);     
			sscanf(str, "simulator_step : %d", &SIMULATOR_STEP); 
			sscanf(str, "send_step : %d", &SEND_STEP);    
			sscanf(str, "recv_step : %d", &RECV_STEP);  
			sscanf(str, "half_time : %d", &HALF_TIME);  
			sscanf(str, "say_msg_size : %d", &SAY_MSG_SIZE);  
			sscanf(str, "use_offside : %3s", use_offside_str); 
			sscanf(str, "offside_active_area : %lf", &OFFSIDE_ACTIVE_AREA);
			sscanf(str, "forbid_kick_off_offside : %3s", forbid_kick_off_offside_str);
			sscanf(str, "offside_kick_margin : %lf", &OFFSIDE_KICK_MARGIN);
			sscanf(str, "slow_down_factor : %d", &SLOW_DOWN_FACTOR);
		}
		fclose(f);
	
		USE_OFFSIDE = bool(strncasecmp(use_offside_str, "off",3));
		FORBID_KICK_OFF_OFFSIDE = bool(strncasecmp(forbid_kick_off_offside_str, "off",3));
	// handle message size changes
		if (MAX_SAY_MSG_SIZE < SAY_MSG_SIZE) {
		  fprintf(stderr, "ERROR: SAY_MSG_SIZE in server.conf has changed. Change MAX_SAY_MSG_SIZE to %d in constants.h", MAX_SAY_MSG_SIZE);
		}
		if (MAX_SAY_MSG_SIZE < SAY_COACH_MSG_SIZE) {
		  fprintf(stderr, "ERROR: SAY_COACH_MSG_SIZE in server.conf has changed. Change MAX_SAY_MSG_SIZE to %d in constants.h", MAX_SAY_MSG_SIZE);
		}

	} else
		fprintf(stderr, "WARNING: server.conf not found, using defaults\n");


	// it should be like this:
	// HALF_TIME_IN_CYCLES = HALF_TIME * 1000 
	//   / SIMULATOR_STEP / SLOW_DOWN_FACTOR;
	
	// but in the soccerserver this formula is used:
	HALF_TIME_IN_CYCLES = HALF_TIME * 1000 / SIMULATOR_STEP;
	
	// set kick range
	KICKRANGE = BALL_SIZE + PLAYER_SIZE + KICKABLE_MARGIN;

	// Initialize flag positions
	const double PITCH_MARGIN = 5.0;
	flag_x[FLAG_BOTTOM_LEFT] = -FIELD_LENGTH / 2;
	flag_y[FLAG_BOTTOM_LEFT] = FIELD_WIDTH / 2;
	flag_x[FLAG_BOTTOM_CENTER] = 0;
	flag_y[FLAG_BOTTOM_CENTER] = FIELD_WIDTH / 2;
	flag_x[FLAG_BOTTOM_RIGHT] = FIELD_LENGTH / 2;
	flag_y[FLAG_BOTTOM_RIGHT] = FIELD_WIDTH / 2;
	flag_x[FLAG_TOP_LEFT] = -FIELD_LENGTH / 2;
	flag_y[FLAG_TOP_LEFT] = -FIELD_WIDTH / 2;
	flag_x[FLAG_TOP_CENTER] = 0;
	flag_y[FLAG_TOP_CENTER] = -FIELD_WIDTH / 2;
	flag_x[FLAG_TOP_RIGHT] = FIELD_LENGTH / 2;
	flag_y[FLAG_TOP_RIGHT] = -FIELD_WIDTH / 2;

	flag_x[FLAG_P_LEFT_BOTTOM] = -FIELD_LENGTH / 2 + PENALTY_AREA_LENGTH;
	flag_y[FLAG_P_LEFT_BOTTOM] = PENALTY_AREA_WIDTH / 2;
	flag_x[FLAG_P_LEFT_CENTER] = -FIELD_LENGTH / 2 + PENALTY_AREA_LENGTH;
	flag_y[FLAG_P_LEFT_CENTER] = 0;
	flag_x[FLAG_P_LEFT_TOP] = -FIELD_LENGTH / 2 + PENALTY_AREA_LENGTH;
	flag_y[FLAG_P_LEFT_TOP] = -PENALTY_AREA_WIDTH / 2;
	flag_x[FLAG_P_RIGHT_BOTTOM] = FIELD_LENGTH / 2 - PENALTY_AREA_LENGTH;
	flag_y[FLAG_P_RIGHT_BOTTOM] = PENALTY_AREA_WIDTH / 2;
	flag_x[FLAG_P_RIGHT_CENTER] = FIELD_LENGTH / 2 - PENALTY_AREA_LENGTH;
	flag_y[FLAG_P_RIGHT_CENTER] = 0;
	flag_x[FLAG_P_RIGHT_TOP] = FIELD_LENGTH / 2 - PENALTY_AREA_LENGTH;
	flag_y[FLAG_P_RIGHT_TOP] = -PENALTY_AREA_WIDTH / 2;

	flag_x[FLAG_CENTER] = 0;
	flag_y[FLAG_CENTER] = 0;

	flag_x[FLAG_GOAL_LEFT_BOTTOM] = -FIELD_LENGTH / 2;
	flag_y[FLAG_GOAL_LEFT_BOTTOM] = GOAL_WIDTH / 2;
	flag_x[FLAG_GOAL_LEFT_TOP] = -FIELD_LENGTH / 2;
	flag_y[FLAG_GOAL_LEFT_TOP] = -GOAL_WIDTH / 2;
	flag_x[FLAG_GOAL_RIGHT_BOTTOM] = FIELD_LENGTH / 2;
	flag_y[FLAG_GOAL_RIGHT_BOTTOM] = GOAL_WIDTH / 2;
	flag_x[FLAG_GOAL_RIGHT_TOP] = FIELD_LENGTH / 2;
	flag_y[FLAG_GOAL_RIGHT_TOP] = -GOAL_WIDTH / 2;

	flag_x[FLAG_BOTTOM_LEFT_50] = -50.0;
	flag_y[FLAG_BOTTOM_LEFT_50] = FIELD_WIDTH / 2 + PITCH_MARGIN;
	flag_x[FLAG_BOTTOM_LEFT_40] = -40.0;
	flag_y[FLAG_BOTTOM_LEFT_40] = FIELD_WIDTH / 2 + PITCH_MARGIN;
	flag_x[FLAG_BOTTOM_LEFT_30] = -30.0;
	flag_y[FLAG_BOTTOM_LEFT_30] = FIELD_WIDTH / 2 + PITCH_MARGIN;
	flag_x[FLAG_BOTTOM_LEFT_20] = -20.0;
	flag_y[FLAG_BOTTOM_LEFT_20] = FIELD_WIDTH / 2 + PITCH_MARGIN;
	flag_x[FLAG_BOTTOM_LEFT_10] = -10.0;
	flag_y[FLAG_BOTTOM_LEFT_10] = FIELD_WIDTH / 2 + PITCH_MARGIN;
	flag_x[FLAG_BOTTOM_0] = 0;
	flag_y[FLAG_BOTTOM_0] = FIELD_WIDTH / 2 + PITCH_MARGIN;
	flag_x[FLAG_BOTTOM_RIGHT_10] = 10.0;
	flag_y[FLAG_BOTTOM_RIGHT_10] = FIELD_WIDTH / 2 + PITCH_MARGIN;
	flag_x[FLAG_BOTTOM_RIGHT_20] = 20.0;
	flag_y[FLAG_BOTTOM_RIGHT_20] = FIELD_WIDTH / 2 + PITCH_MARGIN;
	flag_x[FLAG_BOTTOM_RIGHT_30] = 30.0;
	flag_y[FLAG_BOTTOM_RIGHT_30] = FIELD_WIDTH / 2 + PITCH_MARGIN;
	flag_x[FLAG_BOTTOM_RIGHT_40] = 40.0;
	flag_y[FLAG_BOTTOM_RIGHT_40] = FIELD_WIDTH / 2 + PITCH_MARGIN;
	flag_x[FLAG_BOTTOM_RIGHT_50] = 50.0;
	flag_y[FLAG_BOTTOM_RIGHT_50] = FIELD_WIDTH / 2 + PITCH_MARGIN;

	flag_x[FLAG_TOP_LEFT_50] = -50.0;
	flag_y[FLAG_TOP_LEFT_50] = -FIELD_WIDTH / 2 - PITCH_MARGIN;
	flag_x[FLAG_TOP_LEFT_40] = -40.0;
	flag_y[FLAG_TOP_LEFT_40] = -FIELD_WIDTH / 2 - PITCH_MARGIN;
	flag_x[FLAG_TOP_LEFT_30] = -30.0;
	flag_y[FLAG_TOP_LEFT_30] = -FIELD_WIDTH / 2 - PITCH_MARGIN;
	flag_x[FLAG_TOP_LEFT_20] = -20.0;
	flag_y[FLAG_TOP_LEFT_20] = -FIELD_WIDTH / 2 - PITCH_MARGIN;
	flag_x[FLAG_TOP_LEFT_10] = -10.0;
	flag_y[FLAG_TOP_LEFT_10] = -FIELD_WIDTH / 2 - PITCH_MARGIN;
	flag_x[FLAG_TOP_0] = 0;
	flag_y[FLAG_TOP_0] = -FIELD_WIDTH / 2 - PITCH_MARGIN;
	flag_x[FLAG_TOP_RIGHT_10] = 10.0;
	flag_y[FLAG_TOP_RIGHT_10] = -FIELD_WIDTH / 2 - PITCH_MARGIN;
	flag_x[FLAG_TOP_RIGHT_20] = 20.0;
	flag_y[FLAG_TOP_RIGHT_20] = -FIELD_WIDTH / 2 - PITCH_MARGIN;
	flag_x[FLAG_TOP_RIGHT_30] = 30.0;
	flag_y[FLAG_TOP_RIGHT_30] = -FIELD_WIDTH / 2 - PITCH_MARGIN;
	flag_x[FLAG_TOP_RIGHT_40] = 40.0;
	flag_y[FLAG_TOP_RIGHT_40] = -FIELD_WIDTH / 2 - PITCH_MARGIN;
	flag_x[FLAG_TOP_RIGHT_50] = 50.0;
	flag_y[FLAG_TOP_RIGHT_50] = -FIELD_WIDTH / 2 - PITCH_MARGIN;

	flag_x[FLAG_LEFT_BOTTOM_30] = -FIELD_LENGTH / 2 - PITCH_MARGIN;
	flag_y[FLAG_LEFT_BOTTOM_30] = 30.0;
	flag_x[FLAG_LEFT_BOTTOM_20] = -FIELD_LENGTH / 2 - PITCH_MARGIN;
	flag_y[FLAG_LEFT_BOTTOM_20] = 20.0;
	flag_x[FLAG_LEFT_BOTTOM_10] = -FIELD_LENGTH / 2 - PITCH_MARGIN;
	flag_y[FLAG_LEFT_BOTTOM_10] = 10.0;
	flag_x[FLAG_LEFT_0] = -FIELD_LENGTH / 2 - PITCH_MARGIN;
	flag_y[FLAG_LEFT_0] = 0;
	flag_x[FLAG_LEFT_TOP_10] = -FIELD_LENGTH / 2 - PITCH_MARGIN;
	flag_y[FLAG_LEFT_TOP_10] = -10.0;
	flag_x[FLAG_LEFT_TOP_20] = -FIELD_LENGTH / 2 - PITCH_MARGIN;
	flag_y[FLAG_LEFT_TOP_20] = -20.0;
	flag_x[FLAG_LEFT_TOP_30] = -FIELD_LENGTH / 2 - PITCH_MARGIN;
	flag_y[FLAG_LEFT_TOP_30] = -30.0;

	flag_x[FLAG_RIGHT_BOTTOM_30] = FIELD_LENGTH / 2 + PITCH_MARGIN;
	flag_y[FLAG_RIGHT_BOTTOM_30] = 30.0;
	flag_x[FLAG_RIGHT_BOTTOM_20] = FIELD_LENGTH / 2 + PITCH_MARGIN;
	flag_y[FLAG_RIGHT_BOTTOM_20] = 20.0;
	flag_x[FLAG_RIGHT_BOTTOM_10] = FIELD_LENGTH / 2 + PITCH_MARGIN;
	flag_y[FLAG_RIGHT_BOTTOM_10] = 10.0;
	flag_x[FLAG_RIGHT_0] = FIELD_LENGTH / 2 + PITCH_MARGIN;
	flag_y[FLAG_RIGHT_0] = 0;
	flag_x[FLAG_RIGHT_TOP_10] = FIELD_LENGTH / 2 + PITCH_MARGIN;
	flag_y[FLAG_RIGHT_TOP_10] = -10.0;
	flag_x[FLAG_RIGHT_TOP_20] = FIELD_LENGTH / 2 + PITCH_MARGIN;
	flag_y[FLAG_RIGHT_TOP_20] = -20.0;
	flag_x[FLAG_RIGHT_TOP_30] = FIELD_LENGTH / 2 + PITCH_MARGIN;
	flag_y[FLAG_RIGHT_TOP_30] = -30.0;

	// Initialize goal positions
	goal_x[GOAL_LEFT] = -FIELD_LENGTH / 2;
	goal_y[GOAL_LEFT] = 0;
	goal_x[GOAL_RIGHT] = FIELD_LENGTH / 2;
	goal_y[GOAL_RIGHT] = 0;
}
