//////////////////////////////////////////////////////////////////////
// Mainz Rolling Brains RoboCup agents developed for the Simulator 
// League of the RoboCup 2000 competition
//
// Copyright (C) 1998, 1999, 2000 Team Members 
// of the Mainz Rolling Brains RoboCup Team 
//
// This program under the terms of the GNU General Public License as
// published by the Free Software Foundation; version 2 of the
// License. See the README file in the root directory of our archive.
//////////////////////////////////////////////////////////////////////

//-*- c++ -*-/////////////////////////////////////////////////////////
//
//  $Id: parser.C,v 1.13 2000/10/19 15:27:47 goetz Exp $
// 
//  parser.C - parsing of server messages
// 
//////////////////////////////////////////////////////////////////////


#include "parser.h"

#undef TRACE
#define TRACE(cmd)


/////////////////////////////////////////////////////////////////////////


Parser::Parser(const char* _team, int _number, 
	       playing_side _side, bool _isCoach) {

  TRACE (cout << "Parser: entering constructor" << endl);
  if (_team) {
    team = new char [strlen(_team)+1];
    strcpy (team, _team);
  } else
    team = NULL;
  number = _number;
  side = _side;

  isCoach = _isCoach;
  TRACE (cout << "Parser: leaving constructor" << endl);

}

Parser::~Parser() {

  TRACE (cout << "Parser: entering destructor" << endl);

  delete[] team;

  TRACE (cout << "Parser: leaving destructor" << endl);

}


/////////////////////////////////////////////////////////////////////////

void Parser::setStates(const char* _team, int _number, 
		       playing_side _side, bool _isCoach) {

  TRACE (cout << "Parser: setting States" << endl);
  delete team;
  team = new char [strlen(_team)+1];
  strcpy (team, _team);
  number = _number;
  side = _side;

  isCoach = _isCoach;
  TRACE (cout << "Parser: setting States" << endl);
}


void 
Parser::makeValid(SensorObject &obj, 
		  double dist, double dir, 
		  double dist_change, double dir_change)
{
  obj.valid = true;
  obj.dist = dist;
  obj.dir = dir;
  obj.dist_change = dist_change;
  obj.dir_change = dir_change;
}

void 
Parser::makeValid(ExactObject &obj, 
		  double x, double y, 
		  double x_vel, double y_vel)
{
  obj.valid = true;
  obj.x = x;
  obj.y = y;
  obj.x_vel = x_vel;
  obj.y_vel = y_vel;
}


/////////////////////////////////////////////////////////////////////////


bool 
Parser::parseSee(const char *s, SensorData &data) {

  int i;

  TRACE(cout << "Parser:   parsing see data.." << endl);

  // Invalidate visual information
  data.ball.valid = false;
  for (i=0; i<NUM_GOALS; i++)
    data.goal[i].valid = false;
  for (i=0; i<21; i++)
    data.player[i].valid = false;
  for (i=0; i<NUM_FLAGS; i++)
    data.flag[i].valid = false;
  for (i=0; i<NUM_LINES; i++)
    data.line[i].valid = false;

  // Pointer to first free goal, player, flag and line
  SensorGoal *goal = data.goal;
  SensorPlayer *player = data.player;
  SensorFlag *flag = data.flag;
  SensorLine *line = data.line;

  // Parse object data
  while (*s != ')') {

    char object_name[MAX_SERVER_MSG_SIZE], *name = object_name;
    double dist = INVALID_DIST;
    double dir = INVALID_DIR;
    double dist_change = INVALID_DIST;
    double dir_change = INVALID_DIR;
    double body_dir = INVALID_DIR;
    double face_dir = INVALID_DIR;
    int len = 0;

    // team names may only contain the characters [-_a-zA-Z0-9], and NO ')' !
    // (see Noda's source code -> field.C)
    int actual = sscanf(s, "((%[^)])%n %lf %lf %lf %lf %lf %lf", 
			name, 
			&len, 
			&dist, &dir, 
			&dist_change, &dir_change, 
			&body_dir, &face_dir);
    
    // If only one parameter is given, it's the direction, not the distance
    // (when receiving low quality visual information)
    if (actual == 2) {
      dir = dist;
      dist = INVALID_DIST;
    }

    // Convert angles to radians
    if (IsDirValid(dir))
      dir = DEG_TO_RAD(dir);
    if (IsDirValid(dir_change))
      dir_change = DEG_TO_RAD(dir_change);
    if (IsDirValid(body_dir))
      body_dir = DEG_TO_RAD(body_dir);
    if (IsDirValid(face_dir))
      face_dir = DEG_TO_RAD(face_dir);

    // Convert object name to lower case
    object_name[0] = tolower(object_name[0]);

    // Determine object type
    if (strncmp(name, "b", 1) == 0) {

      // Ball
      makeValid(data.ball, dist, dir, dist_change, dir_change);

    }
    else if (strncmp(name, "g", 1) == 0) {

      // Goal
      makeValid(*goal, dist, dir, dist_change, dir_change);
      
      // Determine type
      name += 1;	// Skip "g"
      if (strcmp(name, " l") == 0) {
	
	goal->which = GOAL_LEFT;
	if (side == SIDE_LEFT) goal->team = TEAM_US;
	else goal->team = TEAM_THEM;

      } else if (strcmp(name, " r") == 0) {

	goal->which = GOAL_RIGHT;
	if (side == SIDE_RIGHT) goal->team = TEAM_US;
	else goal->team = TEAM_THEM;

      } else {

	goal->which = GOAL_UNKNOWN;
	goal->team = TEAM_UNKNOWN;

      }
      goal++;

    } else if (strncmp(name, "p", 1) == 0) {

      // Player, determine team and number
      makeValid(*player, dist, dir, dist_change, dir_change);
      player->number = -1;		// Invalid number
      char teamname[MAX_SERVER_MSG_SIZE];
      char goalie[MAX_SERVER_MSG_SIZE];
      teamname[0] = 0;
      goalie[0] = 0;
      sscanf(name, "p \"%[-_a-zA-Z0-9]\" %d %[a-z]", 
	     teamname, &player->number, goalie);
      
      if (strcmp(teamname, team) == 0) player->team = TEAM_US;
      else if (teamname[0]) player->team = TEAM_THEM;
      else player->team = TEAM_UNKNOWN;

      if (player->number != -1)
	if (strcmp(goalie, "goalie") == 0) player->goalie = GOALIE_TRUE;
	else player->goalie = GOALIE_FALSE;
      else player->goalie = GOALIE_UNKNOWN;
      
      player->body_dir = body_dir;
      player->face_dir = face_dir;
      player++;
      
    } else if (strncmp(name, "f", 1) == 0) {

      // Flag
      makeValid(*flag, dist, dir, dist_change, dir_change);

      // Determine type
      name += 1;	// Skip "f"
      if (strcmp(name, " l t") == 0)
	flag->which = FLAG_TOP_LEFT;
      else if (strcmp(name, " c t") == 0)
	flag->which = FLAG_TOP_CENTER;
      else if (strcmp(name, " r t") == 0)
	flag->which = FLAG_TOP_RIGHT;
      else if (strcmp(name, " l b") == 0)
	flag->which = FLAG_BOTTOM_LEFT;
      else if (strcmp(name, " c b") == 0)
	flag->which = FLAG_BOTTOM_CENTER;
      else if (strcmp(name, " r b") == 0)
	flag->which = FLAG_BOTTOM_RIGHT;

      else if (strcmp(name, " p l t") == 0)
	flag->which = FLAG_P_LEFT_TOP;
      else if (strcmp(name, " p l c") == 0)
	flag->which = FLAG_P_LEFT_CENTER;
      else if (strcmp(name, " p l b") == 0)
	flag->which = FLAG_P_LEFT_BOTTOM;
      else if (strcmp(name, " p r t") == 0)
	flag->which = FLAG_P_RIGHT_TOP;
      else if (strcmp(name, " p r c") == 0)
	flag->which = FLAG_P_RIGHT_CENTER;
      else if (strcmp(name, " p r b") == 0)
	flag->which = FLAG_P_RIGHT_BOTTOM;

      else if (strcmp(name, " c") == 0)
	flag->which = FLAG_CENTER;

      else if (strcmp(name, " g l t") == 0)
	flag->which = FLAG_GOAL_LEFT_TOP;
      else if (strcmp(name, " g l b") == 0)
	flag->which = FLAG_GOAL_LEFT_BOTTOM;
      else if (strcmp(name, " g r t") == 0)
	flag->which = FLAG_GOAL_RIGHT_TOP;
      else if (strcmp(name, " g r b") == 0)
	flag->which = FLAG_GOAL_RIGHT_BOTTOM;

      else if (strcmp(name, " t l 50") == 0)
	flag->which = FLAG_TOP_LEFT_50;
      else if (strcmp(name, " t l 40") == 0)
	flag->which = FLAG_TOP_LEFT_40;
      else if (strcmp(name, " t l 30") == 0)
	flag->which = FLAG_TOP_LEFT_30;
      else if (strcmp(name, " t l 20") == 0)
	flag->which = FLAG_TOP_LEFT_20;
      else if (strcmp(name, " t l 10") == 0)
	flag->which = FLAG_TOP_LEFT_10;
      else if (strcmp(name, " t 0") == 0)
	flag->which = FLAG_TOP_0;
      else if (strcmp(name, " t r 10") == 0)
	flag->which = FLAG_TOP_RIGHT_10;
      else if (strcmp(name, " t r 20") == 0)
	flag->which = FLAG_TOP_RIGHT_20;
      else if (strcmp(name, " t r 30") == 0)
	flag->which = FLAG_TOP_RIGHT_30;
      else if (strcmp(name, " t r 40") == 0)
	flag->which = FLAG_TOP_RIGHT_40;
      else if (strcmp(name, " t r 50") == 0)
	flag->which = FLAG_TOP_RIGHT_50;

      else if (strcmp(name, " b l 50") == 0)
	flag->which = FLAG_BOTTOM_LEFT_50;
      else if (strcmp(name, " b l 40") == 0)
	flag->which = FLAG_BOTTOM_LEFT_40;
      else if (strcmp(name, " b l 30") == 0)
	flag->which = FLAG_BOTTOM_LEFT_30;
      else if (strcmp(name, " b l 20") == 0)
	flag->which = FLAG_BOTTOM_LEFT_20;
      else if (strcmp(name, " b l 10") == 0)
	flag->which = FLAG_BOTTOM_LEFT_10;
      else if (strcmp(name, " b 0") == 0)
	flag->which = FLAG_BOTTOM_0;
      else if (strcmp(name, " b r 10") == 0)
	flag->which = FLAG_BOTTOM_RIGHT_10;
      else if (strcmp(name, " b r 20") == 0)
	flag->which = FLAG_BOTTOM_RIGHT_20;
      else if (strcmp(name, " b r 30") == 0)
	flag->which = FLAG_BOTTOM_RIGHT_30;
      else if (strcmp(name, " b r 40") == 0)
	flag->which = FLAG_BOTTOM_RIGHT_40;
      else if (strcmp(name, " b r 50") == 0)
	flag->which = FLAG_BOTTOM_RIGHT_50;

      else if (strcmp(name, " l t 30") == 0)
	flag->which = FLAG_LEFT_TOP_30;
      else if (strcmp(name, " l t 20") == 0)
	flag->which = FLAG_LEFT_TOP_20;
      else if (strcmp(name, " l t 10") == 0)
	flag->which = FLAG_LEFT_TOP_10;
      else if (strcmp(name, " l 0") == 0)
	flag->which = FLAG_LEFT_0;
      else if (strcmp(name, " l b 10") == 0)
	flag->which = FLAG_LEFT_BOTTOM_10;
      else if (strcmp(name, " l b 20") == 0)
	flag->which = FLAG_LEFT_BOTTOM_20;
      else if (strcmp(name, " l b 30") == 0)
	flag->which = FLAG_LEFT_BOTTOM_30;

      else if (strcmp(name, " r t 30") == 0)
	flag->which = FLAG_RIGHT_TOP_30;
      else if (strcmp(name, " r t 20") == 0)
	flag->which = FLAG_RIGHT_TOP_20;
      else if (strcmp(name, " r t 10") == 0)
	flag->which = FLAG_RIGHT_TOP_10;
      else if (strcmp(name, " r 0") == 0)
	flag->which = FLAG_RIGHT_0;
      else if (strcmp(name, " r b 10") == 0)
	flag->which = FLAG_RIGHT_BOTTOM_10;
      else if (strcmp(name, " r b 20") == 0)
	flag->which = FLAG_RIGHT_BOTTOM_20;
      else if (strcmp(name, " r b 30") == 0)
	flag->which = FLAG_RIGHT_BOTTOM_30;

      else
	flag->which = FLAG_UNKNOWN;
      flag++;

    } else if (strncmp(name, "l", 1) == 0) {

      // Line
      makeValid(*line, dist, dir, dist_change, dir_change);

      // Determine type
      name += 1;	// Skip "l"
      if (strcmp(name, " t") == 0)
	line->which = LINE_TOP;
      else if (strcmp(name, " l") == 0)
	line->which = LINE_LEFT;
      else if (strcmp(name, " r") == 0)
	line->which = LINE_RIGHT;
      else if (strcmp(name, " b") == 0)
	line->which = LINE_BOTTOM;
      else
	line->which = LINE_UNKNOWN;
      line++;

    } else return false; // illegal object name

    // Skip to start of next object description
    s += len;
    while (*s && *s != ')') s++;
    s++;
    while (*s == ' ') s++;
  }
  
  return true;
}


bool 
Parser::parseExact(const char *s, SensorData &data) {

  int i;

  TRACE(cout << "Parser:   parsing coach look reply or a coach see data.." 
	<< endl);
  
  // Invalidate look information
  data.exactball.valid = false;
  for (i=0; i<22; i++)
    data.exactplayer[i].valid = false;

  // Pointer to first free player and goal
  ExactPlayer *player = data.exactplayer;
  ExactGoal   *goal   = data.exactgoal; 
  
  // Parse object data
  while (*s != ')') {

    char object_name[MAX_SERVER_MSG_SIZE], *name = object_name;
    double x;
    double y;
    double body_dir = INVALID_DIR;
    double face_dir = INVALID_DIR;
    double vx = 0;
    double vy = 0;
    int len = 0;

    // team names may only contain the characters [-_a-zA-Z0-9], and NO ')' !
    // (see Noda's source code -> field.C)
    int actual = sscanf(s, "((%[^)])%n %lf %lf %lf %lf %lf %lf", name, 
			&len, &x, &y, &body_dir, &face_dir, &vx, &vy);
    
    // If only four parameters are given, the dirs are the velocities !
    // (the ball)
    if (actual == 5) {
      vx = body_dir;
      body_dir = INVALID_DIR;
      vy = face_dir;
      face_dir = INVALID_DIR;
    }

    // Convert angles to radians
    if (IsDirValid(body_dir))
      body_dir = DEG_TO_RAD(body_dir);
    if (IsDirValid(face_dir))
      face_dir = DEG_TO_RAD(face_dir);

    // Convert object name to lower case
    object_name[0] = tolower(object_name[0]);

    // Determine object type
    if (strncmp(name, "ball", 4) == 0) {

      // Ball
      makeValid(data.exactball, x, y, vx, vy);

    } 
    else if (strncmp(name, "goal", 4) == 0) {

      // Goal
      makeValid(*goal, x, y, vx, vy);

      // Determine type
      name += 4;	// Skip "g"
      if (strcmp(name, " l") == 0) {

	goal->which = GOAL_LEFT;
	if (side == SIDE_LEFT) goal->team = TEAM_US;
	else goal->team = TEAM_THEM;

      } else if (strcmp(name, " r") == 0) {

	goal->which = GOAL_RIGHT;
	if (side == SIDE_RIGHT) goal->team = TEAM_US;
	else goal->team = TEAM_THEM;

      } else {

	goal->which = GOAL_UNKNOWN;
	goal->team = TEAM_UNKNOWN;

      }
      goal++;

    } else if (strncmp(name, "player", 6) == 0) {

      // Player, determine team and number
      makeValid(*player, x, y, vx, vy);
      player->number = -1;		// Invalid number
      char teamname[MAX_SERVER_MSG_SIZE];
      teamname[0] = 0;
      sscanf(name, "player %[-_a-zA-Z0-9] %d", 
	     teamname, &player->number);

      if (strcmp(teamname, team) == 0) player->team = TEAM_US;
      else if (teamname[0]) player->team = TEAM_THEM;
      else player->team = TEAM_UNKNOWN;

      // coaches do not receive goalie flag
      player->goalie = GOALIE_UNKNOWN;

      player->body_dir = body_dir;
      player->face_dir = face_dir;

      player++;
      
    } else
      return false;
    // Skip to start of next object description
    s += len;
    while (*s && *s != ')') s++;
    s++;
    while (*s == ' ') s++;
  }
  
  return true;
}


bool 
Parser::parseHear(const char *s, SensorData &data) {

  TRACE(cout << "Parser:   parsing heard message.." << endl);

  // Invalidate aural information
  data.message[0] = 0;
  data.mode = PLAY_MODE_UNKNOWN;
  data.sender_dir = INVALID_DIR;
  data.sender_name[0] = 0;
  
  // Determine sender type and the actual message
  if (sscanf(s, "referee %[-0-9a-zA-Z .+*/?<>_]", data.message) == 1) {
    // could be either play mode, or a relayed message from the coach

    data.mode = parsePlayMode(data.message);

    if (data.mode == PLAY_MODE_UNKNOWN) {
      TRACE (cout << "Parser:   message from the coach" << endl);
      data.sender = SENDER_COACH;
    } else {
      TRACE (cout << "Parser:   message from the referee" << endl);
      data.sender = SENDER_REFEREE;
      if ( (data.mode == PLAY_MODE_GOAL_US) ||
	   (data.mode == PLAY_MODE_GOAL_THEM) ) {
	if (!parseGoalScore(data.message, data.score)) return false;
      }
    }
    
  } else if (sscanf(s, "self %[-0-9a-zA-Z .+*/?<>_]", data.message) == 1) {
    
    data.sender = SENDER_SELF;
    TRACE (cout << "Parser:   message from self" << endl);

  } else if (sscanf(s, "%lf %[-0-9a-zA-Z .+*/?<>_]", 
		    &data.sender_dir, data.message) == 2) {

    data.sender = SENDER_DIR;
    data.sender_dir = DEG_TO_RAD(data.sender_dir);
    TRACE (cout << "Parser:   message from a certain direction" << endl);


  } else if (sscanf(s, "(%[-0-9a-zA-Z .+*/?<>_]) %[-0-9a-zA-Z .+*/?<>_]", 
		    data.sender_name, data.message) == 2) {

    data.sender = SENDER_PLAYER;
    TRACE (cout << "Parser:   message from a named player" << endl);
	   
  } else if (sscanf(s, "online_coach_left %[-0-9a-zA-Z .+*/?<>_]", 
		    data.message) == 1) {

    data.sender = SENDER_ONLINECOACH;
    TRACE (cout << "Parser:   message from onlinecoach left" << endl);

  } else if (sscanf(s, "online_coach_right %[-0-9a-zA-Z .+*/?<>_]", 
		    data.message) == 1) {
    
    data.sender = SENDER_ONLINECOACH;
    TRACE (cout << "Parser:   message from an onlinecoach right" << endl);

  } else return false;

  return true;

}


bool 
Parser::parseSenseBody(const char *s, SensorData &data) {

  TRACE(cout << "Parser:   parsing sense body data.." << endl);

  char quality[MAX_SAY_MSG_SIZE];
  char width[MAX_SAY_MSG_SIZE];
  
  if (sscanf(s, "(view_mode %[a-z] %[a-z]) (stamina %lf %lf) (speed %lf %lf) (head_angle %lf) (kick %d) (dash %d) (turn %d) (say %d) (turn_neck %d) ()",
	     quality, width, 
	     &data.stamina, &data.effort, 
	     &data.speed,
	     &data.speed_angle,
	     &data.head_angle, 
	     &data.kick_count, &data.dash_count, &data.turn_count, 
	     &data.say_count, &data.turn_neck_count) == 12)
    
    {

    // Parse view quality
    if (strcmp(quality, "high") == 0) 
      data.my_view_quality = VIEW_QUALITY_HIGH;
    else if (strcmp(quality, "low") == 0) 
      data.my_view_quality = VIEW_QUALITY_LOW;
    else return false;

    // Parse view width
    if (strcmp(width, "narrow") == 0)
      data.my_view_width = VIEW_WIDTH_NARROW;
    else if (strcmp(width, "normal") == 0)
      data.my_view_width = VIEW_WIDTH_NORMAL;
    else if (strcmp(width, "wide") == 0)
      data.my_view_width = VIEW_WIDTH_WIDE;
    else return false;

    // transform head angle into radians
    data.head_angle = DEG_TO_RAD(data.head_angle);

    // transform speed angle into radians
    data.speed_angle = DEG_TO_RAD(data.speed_angle);

    return true;

    } else return false;

}

play_mode 
Parser::parsePlayMode(const char *s) {

  TRACE(cout << "Parser:   parsing play mode.." << endl);

  if (strcmp(s, "before_kick_off") == 0)
    return PLAY_MODE_BEFORE_KICK_OFF;
  else if (strcmp(s, "time_over") == 0)
    return PLAY_MODE_TIME_OVER;
  else if (strcmp(s, "play_on") == 0)
    return PLAY_MODE_PLAY_ON;
  else if (strcmp(s, "kick_off_l") == 0)
    return side == SIDE_LEFT ? PLAY_MODE_KICK_OFF_US : PLAY_MODE_KICK_OFF_THEM;
  else if (strcmp(s, "kick_off_r") == 0)
    return side == SIDE_RIGHT ? PLAY_MODE_KICK_OFF_US : PLAY_MODE_KICK_OFF_THEM;
  else if (strcmp(s, "kick_in_l") == 0)
    return side == SIDE_LEFT ? PLAY_MODE_KICK_IN_US : PLAY_MODE_KICK_IN_THEM;
  else if (strcmp(s, "kick_in_r") == 0)
    return side == SIDE_RIGHT ? PLAY_MODE_KICK_IN_US : PLAY_MODE_KICK_IN_THEM;
  else if (strcmp(s, "free_kick_l") == 0)
    return side == SIDE_LEFT ? PLAY_MODE_FREE_KICK_US : PLAY_MODE_FREE_KICK_THEM;
  else if (strcmp(s, "free_kick_r") == 0)
    return side == SIDE_RIGHT ? PLAY_MODE_FREE_KICK_US : PLAY_MODE_FREE_KICK_THEM;
  else if (strcmp(s, "corner_kick_l") == 0)
    return side == SIDE_LEFT ? PLAY_MODE_CORNER_KICK_US : PLAY_MODE_CORNER_KICK_THEM;
  else if (strcmp(s, "corner_kick_r") == 0)
    return side == SIDE_RIGHT ? PLAY_MODE_CORNER_KICK_US : PLAY_MODE_CORNER_KICK_THEM;
  else if (strcmp(s, "goal_kick_l") == 0)
    return side == SIDE_LEFT ? PLAY_MODE_GOAL_KICK_US : PLAY_MODE_GOAL_KICK_THEM;
  else if (strcmp(s, "goal_kick_r") == 0)
    return side == SIDE_RIGHT ? PLAY_MODE_GOAL_KICK_US : PLAY_MODE_GOAL_KICK_THEM;
  else if (strcmp(s, "drop_ball") == 0)
    return PLAY_MODE_DROP_BALL;
  else if (strcmp(s, "offside_l") == 0)
    return side == SIDE_LEFT ? PLAY_MODE_OFFSIDE_US : PLAY_MODE_OFFSIDE_THEM;
  else if (strcmp(s, "offside_r") == 0)
    return side == SIDE_RIGHT ? PLAY_MODE_OFFSIDE_US : PLAY_MODE_OFFSIDE_THEM;
  else if (strcmp(s, "half_time") == 0)
    return PLAY_MODE_HALF_TIME;
  else if (strcmp(s, "time_extended") == 0)
    return PLAY_MODE_TIME_EXTENDED;
  else if (strncmp(s, "goal_l", 6) == 0)
    return side==SIDE_LEFT ? PLAY_MODE_GOAL_US : PLAY_MODE_GOAL_THEM;
  else if (strncmp(s, "goal_r", 6) == 0)
    return side==SIDE_RIGHT ? PLAY_MODE_GOAL_US : PLAY_MODE_GOAL_THEM;
  else return PLAY_MODE_UNKNOWN;
}

bool
Parser::parseGoalScore(const char* s, int &score) {
  char side;
  if (sscanf (s, "goal_%c_%d", &side, &score) == 2) return true;
  else return false;
}


/////////////////////////////////////////////////////////////////////////


bool 
Parser::parseServerString(const char *s, SensorData &data) {
  
  TRACE(cout << "Parser: parsing server string.." << endl;)

  int len = 0;

  // Switch on message type
  if (sscanf(s, "(see %d %n", &data.time, &len) == 1) {

    if (isCoach) { // coach receives different (see) information !
      data.type = SENSOR_TYPE_EXACT;
      return parseExact(s + len, data);
    }
    else { // a normal player clients visual information
      data.type = SENSOR_TYPE_SEE;
      return parseSee(s + len, data);
    }

  } else if (sscanf(s, "(hear %d %n", &data.time, &len) == 1) {

    // Aural information from referee, players or coach
    data.type = SENSOR_TYPE_HEAR;
    return parseHear(s + len, data);

  } else if (sscanf(s, "(sense_body %d %n", &data.time, &len) == 1) {

    // Sense body information
    data.type = SENSOR_TYPE_SENSE_BODY;
    return parseSenseBody(s + len, data);

  } else if (sscanf(s, "(ok look %d %n", &data.time, &len) == 1) {
    
    // exact information as requested by coach
    data.type = SENSOR_TYPE_EXACT;
    return parseExact (s+len, data);

  } else if (sscanf(s, "(ok %n", &len) == 0) {
    
    // other ok replies to coach not parsed any further
    data.time = -1;
    data.type = SENSOR_TYPE_OK;
    return true; // parsing was successful
    
  } else if (sscanf(s, "(error %n", &len) == 0) {
    
    // all error replies by server will be prompted
    data.time = -1;
    data.type = SENSOR_TYPE_ERROR;

    cerr << "Parser (" << team << " No." << number
	 << ") received error message: " << s << endl;

    return true;  // parsing was successful
    
  } else return false;
 
}
