//////////////////////////////////////////////////////////////////////
// Mainz Rolling Brains RoboCup agents developed for the Simulator 
// League of the RoboCup 2000 competition
//
// Copyright (C) 1998, 1999, 2000 Team Members 
// of the Mainz Rolling Brains RoboCup Team 
//
// This program under the terms of the GNU General Public License as
// published by the Free Software Foundation; version 2 of the
// License. See the README file in the root directory of our archive.
//////////////////////////////////////////////////////////////////////

// -*- c++ -*-////////////////////////////////////////////////////////
//
// $Id: player.C,v 1.12 2000/10/19 15:27:47 goetz Exp $
//
// player.C - interface for player client
//
///////////////////////////////////////////////////////////////////////


#include "player.h"

#undef TRACE
#define TRACE(cmd)

Player::Player (const char* team_name,             
		bool        isGoalie, 
		const char* host,
		int         port,
		double      MapWindowMagnify,
		bool        logging_snap,
		int         reconnectNumber,
		int         numSnapshots) :

  skills(NULL), 
  comm(NULL),
  world(NULL),
  mapwindow(NULL) {

  TRACE (cout << "Player: reading constants" << endl);

  if (MapWindowMagnify>0)
    mapwindow = new WorldMap_r("MRB Worldmodel display", MapWindowMagnify);
  else mapwindow = NULL;
  
  TRACE (cout << "Player: generating Comm" << endl);
  
  comm = new Comm(team_name);

  TRACE (cout << "Player: initializing Comm" << endl);
  
  comm_error ce;
  if (isGoalie) 
    ce = comm->init(CLIENT_GOALIE, host, port, 
		    reconnectNumber, logging_snap);
  else 
    ce = comm->init(CLIENT_FIELDPLAYER, host, port, 
		    reconnectNumber, logging_snap);
  
  switch (ce) {
  case (COMM_OK): 
    
    TRACE (cout << "Player: generating Worldmodel and AdvSkills" << endl);
    
    world  = new Worldmodel (numSnapshots, comm);
    skills = new AdvSkills (comm);
    
    writeSnaps = false;
    if (logging_snap) {
      writeSnaps = true;
      // open snap_log file
      ostrstream tmp;
      tmp << SNAP_LOG_FILENAME << getMyTeam() << getMyNumber() 
	  << SNAP_LOG_EXTENSION;
      char* filename = tmp.str();
      out.open(filename, ios::out);
      delete filename;
    }
    
    break;
  case (COMM_ERROR): 
    cerr << "Player: unspecified error when initializing !" << endl; break;
  case (COMM_ERROR_SERVER_NOT_FOUND):
    cerr << "Player: could not find server's host !"<< endl; break;
  case (COMM_ERROR_SOCKET_INIT):
    cerr << "Player: could not create and bind socket !" << endl; break;
  case (COMM_ERROR_SERVER_INIT):
    cerr << "Player: could not deliver checkin message to server !" 
	 << endl; break;
  case (COMM_ERROR_SIGNAL_INIT):
    cerr << "Player: could not set up signal handlrs !" << endl; break;
  case (COMM_ERROR_PARSE):
    cerr << "Player: could not parse init response !" << endl; break;
  case (COMM_ERROR_NO_DATA):
    cerr << "Player: did not receive response from server !" << endl; break;
  }
   
  if (mapwindow) {
    
    TRACE (cout << "Player: opening Mapwindow" << endl);
    mapwindow->Open();

  }
  
} 


Player::~Player () {

  TRACE (cout << "Player: entering destructor, closing down" << endl);

  if (mapwindow) {
    mapwindow->Close();
    delete(mapwindow);
  }

  delete (skills);
  delete (world);
  delete (comm);
  
  if (writeSnaps) out.close();

  TRACE (cout << "Player: leaving destructor" << endl);

}


void 
Player::play() {

  if (comm->isServerAlive()) {
    TRACE (cout << "Player: started" << endl);
    
    while ( comm->isServerAlive() && !comm->isGameOver() ) {
      
      TRACE (cout << "Player: waiting for signals and handling them" << endl);
      comm->handleSignals();
      
      if (comm->isActionNeeded()) {
	
	TRACE (cout << "Player: need action for time " << getTime() << endl); 
	
	TRACE (cout << "Player: adjusting worldmodel" << endl);
	world->adjust();
	
	TRACE (cout << "Player: getting snapshot of time " 
	       << getTime() << endl);

	TRACE (cout << "Player: acting" << endl);
	
	act();

	Snapshot* snap = world->getSnapshot(getTime());
	
	if (snap == NULL) {
	  cerr << "Player (" << getMyTeam() << " No." << getMyNumber()
	       <<"): cannot generate Snapshot of time " << getTime() << endl;
	} else {
	  
	  if (mapwindow) mapwindow->Map(snap, MAX_MAP_AGE);
	  
	  if ( (writeSnaps) && comm->isTimeRunning() ) {
	    snap->binaryWriteToStream(out);
	  }
	  
	} // got a snapshot
	
      } else {
	TRACE(cout << "Player: do not need new action yet" << endl);
      }
      
    } // while server alive
    
    TRACE (cout << "Player: server is gone, leaving" << endl);
    
  } else { 
    // server never was alive
    cerr << "Player: could not connect to sserver, closing down" << endl;
  }
  
} //play
