//////////////////////////////////////////////////////////////////////
// Mainz Rolling Brains RoboCup agents developed for the Simulator 
// League of the RoboCup 2000 competition
//
// Copyright (C) 1998, 1999, 2000 Team Members 
// of the Mainz Rolling Brains RoboCup Team 
//
// This program under the terms of the GNU General Public License as
// published by the Free Software Foundation; version 2 of the
// License. See the README file in the root directory of our archive.
//////////////////////////////////////////////////////////////////////

//////////////////////////////////////////////////////////////////////
//
// $Id: worldmap.C,v 1.18 2000/10/19 15:27:48 goetz Exp $	
// $Manager: goetz$
// 
// 
//  worldmap.C - World model display
// 
//  ( Christian Bauer <cebix> )
//    Goetz Schwandtner
//
//
//////////////////////////////////////////////////////////////////////

#include <stdio.h>

#include "worldmap.h"

#include "constants.h"
#include "coordinates.h"

#define BALL_TRACE(printcommand) 

// Internal constants

static char *num2str[] = {"? ", "1 ", "2 ", "3 ", "4 ", "5 ",
			 "6 ", "7 ", "8 ", "9 ", "10", "11"};
static char *defaulttitle ="MRB Worldmodel display";

static char *fontname="fixed";

/*
 *  Constructor and destructor
 */

WorldMap::WorldMap(const char *windowtitle, double mag) :
  MAGNIFY(mag),
  VELOCITY_MAG(10.0),
  PLAYER_MAG(3.0),
  PSIZE(int(PLAYER_SIZE * MAGNIFY * PLAYER_MAG)),
  VIEW_CONE_LENGTH(PSIZE * 150),
  MAX_AGE_AS_NOW(1),
  WIN_WIDTH(int(FIELD_LENGTH * MAGNIFY)),
  WIN_HEIGHT(int(FIELD_WIDTH * MAGNIFY))
{
  if (windowtitle) {
    title = new char[strlen(windowtitle)+1];
    strcpy(title, windowtitle);
  } else {
    title = new char[strlen(defaulttitle)+1];
    strcpy(title, defaulttitle);
  }
  the_disp = NULL;
}

WorldMap::~WorldMap()
{
  if (title)
    delete [] title;
  // close window if not done (close checks this)
  Close();
}

/*
 *  Open map window
 */

bool WorldMap::Open()
{
	// Establish connection to X server
	the_disp = XOpenDisplay(NULL);
	if (the_disp == NULL) {
		printf("CebixGraph FATAL: Can't connect to X server\n");
		return false;
	}
	the_screen = DefaultScreen(the_disp);
	black = BlackPixel(the_disp, the_screen);
	white = WhitePixel(the_disp, the_screen);
	the_root = RootWindow(the_disp, the_screen);
	the_cmap = DefaultColormap(the_disp, the_screen);

	// Open window
	the_win = XCreateSimpleWindow(the_disp, the_root, 0, 0, WIN_WIDTH, WIN_HEIGHT, 0, black, white);
	XStoreName(the_disp, the_win, title);
	XMapRaised(the_disp, the_win);

	// Create pixmap and attach to window
	the_map = XCreatePixmap(the_disp, the_win, WIN_WIDTH, WIN_HEIGHT, DefaultDepth(the_disp, the_screen));
	XSetWindowBackgroundPixmap(the_disp, the_win, the_map);

	// Create GCs
	win_gc = XCreateGC(the_disp, the_win, 0, NULL);
	map_gc = XCreateGC(the_disp, the_map, 0, NULL);
	Font font = XLoadFont(the_disp, fontname);
	XSetFont(the_disp, map_gc, font);

	// Get colors
	XParseColor(the_disp, the_cmap, "rgb:c/2/2", &red_color);
	XAllocColor(the_disp, the_cmap, &red_color);
	red = red_color.pixel;
	XParseColor(the_disp, the_cmap, "rgb:f/6/6", &lred_color);
	XAllocColor(the_disp, the_cmap, &lred_color);
	lred = lred_color.pixel;
	XParseColor(the_disp, the_cmap, "rgb:0/b/0", &green_color);
	XAllocColor(the_disp, the_cmap, &green_color);
	green = green_color.pixel;
	XParseColor(the_disp, the_cmap, "rgb:4/f/4", &lgreen_color);
	XAllocColor(the_disp, the_cmap, &lgreen_color);
	lgreen = lgreen_color.pixel;
	XParseColor(the_disp, the_cmap, "rgb:A/A/f", &blue_color);
	XAllocColor(the_disp, the_cmap, &blue_color);
	blue = blue_color.pixel;
	XParseColor(the_disp, the_cmap, "rgb:c/c/0", &yellow_color);
	XAllocColor(the_disp, the_cmap, &yellow_color);
	yellow = yellow_color.pixel;
	XParseColor(the_disp, the_cmap, "rgb:f/f/2", &lyellow_color);
	XAllocColor(the_disp, the_cmap, &lyellow_color);
	lyellow = lyellow_color.pixel;
	XParseColor(the_disp, the_cmap, "rgb:2/2/2", &darkgrey_color);
	XAllocColor(the_disp, the_cmap, &darkgrey_color);
	darkgrey = darkgrey_color.pixel;
	XParseColor(the_disp, the_cmap, "rgb:4/4/4", &lightgrey_color);
	XAllocColor(the_disp, the_cmap, &lightgrey_color);
	lightgrey = lightgrey_color.pixel;

	XParseColor(the_disp, the_cmap, "rgb:f/0/f", &pink_color);
	XAllocColor(the_disp, the_cmap, &pink_color);
	pink = pink_color.pixel;
	XParseColor(the_disp, the_cmap, "rgb:b/0/b", &dpink_color);
	XAllocColor(the_disp, the_cmap, &dpink_color);
	dpink = dpink_color.pixel;

	// Wait until window is ready
	XSync(the_disp, false);
	return(true);
}


/*
 *  Close map window
 */

void WorldMap::Close()
{
	if (the_disp != NULL) {
		XFreeGC(the_disp, map_gc);
		XFreeGC(the_disp, win_gc);
		XDestroyWindow(the_disp, the_win);
		XCloseDisplay(the_disp);
		the_disp = NULL;
	}
}


/*
 *  Draw snapshot
 */

void WorldMap::Map(Snapshot *snap, int max_age)
{
	if (the_disp == NULL)
	  return;

	if (snap == NULL)
	  return;

	// Debug information
	BALL_TRACE( printf("worldmap: (ball %f %f %d) \n",
			   snap->ball.pos.getX(), snap->ball.pos.getY(), snap->ball.age); )

	// Clear pixmap
	XSetForeground(the_disp, map_gc, black);
	XFillRectangle(the_disp, the_map, map_gc, 0, 0, WIN_WIDTH, WIN_HEIGHT);

	// Draw viewcone (not for coach!)
	if (snap->my.number > 0)
	  drawviewcone(snap->my.pos, snap->my.faceDir,
		       snap->senseBody.getViewAngle(), snap->senseBody.viewQuality);
	
	// Draw lines of field
	drawlines();

	// Draw myself (don't assume coach as me in the midde of the field!)
	if (snap->my.number > 0)
	  drawownplayer(snap->my);

	// Draw ball
	drawball(snap->ball);
	
	// Draw my team
	drawmyteam(snap, max_age);
	
	// Draw other team
	drawotherteam(snap, max_age);
	    
	// Draw pixmap
	XClearWindow(the_disp, the_win);
	XFlush(the_disp);
}

void WorldMap::drawCircle(const double x, const double y, const double rad, 
			  const bool lines, const bool bright)
{
  if (rad <= 0) return;
  
  if (bright)
    XSetForeground(the_disp, map_gc, pink);
  else
    XSetForeground(the_disp, map_gc, dpink);

  XDrawArc(the_disp, the_map, map_gc, 
	   mx(x-rad), my(y-rad),
	   md(2*rad), md(2*rad),
	   0, 360 * 64);

  if (lines) {
  XDrawLine(the_disp, the_map, map_gc,
	    mx(x), my(y-rad),
	    mx(x), my(y+rad));
  XDrawLine(the_disp, the_map, map_gc,
	    mx(x-rad), my(y),
	    mx(x+rad), my(y));
  }

  // Draw pixmap
  XClearWindow(the_disp, the_win);
  XFlush(the_disp);
}

void WorldMap::drawLine(const double x1, const double y1,
			const double x2, const double y2,
			const bool bright)
{
  if (bright)
    XSetForeground(the_disp, map_gc, pink);
  else
    XSetForeground(the_disp, map_gc, dpink);

  XDrawLine(the_disp, the_map, map_gc,
	    mx(x1), my(y1),
	    mx(x2), my(y2));

  // Draw pixmap
  XClearWindow(the_disp, the_win);
  XFlush(the_disp);
}

void WorldMap::drawText(const double x, const double y, 
			const char *text, const bool bright)
{
  if (bright)
    XSetForeground(the_disp, map_gc, pink);
  else
    XSetForeground(the_disp, map_gc, dpink);
  
  XDrawString(the_disp, the_map, map_gc,
	      mx(x), my(y), text, strlen(text));
  
  // Draw pixmap
  XClearWindow(the_disp, the_win);
  XFlush(the_disp);
}


void WorldMap::drawviewcone(Vector2d pos, Angle orient, 
			    Angle viewangle, view_quality viewqu)
{
	XPoint viewcone[4];  // Points for drawing viewcone as polygon
	int x = mx(pos.getX());
	int y = my(pos.getY());
 
		// Draw viewcone (grey)
	// set viewcone points: origin
	viewcone[0].x = x;
	viewcone[0].y = y;
	
	// set left and right margin points of view cone

	viewcone[1].x = x + md(cos(orient + viewangle) * VIEW_CONE_LENGTH);
	viewcone[1].y = y + md(sin(orient + viewangle) * VIEW_CONE_LENGTH);
	viewcone[2].x = x + md(cos(orient) * VIEW_CONE_LENGTH);
	viewcone[2].y = y + md(sin(orient) * VIEW_CONE_LENGTH);
	viewcone[3].x = x + md(cos(orient - viewangle) * VIEW_CONE_LENGTH);
	viewcone[3].y = y + md(sin(orient - viewangle) * VIEW_CONE_LENGTH);

	// Draw low quality and high quality in different colors
	if (viewqu == VIEW_QUALITY_LOW)
	  XSetForeground(the_disp, map_gc, darkgrey);
	else
	  XSetForeground(the_disp, map_gc, lightgrey);

	XFillPolygon(the_disp, the_map, map_gc,
	  viewcone, 4, Nonconvex, CoordModeOrigin);	
}

void WorldMap::drawlines()
{
  XSetForeground(the_disp, map_gc, white);
  XDrawLine(the_disp, the_map, map_gc,		// Center line
	    mx(0), my(-FIELD_WIDTH/2),
	    mx(0), my(FIELD_WIDTH/2));
  XDrawRectangle(the_disp, the_map, map_gc,	// Left penalty area
		 mx(-FIELD_LENGTH/2) - 1, my(-PENALTY_AREA_WIDTH/2),
		 md(PENALTY_AREA_LENGTH), md(PENALTY_AREA_WIDTH));
  XDrawRectangle(the_disp, the_map, map_gc,	// Right penalty area
		 mx(FIELD_LENGTH/2 - PENALTY_AREA_LENGTH), my(-PENALTY_AREA_WIDTH/2),
		 md(PENALTY_AREA_LENGTH), md(PENALTY_AREA_WIDTH));
  XDrawRectangle(the_disp, the_map, map_gc,	// Left goal area
		 mx(-FIELD_LENGTH/2) - 1, my(-GOAL_AREA_WIDTH/2),
		 md(GOAL_AREA_LENGTH), md(GOAL_AREA_WIDTH));
  XDrawRectangle(the_disp, the_map, map_gc,	// Right goal area
		 mx(FIELD_LENGTH/2 - GOAL_AREA_LENGTH), my(-GOAL_AREA_WIDTH/2),
		 md(GOAL_AREA_LENGTH), md(GOAL_AREA_WIDTH));
  XDrawRectangle(the_disp, the_map, map_gc,	// Left goal
		 mx(-FIELD_LENGTH/2), my(-GOAL_WIDTH/2),
		 1, md(GOAL_WIDTH));
  XDrawRectangle(the_disp, the_map, map_gc,	// Right goal
		 mx(FIELD_LENGTH/2) - 2, my(-GOAL_WIDTH/2),
		 1, md(GOAL_WIDTH));
  XDrawArc(the_disp, the_map, map_gc,		// Center circle
	   mx(-CENTER_CIRCLE_R), my(-CENTER_CIRCLE_R),
	   md(CENTER_CIRCLE_R*2), md(CENTER_CIRCLE_R*2),
	   0, 360 * 64);
}

void WorldMap::drawownplayer(WorldPlayer ownplayer)
{
  // My Coordinates 

  int x = mx(ownplayer.pos.getX());
  int y = my(ownplayer.pos.getY());

  XSetForeground(the_disp, map_gc, blue);

  XDrawArc(the_disp, the_map, map_gc,		// Position
	   x - PSIZE, y - PSIZE,
	   PSIZE*2, PSIZE*2,
	   0, 360 * 64);
  XDrawLine(the_disp, the_map, map_gc,	       // Velocity
	    x, y,
	    x + mv(ownplayer.v.getX()), y + mv(ownplayer.v.getY()));

  XFillArc(the_disp, the_map, map_gc,		// Body direction
	   x - PSIZE, y - PSIZE,
	   PSIZE*2, PSIZE*2,
	   -int(RAD_TO_DEG(ownplayer.bodyDir + PI_4) * 64), 
	   90 * 64);
    
  XDrawString(the_disp, the_map, map_gc,	// Number
	      x - PSIZE, y + PSIZE + 8,
	      num2str[ownplayer.number], 2);
}

void WorldMap::drawball(WorldBall ball)
{
	int x = mx(ball.pos.getX());
	int y = my(ball.pos.getY());
	
	if (ball.age > MAX_AGE_AS_NOW)
	  XSetForeground(the_disp, map_gc, yellow);
	else
	  XSetForeground(the_disp, map_gc, lyellow);

	XDrawArc(the_disp, the_map, map_gc,	// Position
		x - PSIZE, y - PSIZE,
		PSIZE*2, PSIZE*2,
		0, 360 * 64);
	XDrawLine(the_disp, the_map, map_gc,	// Velocity
		x, y,
		x + mv(ball.v.getX()), y + mv(ball.v.getY()));
	XDrawLine(the_disp, the_map, map_gc,	// Age
		x - PSIZE, y - PSIZE - 2,
		x - PSIZE + ball.age * PSIZE , y - PSIZE - 2);
}

void WorldMap::drawmyteam(Snapshot *snap, int max_player_age)
{
  for (int i=1; i<=11; i++) {
    if (i == snap->my.number)
      continue;
    if (snap->ourTeam(i).age > MAX_AGE_AS_NOW)
      XSetForeground(the_disp, map_gc, green);
    else
      XSetForeground(the_disp, map_gc, lgreen);    
    drawplayer(snap->ourTeam(i), max_player_age);
  }
}

void WorldMap::drawotherteam(Snapshot *snap, int max_player_age)
{
  for (int i=1; i<=11; i++) {
    if (snap->otherTeam(i).age > MAX_AGE_AS_NOW)
      XSetForeground(the_disp, map_gc, red);
    else
      XSetForeground(the_disp, map_gc, lred);
    drawplayer(snap->otherTeam(i), max_player_age);
  }
}

void WorldMap::drawplayer(WorldPlayer player, int max_age)
{
  if (player.age > max_age)
    return;

  int x = mx(player.pos.getX());
  int y = my(player.pos.getY());
  int num = player.number;

  XDrawArc(the_disp, the_map, map_gc,		// Position
	   x - PSIZE, y - PSIZE,
	   PSIZE*2, PSIZE*2,
	   0, 360 * 64);
  XDrawLine(the_disp, the_map, map_gc,	        // Velocity
	    x, y,
	    x + mv(player.v.getX()), y + mv(player.v.getY()));
  XFillArc(the_disp, the_map, map_gc,		// body direction
	   x - PSIZE, y - PSIZE,
	   PSIZE*2, PSIZE*2,
	   -int(RAD_TO_DEG(player.bodyDir + PI_4) * 64), 
	   90 * 64);
  XDrawLine(the_disp, the_map, map_gc,	        // Age
	    x - PSIZE, y - PSIZE - 2,
	    x - PSIZE + player.age * PSIZE, y - PSIZE - 2);
  if (num <= 0) num = 0;
  XDrawString(the_disp, the_map, map_gc,	// Number
	      x - PSIZE, y + PSIZE + 8,
	      num2str[num], 2);
}


