//////////////////////////////////////////////////////////////////////
// Mainz Rolling Brains RoboCup agents developed for the Simulator 
// League of the RoboCup 2000 competition
//
// Copyright (C) 1998, 1999, 2000 Team Members 
// of the Mainz Rolling Brains RoboCup Team 
//
// This program under the terms of the GNU General Public License as
// published by the Free Software Foundation; version 2 of the
// License. See the README file in the root directory of our archive.
//////////////////////////////////////////////////////////////////////

//////////////////////////////////////////////////////////////////////
//
// $Id: worldmap_r.C,v 1.12 2000/10/19 15:27:48 goetz Exp $	
// $Manager: goetz$
// 
// 
//  worldmap_r.C - World model display
// 
//  ( Christian Bauer <cebix> )
//    Goetz Schwandtner
//
//
//////////////////////////////////////////////////////////////////////

#include <stdio.h>
#include <string.h>

#include "worldmap_r.h"

#include "constants.h"
#include "coordinates.h"

// Internal constants

static char *num2str[] = {"? ", "1 ", "2 ", "3 ", "4 ", "5 ",
			 "6 ", "7 ", "8 ", "9 ", "10", "11"};

static char *rel_defaulttitle ="MRB Worldmodel display";

static char *rel_append_title =" (relative)";

static char *fontname="fixed";
/*
 *  Constructor and destructor
 */

WorldMap_r::WorldMap_r(const char *windowtitle, double mag, double relmag) :
  WorldMap(windowtitle, mag)
{
  if (relmag) {
    if (windowtitle) {
      rel_title = new char[strlen(windowtitle)+
			  strlen(rel_append_title)+1];
      strcpy(rel_title, windowtitle);
    } else {
      rel_title = new char[strlen(rel_defaulttitle)+
			  strlen(rel_append_title)+1];
      strcpy(rel_title, rel_defaulttitle);
    }
    // add " (relative)" to window title for relative map window
    strcat(rel_title, rel_append_title);
    
    // change constants for relative map window
    REL_MAGNIFY = relmag > 0 ? relmag : MAGNIFY;
    REL_PSIZE = int(PLAYER_SIZE * REL_MAGNIFY * PLAYER_MAG);
    REL_VIEW_CONE_LENGTH = REL_PSIZE * 150;
    REL_WIN_WIDTH = int(FIELD_LENGTH * REL_MAGNIFY);
    REL_WIN_HEIGHT = int(FIELD_WIDTH * REL_MAGNIFY);
    // tell other functions about relative map window
    has_rel_map = true;
  } else {
    has_rel_map = false;
  }
}

WorldMap_r::~WorldMap_r()
{
  if (rel_title)
    delete [] rel_title;
  // close window if not done (close checks this)
  Close();
}

/*
 *  Open map window
 */

bool WorldMap_r::Open()
{
  // Open absolute map window and prepare X connection ...
  if (!WorldMap::Open())
    return false;
  
  // just in case something went terribly wrong ...
  if (!the_disp)
    return false;

  // is there a relative map window wanted?
  if (!has_rel_map)
    return false;
  
  // Open window
  rel_win = XCreateSimpleWindow(the_disp, the_root, 0, 0, 
				REL_WIN_WIDTH, REL_WIN_HEIGHT, 
				0, black, white);
  XStoreName(the_disp, rel_win, rel_title);
  XMapRaised(the_disp, rel_win);
  
  // Create pixmap and attach to window
  rel_map = XCreatePixmap(the_disp, rel_win, REL_WIN_WIDTH, 
			  REL_WIN_HEIGHT, 
			  DefaultDepth(the_disp, the_screen));
  XSetWindowBackgroundPixmap(the_disp, rel_win, rel_map);

  // Create GCs
  rel_gc = XCreateGC(the_disp, rel_win, 0, NULL);
  rel_map_gc = XCreateGC(the_disp, rel_map, 0, NULL);
    
  Font font = XLoadFont(the_disp, fontname);
  XSetFont(the_disp, rel_map_gc, font);

  // Wait until window is ready
  XSync(the_disp, false);
  return(true);
}


/*
 *  Close map window
 */

void WorldMap_r::Close()
{
  if ((the_disp != NULL) && has_rel_map) {
    XFreeGC(the_disp, rel_map_gc);
    XFreeGC(the_disp, rel_gc);
    XDestroyWindow(the_disp, rel_win);
  }
  WorldMap::Close();
}


/*
 *  Draw snapshot
 */

void WorldMap_r::Map(Snapshot *snap, int max_age)
{
  if (the_disp == NULL)
    return;
  
  if (snap == NULL)
    return;
  
  
  // Draw absolute worldmap
  WorldMap::Map(snap, max_age);

  if (!has_rel_map)
    return;

  // Clear pixmap
  XSetForeground(the_disp, rel_map_gc, black);
  XFillRectangle(the_disp, rel_map, rel_map_gc, 0, 0, 
		 REL_WIN_WIDTH, REL_WIN_HEIGHT);

  // Draw viewcone (not for coach)
  if (snap->my.number > 0)
    drawviewcone(snap->my.faceDir,
		 snap->senseBody.getViewAngle(), snap->senseBody.viewQuality);

  // Set position of own player to ownpos
  setownposandvel(snap->my);

  // Draw myself (not for coach)
  if (snap->my.number > 0)
    drawownplayer(snap->my);

  // Draw ball
  drawball(snap->ball);
	
  // Draw my team
  drawmyteam(snap, max_age);
	
  // Draw other team
  drawotherteam(snap, max_age);
  
  // Draw pixmap
  XClearWindow(the_disp, rel_win); 
  XFlush(the_disp);
}

void WorldMap_r::drawviewcone(Angle orient, Angle viewangle, 
			      view_quality viewqu)
{
	XPoint viewcone[4];  // Points for drawing viewcone as polygon
	int x = mx(0);
	int y = my(0);
 
		// Draw viewcone (grey)
	// set viewcone points: origin
	viewcone[0].x = x;
	viewcone[0].y = y;
	
	// set left and right margin points of view cone

	viewcone[1].x = x + md(cos(orient + viewangle) * VIEW_CONE_LENGTH);
	viewcone[1].y = y + md(sin(orient + viewangle) * VIEW_CONE_LENGTH);
	viewcone[2].x = x + md(cos(orient) * VIEW_CONE_LENGTH);
	viewcone[2].y = y + md(sin(orient) * VIEW_CONE_LENGTH);
	viewcone[3].x = x + md(cos(orient - viewangle) * VIEW_CONE_LENGTH);
	viewcone[3].y = y + md(sin(orient - viewangle) * VIEW_CONE_LENGTH);
	// Draw low quality and high quality in different colors
	if (viewqu == VIEW_QUALITY_LOW)
	  XSetForeground(the_disp, rel_map_gc, darkgrey);
	else
	  XSetForeground(the_disp, rel_map_gc, lightgrey);

	XFillPolygon(the_disp, rel_map, rel_map_gc,
	  viewcone, 4, Nonconvex, CoordModeOrigin);
}

void WorldMap_r::setownposandvel(WorldPlayer ownplayer)
{
  ownx = ownplayer.pos.getX();
  owny = ownplayer.pos.getY();
  ownvx = ownplayer.v.getX();
  ownvy = ownplayer.v.getY();
}

void WorldMap_r::drawownplayer(WorldPlayer ownplayer)
{
  // My Coordinates 

  int x = mx(0);
  int y = my(0);
  
  XSetForeground(the_disp, rel_map_gc, blue);

  XDrawArc(the_disp, rel_map, rel_map_gc,		// Position
	   x - PSIZE, y - PSIZE,
	   PSIZE*2, PSIZE*2,
	   0, 360 * 64);
  XDrawLine(the_disp, rel_map, rel_map_gc,	       // Velocity
	    x, y,
	    x + mv(ownplayer.v.getX() - ownvx), y + mv(ownplayer.v.getY() - ownvy));
  
  XFillArc(the_disp, rel_map, rel_map_gc,		// Body direction
	   x - PSIZE, y - PSIZE,
	   PSIZE*2, PSIZE*2,
	   -int(RAD_TO_DEG(ownplayer.bodyDir + PI_4) * 64), 
	   90 * 64);
    
  XDrawString(the_disp, rel_map, rel_map_gc,	// Number
	      x - PSIZE, y + PSIZE + 8,
	      num2str[ownplayer.number], 2);
}

void WorldMap_r::drawball(WorldBall ball)
{
	int x = mx(ball.pos.getX() - ownx);
	int y = my(ball.pos.getY() - owny);

	if (ball.age > MAX_AGE_AS_NOW) 
	  XSetForeground(the_disp, rel_map_gc, yellow);
	else 
	  XSetForeground(the_disp, rel_map_gc, lyellow);
	
	XDrawArc(the_disp, rel_map, rel_map_gc,	// Position
		x - PSIZE, y - PSIZE,
		PSIZE*2, PSIZE*2,
		0, 360 * 64);
	XDrawLine(the_disp, rel_map, rel_map_gc,	// Velocity
		x, y,
		x + mv(ball.v.getX()-ownvx), y + mv(ball.v.getY()-ownvy));
	XDrawLine(the_disp, rel_map, rel_map_gc,	// Age
		x - PSIZE, y - PSIZE - 2,
		x - PSIZE + ball.age * PSIZE , y - PSIZE - 2);
}

void WorldMap_r::drawmyteam(Snapshot *snap, int max_player_age)
{
  for (int i=1; i<=11; i++) {
    if (i == snap->my.number)
      continue;
    if (snap->ourTeam(i).age > MAX_AGE_AS_NOW)
      XSetForeground(the_disp, rel_map_gc, green);
    else
      XSetForeground(the_disp, rel_map_gc, lgreen);        
    drawplayer(snap->ourTeam(i), max_player_age);
  }
}

void WorldMap_r::drawotherteam(Snapshot *snap, int max_player_age)
{
  for (int i=1; i<=11; i++) {
    if (snap->otherTeam(i).age > MAX_AGE_AS_NOW)
      XSetForeground(the_disp, rel_map_gc, red);
    else
      XSetForeground(the_disp, rel_map_gc, lred);
    drawplayer(snap->otherTeam(i), max_player_age);
  }
}

void WorldMap_r::drawplayer(WorldPlayer player, int max_age)
{
  if (player.age > max_age)
    return;

  int x = mx(player.pos.getX() - ownx);
  int y = my(player.pos.getY() - owny);
  int num = player.number;

  XDrawArc(the_disp, rel_map, rel_map_gc,		// Position
	   x - PSIZE, y - PSIZE,
	   PSIZE*2, PSIZE*2,
	   0, 360 * 64);
  XDrawLine(the_disp, rel_map, rel_map_gc,	        // Velocity
	    x, y,
	    x + mv(player.v.getX()-ownvx), y + mv(player.v.getY()-ownvy));
  XFillArc(the_disp, rel_map, rel_map_gc,		// face direction
	   x - PSIZE, y - PSIZE,
	   PSIZE*2, PSIZE*2,
	   -int(RAD_TO_DEG(player.bodyDir + PI_4) * 64), 
	   90 * 64);
  XDrawLine(the_disp, rel_map, rel_map_gc,	        // Age
	    x - PSIZE, y - PSIZE - 2,
	    x - PSIZE + player.age * PSIZE, y - PSIZE - 2);
  if (num <= 0) num = 0;
  XDrawString(the_disp, rel_map, rel_map_gc,	// Number
	      x - PSIZE, y + PSIZE + 8,
	      num2str[num], 2);
}

