//////////////////////////////////////////////////////////////////////
// Mainz Rolling Brains RoboCup agents developed for the Simulator 
// League of the RoboCup 2000 competition
//
// Copyright (C) 1998, 1999, 2000 Team Members 
// of the Mainz Rolling Brains RoboCup Team 
//
// This program under the terms of the GNU General Public License as
// published by the Free Software Foundation; version 2 of the
// License. See the README file in the root directory of our archive.
//////////////////////////////////////////////////////////////////////

//////////////////////////////////////////////////////////////////////// 
//   $Id: worldmodel.h,v 1.10 2000/10/19 15:27:49 goetz Exp $	
//   $Manager: cm$
//
//   worldmodel.h
//
////////////////////////////////////////////////////////////////////////

#ifndef _WORLDMODEL_H_
#define _WORLDMODEL_H_

#include "sensor.h"         // sensor data types
#include "comm.h"           // communication module
#include "coordinates.h"    // angle class and vector class for 2-dimensional space
#include "constants.h"      // constants used by soccerserver
#include "snapshot.h"       // declaration of snapshot data types


//////////////////////////////////////////////////////////////////////
//
// Class WorldAction
//
//////////////////////////////////////////////////////////////////////

class WorldAction {
public:
  WorldAction() 
    : time( 0 ), type( NO_COMMAND ), moment( 0.0 ), power( 0.0 ), pos ( 0.0, 0.0 ) {}

  int           time;       // time index (for snapshot::action array)
  CycleCommand  type;       // different types of actions (see "comm.h")

  double        moment;     // valid if type == TURN_COMMAND or CATCH_COMMAND or KICK_COMMAND
  double        power;      // valid if type == DASH_COMMAND or KICK_COMMAND
  Vector2d      pos;        // valid if type == MOVE_COMMAND
};


//////////////////////////////////////////////////////////////////////
//
// Class Worldmodel
//
//////////////////////////////////////////////////////////////////////

class Worldmodel {
public:
  Worldmodel( int _numSnapshots, Comm* _comm );
  
  Worldmodel( const Worldmodel& wm );
  
  Worldmodel& Worldmodel::operator = ( const Worldmodel& wm );
  
  ~Worldmodel();
  
  int            getVisualTime() { return visualTime; }
  int            getSenseTime() { return senseTime; }
  
  Snapshot*      getSnapshot( int time );    // get simulated world at a certain time
  Snapshot*      getCurrentSnapshot() { return getSnapshot( comm->getServerTime() ); }

  bool           adjust();         // fill in snapshot array with new data (if there is)

  // feed audio information into worldmodel
  bool           feedbackAudio( AudioMsg &audioMsg );

  // write worldmodel information for timestep maxTime into AudioMsg stream
  bool           getSayMsg( AudioMsg &audioMsg );

private:
  Comm*          comm;             // communication module
  
  Snapshot*      snapshots;        // Simulations of the world at different time steps
  WorldAction*   actions;          // player's actions at different time steps
  
  int            numSnapshots;     // number of entries in "snapshots" and "actions"
  
  // get Index for lookup in snapshot/actions array
  // value for time<0 is 0 (just to avoid calls like snapshots[-1]...)
  int            getTimeIndex( int time ) { return ( time >= 0 ) ? time % numSnapshots : 0; }
  
  int            visualTime;       // time of latest valid visual or audial information
  int            senseTime;        // time of latest valid sense information

  int            maxTime;          // time of latest snapshot created
  
  int            lastCatchTime;    // time when last catch command was sent
 
  void           simVisualData( int time );   // simulate visual/sense data in snapshot for
  void           simSenseData( int time );    // (time + 1) with data in snapshot for (time)

  // update worldmodel with a snapshot received via audio message
  bool           updateWithSnapshot( Snapshot audioSnap, int sender );

  // ask comm-module for last action sent by agent and fill in "actions" array
  void           getAction();

};


#endif  // _WORLDMODEL_H_
