//////////////////////////////////////////////////////////////////////
// Mainz Rolling Brains RoboCup agents developed for the Simulator 
// League of the RoboCup 2000 competition
//
// Copyright (C) 1998, 1999, 2000 Team Members 
// of the Mainz Rolling Brains RoboCup Team 
//
// This program under the terms of the GNU General Public License as
// published by the Free Software Foundation; version 2 of the
// License. See the README file in the root directory of our archive.
//////////////////////////////////////////////////////////////////////

//////////////////////////////////////////////////////////////////////
//
// MRB player class  GoalKickEval
//
// \$Id: goalkickeval.C,v 1.27 2000/10/19 15:27:52 goetz Exp $
// \$manager: $schappel
//
//////////////////////////////////////////////////////////////////////

#include "goalkickeval.h"

////////////////////////////
//
// Constructor
//
///////////////////////////

GoalKickEval::GoalKickEval():
  // quasi-constants, perhaps with an adaptive behaviour?  
  // estimated values!?

  maxAge(20)
{
  superKickDuration = Config::getValueAsInt("goal_superKickDuration");       
  fixedKickWithTurnDuration = Config::getValueAsInt("goal_fixedKickWithTurnDuration");
  fixedKickDuration = Config::getValueAsInt("goal_fixedKickDuration");      
  kickAwayDuration  = Config::getValueAsInt("goal_kickAwayDuration");

  superKickDist = Config::getValueAsDouble("goal_superKickDist");
  fixedKickDist = Config::getValueAsDouble("goal_fixedKickDist");
  kickAwayDist  = Config::getValueAsDouble("goal_kickAwayDist");

  corridorWidth      = Config::getValueAsDouble("goal_corridorWidth");
  corridorWidthSmall = Config::getValueAsDouble("goal_corridorWidthSmall");

  distGradeIsImpDist    = Config::getValueAsDouble("goal_distGradeIsImpDist");
  distGradeIsBadDist    = Config::getValueAsDouble("goal_distGradeIsBadDist");
  distGradeIsPoorDist    = Config::getValueAsDouble("goal_distGradeIsPoorDist");
  distGradeIsAverageDist = Config::getValueAsDouble("goal_distGradeIsAverageDist");
  distGradeIsGoodDist    = Config::getValueAsDouble("goal_distGradeIsGoodDist");
  anyCaseGoalKickDist    = Config::getValueAsDouble("goal_anyCaseGoalKickDist");

  goalieNearDist = Config::getValueAsDouble("goal_goalieNearDist");
  goalLineImpDist = Config::getValueAsDouble("goal_goalLineImpDist");
};

////////////////////////////
//
// Destructor
//
///////////////////////////

GoalKickEval::~GoalKickEval()
{};


////////////////////////////////////////////////////////////////////////////////////
//
// public methods
//
////////////////////////////////////////////////////////////////////////////////////


//////////////////////////////////////
//
// Grade GoalKickEval::ourEval
//
/////////////////////////////////////

Grade GoalKickEval::ourEval(const Snapshot* worldSnap, Vector2d playerpos) const
{
  Grade grade;

  // do we have a valid snapshot?
  if(worldSnap == NULL)
    {
      cerr << "GOALEVAL: ourEval -> snapshot is NULL" << endl;
      grade.set_IMPOSSIBLE();
      return grade;
    }


  // the player is standing in his own half of the field, so goalkick is
  // impossible
  
  if(playerpos.getX() < FIELD_LENGTH /4.)
    {
      GOALEVALTRACE("ourEval -> player's xposition is too small ( " 
		    << playerpos.getX() << " ) so grade = IMPOSSIBLE");
      grade.set_IMPOSSIBLE();
      return grade;
    }

  // calculate the minimum of the players distance to the left edge
  // the middle or the right edge of the goal for calculating a first
  // goal target

  double minGoalDist;
  Vector2d goalTarget;

  minGoalDist = minDistToGoal(playerpos, goalTarget);

  // how much time does the player have for kicking ?

  int availableKickTime = availableTimeForKick(worldSnap, playerpos);
  GOALEVALTRACE("ourEval -> available time for kick = " << availableKickTime);

  // is a kick with length minDistToGoal possible with the 
  // above calculated available time for kicking?

  double posKickLength = possibleKickLength(availableKickTime);
  
  if(posKickLength < minGoalDist)
    {
      grade.set_BAD();
      GOALEVALTRACE("ourEval -> possibleKickLength " << posKickLength 
		    << " < min. Distanz zum Tor "
		    << minGoalDist
		    << "; grade = "
		    << grade);
      return grade;
    }

  // searching for the best corridor and evaluating it
  grade = searchAndEvalOurBestCorridor(worldSnap, playerpos,
				       posKickLength, goalTarget);
  return grade;

};

////////////////////////////////////
//
// Grade GoalKickEval::otherEval
//
///////////////////////////////////

Grade GoalKickEval::otherEval(const Snapshot* worldSnap, Vector2d playerpos) const
{
  Grade grade;

  // do we have a valid snapshot?
  if(worldSnap == NULL)
    {
      cerr << "GOALEVAL: otherEval -> snapshot is NULL" << endl;
      grade.set_IMPOSSIBLE();
      return grade;
    }


  // the player is standing in his own half of the field, so goalkick is
  // impossible
  
  if(playerpos.getX() > -FIELD_LENGTH / 4.)
    {
      GOALEVALTRACE("otherEval -> players xposition is too big ( "
		    << playerpos.getX() << " ) so grade = IMPOSSIBLE");
      grade.set_IMPOSSIBLE();
      return grade;
    }

  // calculate the minimum of the players distance to the left edge
  // the middle or the right edge of the goal for calculating a first
  // goal target

  double minGoalDist;
  Vector2d goalTarget;

  minGoalDist = minDistToGoal(playerpos, goalTarget, OTHER_TEAM);

  // how much time does the player have for kicking ?
  int availableKickTime = availableTimeForKick(worldSnap, playerpos, OTHER_TEAM);
  GOALEVALTRACE("otherEval -> available time for kick = " << availableKickTime);

  // is a kick with length minDistToGoal possible with the 
  // above calculated available time for kicking?

  double posKickLength = possibleKickLength(availableKickTime);
  
  if(posKickLength < minGoalDist)
    {
      grade.set_BAD();
      GOALEVALTRACE("otherEval -> possibleKickLength " << posKickLength 
		    << " < min. Distanz zum Tor "
		    << minGoalDist
		    << "; grade = "
		    << grade);
      return grade;
    }

  // searching for the best corridor and evaluating it
  grade = searchAndEvalOtherBestCorridor(worldSnap, playerpos,
					 posKickLength, goalTarget);
  return grade;
};  




////////////////////////////////////////////////////////////////////////////////////
//
// private methods
//
////////////////////////////////////////////////////////////////////////////////////


///////////////////////////////////////////
//
// GoalKickEval::distToLeftEdge
//
//////////////////////////////////////////

double GoalKickEval::distToLeftEdge(Vector2d playerpos,
				    team_type team) const
{
  double goaldist;
  if(team == OUR_TEAM)
    goaldist = dist(playerpos, 
		    Vector2d(FIELD_LENGTH /2., -GOAL_WIDTH / 2. +1) );
  else
    goaldist = dist(playerpos, 
		    Vector2d(-FIELD_LENGTH /2., -GOAL_WIDTH / 2. +1) ); 
  return goaldist;
}

///////////////////////////////////////////
//
// GoalKickEval::distToMiddle
//
//////////////////////////////////////////

double GoalKickEval::distToMiddle(Vector2d playerpos,
				  team_type team) const
{
  double goaldist;
  if(team == OUR_TEAM)
    goaldist = dist(playerpos, 
		    Vector2d(FIELD_LENGTH /2., 0) );
  else
    goaldist = dist(playerpos, 
		    Vector2d(-FIELD_LENGTH /2., 0 ) ); 
  return goaldist;
}


///////////////////////////////////////////
//
// GoalKickEval::distToRightEdge
//
//////////////////////////////////////////

double GoalKickEval::distToRightEdge(Vector2d playerpos,
				     team_type team) const
{
  double goaldist;
  if(team == OUR_TEAM)
    goaldist = dist(playerpos, 
		    Vector2d(FIELD_LENGTH /2., GOAL_WIDTH / 2. - 1) );
  else
    goaldist = dist(playerpos, 
		    Vector2d(-FIELD_LENGTH /2., -GOAL_WIDTH / 2. - 1) ); 
  return goaldist;
}


///////////////////////////////////////////
//
// GoalKickEval::minDistToGoal
//
//////////////////////////////////////////

double GoalKickEval::minDistToGoal(Vector2d playerpos,
				   Vector2d &target, team_type team) const
{
  Vector2d leftEdgeOfGoal;
  Vector2d middleOfGoal;

  Vector2d rightEdgeOfGoal;

  double distToLeftEdge;
  double distToMiddleOfGoal;
  double distToRightEdge;

  if(team == OUR_TEAM)
    {
      leftEdgeOfGoal = Vector2d(FIELD_LENGTH /2., -GOAL_WIDTH /2. + 1);
      middleOfGoal = Vector2d(FIELD_LENGTH / 2., 0);
      rightEdgeOfGoal = Vector2d(FIELD_LENGTH / 2., GOAL_WIDTH / 2. - 1);
      
      distToLeftEdge = dist(playerpos, leftEdgeOfGoal);
      distToMiddleOfGoal = dist(playerpos, middleOfGoal);
      distToRightEdge = dist(playerpos, rightEdgeOfGoal);
    }
  else
    {
      leftEdgeOfGoal = Vector2d(-FIELD_LENGTH /2., -GOAL_WIDTH /2. + 1);
      middleOfGoal = Vector2d(-FIELD_LENGTH / 2., 0);
      rightEdgeOfGoal = Vector2d(-FIELD_LENGTH / 2., GOAL_WIDTH / 2. - 1);
      distToLeftEdge = dist(playerpos, leftEdgeOfGoal);
      distToMiddleOfGoal = dist(playerpos, middleOfGoal);
      distToRightEdge = dist(playerpos, rightEdgeOfGoal);
    }
    
  double minDist = 0;

  if( distToLeftEdge < distToMiddleOfGoal)
    if(distToRightEdge < distToLeftEdge)
      {
	minDist = distToRightEdge;
	target = rightEdgeOfGoal;
      }
    else
      {
	minDist = distToLeftEdge;
	target = leftEdgeOfGoal;
      }
  else
    if(distToRightEdge < distToMiddleOfGoal)
      {
	minDist = distToRightEdge;
	target = rightEdgeOfGoal;
      }
    else
      {
	minDist = distToMiddleOfGoal;
	target = middleOfGoal;
      }
  
  return minDist;
	
}


///////////////////////////////////////////
//
// GoalKickEval::availableTimeForKick
//
//////////////////////////////////////////


int GoalKickEval::availableTimeForKick(const Snapshot *worldSnap, Vector2d playerpos, 
				       team_type team) const
{
  if(worldSnap == NULL)
    {
      cerr << "GOALEVAL: availableTimeForKick -> invalid worldSnap)" << endl;
      return 0;
    }

  // identify the closest opponent

  int closestOppNo;
  if(team == OUR_TEAM)
    // identify the member of the other team being next to the player with no "player"
    // of our team
    closestOppNo = worldSnap->otherTeamClosestTo(playerpos,
						 maxAge);
  else
    // identify the member of our team being next to the player with no "player"
    // of the other team
    closestOppNo = worldSnap->ourTeamClosestTo(playerpos,
					       maxAge);
    
  // player sees no opponent, so everything is allright
  if (closestOppNo == UNKNOWN) 
    return superKickDuration;               

  // calculating the distance to the nearest player

  double distToOppPlayer;
  
  if (team == OUR_TEAM)
    distToOppPlayer 
      = dist(worldSnap->otherTeam(closestOppNo).pos, playerpos);
  else
    distToOppPlayer
      = dist(worldSnap->ourTeam(closestOppNo).pos, playerpos);

  // opponent is far enough away to do a superkick
  if ( distToOppPlayer > PLAYER_SPEED_MAX * superKickDuration)
    return superKickDuration;

  // opponent is far enough away to do a turn befor kicking
  if (distToOppPlayer > PLAYER_SPEED_MAX * fixedKickWithTurnDuration)
    return fixedKickWithTurnDuration;

  // opponent is far enough away to do kicking with best position of the ball
  if (distToOppPlayer > PLAYER_SPEED_MAX * fixedKickDuration)
    return fixedKickDuration;

  // opponent is near, so only a kick away is possible
  return kickAwayDuration;

}


///////////////////////////////////////////
//
// GoalKickEval::possibleKickLength
//
//////////////////////////////////////////

double GoalKickEval::possibleKickLength(int time) const
{
  if(time == superKickDuration) 
    return superKickDist;
  
  if(time == fixedKickWithTurnDuration)
    return fixedKickDist;

  return kickAwayDist;

}

///////////////////////////////////////////////
//
// GoalKickEval::distToLine
//
//////////////////////////////////////////////

double GoalKickEval::distToLine(Vector2d object, Vector2d source, 
				Vector2d target) const
{
  double sx = source.getX();
  double sy = source.getY();

  double tx = target.getX();
  double ty = target.getY();

  double ox = object.getX();
  double oy = object.getY();

  // project object onto the line source->target
  double proj = ( ( (sx-tx) * (sx-ox) + (sy-ty) * (sy-oy) ) /
		  ( (sx-tx) * (sx-tx) + (sy-ty) * (sy-ty) ) );
  
  if ( proj < 0 )
    // projection point is in front of the source
    return dist(object, source);
  
  if( proj > 1)
    return dist(object, target); 

  // calculate distance between object and the line source->target
  return sqrt( sqr( ox - sx - proj * (tx-sx) ) + sqr( oy - sy - proj * (ty-sy) ) );

}

///////////////////////////////////////////
//
// GoalKickEval::evaluateCorridor
//
//////////////////////////////////////////

Grade GoalKickEval::evaluateCorridor(const Snapshot *worldSnap, Vector2d playerpos,  
				     Vector2d target, Angle angle,
				     Grade startGrade,
				     bool, //  goalieInMiddleCorridor,
				     team_type team) const
{
  Grade grade, distgrade, corridorgrade;

  // calculating the grade for the dist between the player and target
  // if the distance from the player to the target
  //       is greater than distGradeIsImpDist : distgrade = IMPOSSIBLE
  //       is greater than distGradeIsBadDist : distgrade = BAD
  //       is greater than distGradeIsPoorDist : distgrade = POOR
  //       is greater than distGradeIsAverageDist and smaller than distGradeIsPoorDist: distgrade = AVERAGE
  //       is greater than distGradeIsGoodDist and smaller than distGradeIsAverageDist: distgrade = GOOD
  //       is smaller than distGradeIsGoodDist : distgrade = PERFECT
  
  double distToTarget = dist(playerpos, target);

  if(distToTarget < anyCaseGoalKickDist)
    // the distance to the target is very small, so try to kick the ball an every case
    {
      grade.set_PERFECT();
      GOALEVALTRACE("evaluateCorr->distance to target " << target
		    << " is " << distToTarget 
		    << " and smaller than " << anyCaseGoalKickDist
		    << " grade = " << grade);
      return grade;
    }

  if (distToTarget > distGradeIsImpDist ) {
    // This is near to technically impossible (in a normal game), so set
    // grade to impossible and return
    distgrade.set_IMPOSSIBLE();

    // trace and return are for speed reasons because our agents
    // often have timing problems
    GOALEVALTRACE("evaluateCorr->distance of player to target " << target << " is " 
		  << distToTarget << ", so distgrade = "
		  << distgrade);
    return distgrade;
  } else if( distToTarget > distGradeIsBadDist )
    distgrade.set_BAD();
  else if( distToTarget > distGradeIsPoorDist )
    distgrade.set_POOR();
  else if( distToTarget > distGradeIsAverageDist) 
    distgrade.set_AVERAGE();
  else if( distToTarget > distGradeIsGoodDist )
    distgrade.set_GOOD();
 
  else if( distToTarget > anyCaseGoalKickDist )
    distgrade.set_PERFECT();
  else {
    // in this case don't care about corridors
    GOALEVALTRACE("evaluateCorr->distance of player to target " << target << " is " 
		  << distToTarget << ", so distgrade = "
		  << distgrade << "without any downgrade." );
    
    distgrade.set_PERFECT();
    return distgrade; 
  }

  GOALEVALTRACE("evaluateCorr->distance of player to target " << target << " is " 
		<< distToTarget << ", so distgrade = "
		<< distgrade);
  
  // calculating the grade for the corridor itself and then taking the worse of this
  // grade and the distgrade

  corridorgrade = startGrade;

  int numberOfOpponents;
  
  if(team == OUR_TEAM)
    numberOfOpponents 
      = worldSnap->otherTeamInCorridor(playerpos, 
				       target, corridorWidth, PI_6, maxAge);
  else
    numberOfOpponents 
      = worldSnap->ourTeamInCorridor(playerpos, 
				     target, corridorWidth, PI_6, maxAge);
    
  if(numberOfOpponents == 0)
    {
      // corridorgrade stays at the startvalue
      GOALEVALTRACE("evaluateCorr->corridor towards position " << target << " with width " 
		    << corridorWidth << " is free; corridorgrade = "
		    << corridorgrade);
	
    }
  else
    // corridorgrade will be downgraded
    {
      corridorgrade.downgrade();
      // there are opponents in the corridor, so let us look if there is no opponent
      // whoms kickrange will be crossed by the ball (supposed none of the players
      // is moving) 

      if(team == OUR_TEAM)
	
	numberOfOpponents =
	  worldSnap->otherTeamInCorridor(playerpos, 
					 target, corridorWidthSmall, angle, maxAge);
      else
	numberOfOpponents =
	  worldSnap->ourTeamInCorridor(playerpos, 
				       target, corridorWidthSmall, angle, maxAge);
      
      if(numberOfOpponents == 0)
	{
	  GOALEVALTRACE("evaluateCorr->corridor towards position " << target <<" with smaller width ("
			<< corridorWidthSmall << ") is free; corridorgrade = "
			<< corridorgrade);
	}
      else
	// corridorgrade will be downgraded again
	{
	  corridorgrade.downgrade();
	  // there are even opponents in the corridor with width corridorWidth
	  // let us look if the corridor with width 0.5 * corridorWidth is free
	  if(team == OUR_TEAM)
	    numberOfOpponents = 
	      worldSnap->otherTeamInCorridor(playerpos, 
					     target, 0.5*corridorWidthSmall, angle, maxAge);
	  else
	    numberOfOpponents = 
	      worldSnap->ourTeamInCorridor(playerpos, 
					   target, 0.5*corridorWidthSmall, angle, maxAge);
	    
	  if(numberOfOpponents == 0)
	    {
	      GOALEVALTRACE("evaluateCorr->corridor towards position " << target <<" with smallest width ("
			    << 0.5*corridorWidthSmall << ") is free; corridorgrade = "
			    << corridorgrade);
	    }
	  else
	    // corridorgrade will be downgraded for the third time
	    {
	      corridorgrade.downgrade();
	      GOALEVALTRACE("evaluateCorr->even corridor towards position " << target <<" with smallest width ("
			    << 0.5*corridorWidthSmall << ") is not free; corridorgrade = "
			    << corridorgrade);
	    }
	}
    }
  
  // caculating the worse of distgrade and corridorgrade
  if (distgrade.is_worse_than(corridorgrade))
    grade = distgrade;
  else 
    grade = corridorgrade;

  GOALEVALTRACE("evaluateCorr->worse of corridorgrade and distgrade is " << grade);

  return grade;
}

///////////////////////////////////////////////
//
// GoalKickEval::baseSearchAndEvalForCorridor
//
//////////////////////////////////////////////
 
Grade GoalKickEval::baseSearchAndEvalForCorridor(const Snapshot *worldSnap, Vector2d playerpos,
						 Vector2d &target, bool useLeft, bool useMiddle,
						 bool useRight) const
{
  Grade grade;
  grade.set_PERFECT();


  // security distances to the border:
  const double SECURITY_DIST = KICKABLE_MARGIN + BALL_SIZE + PLAYER_SIZE;   // = 1m

  // If we are at one of the long lines, only superkick is possible, so
  // it does not make sense if we are to near to one border line
  ////////////////////////////////////////
  
  if ( fabs( playerpos.getY() ) > FIELD_WIDTH/2.0 - SECURITY_DIST ) 
    {
      grade.set_IMPOSSIBLE();
      GOALEVALTRACE ("baseSearchAndEval -> I am to near to long line: my pos= " 
		 << playerpos << "; grade = " << grade);
      return grade;
    }
  
  // If we are near to the opponent's goal line, check if
  // we are not near the goal
  ////////////////////////////////////////
  if ( (playerpos.getX() > FIELD_LENGTH/2.0 - SECURITY_DIST) 
    && (fabs(playerpos.getY()) > GOAL_WIDTH/2.0 + goalLineImpDist) ) 
    {
      grade.set_IMPOSSIBLE();
      GOALEVALTRACE ("baseSearchAndEval -> I am to near to goal line & too far away from goal:"
		 << "my pos= " << playerpos << "; grade = " << grade);
      return grade;
    }
  
  // does the player stand near the goalline so that the angle for a goalkick is
  // very small ?
  // if so the grade for goalkick can at most be AVERAGE
  
  if( ( playerpos.getX() > FIELD_LENGTH / 2. - 2 )
      && ( fabs(playerpos.getY()) > PENALTY_AREA_WIDTH / 4.) )
    {
      grade.set_AVERAGE();
      GOALEVALTRACE("baseSearchAndEval -> distance to goalline too small, grade set to "
		    << grade); 
    }
  else
    if( ( playerpos.getX() > FIELD_LENGTH / 2. - 3 )
	&& ( fabs(playerpos.getY()) > PENALTY_AREA_WIDTH / 2.) )
      {
	grade.set_AVERAGE();
	GOALEVALTRACE("baseSearchAndEval -> distance to goalline too small, grade set to "
		      << grade); 
      }
      
      

  Vector2d leftEdgeOfGoal = Vector2d(FIELD_LENGTH /2., -GOAL_WIDTH /2. + 1);
  Vector2d middleOfGoal = Vector2d(FIELD_LENGTH / 2., 0);
  Vector2d rightEdgeOfGoal = Vector2d(FIELD_LENGTH / 2., GOAL_WIDTH / 2. - 1);


  // drawing the corridors
  GOALDRAWTRACE(CORRIDOR(playerpos, leftEdgeOfGoal, corridorWidth, PI_6));
  GOALDRAWTRACE(CORRIDOR(playerpos, middleOfGoal, corridorWidth, PI_6));
  GOALDRAWTRACE(CORRIDOR(playerpos, rightEdgeOfGoal, corridorWidth, PI_6));

  // in which of the corridors does the goalie stand ? 
  // corridors with goalie in them shall not be used

  bool goalieInMiddleCorridor = false;
  bool goalieInLeftCorridor = false;
  bool goalieInRightCorridor = false;

  if( (worldSnap->otherGoalieNo != UNKNOWN) && 
      (worldSnap->otherTeam(worldSnap->otherGoalieNo).age < maxAge) )
    {
      goalieInLeftCorridor = 
	worldSnap->otherTeam(worldSnap->otherGoalieNo).inCorridor(playerpos, 
								  leftEdgeOfGoal, 
								  corridorWidth,
								  PI_6);
      if(goalieInLeftCorridor)
	GOALEVALTRACE("baseSearchAndEval -> goalie in left corridor");
      useLeft = useLeft && !goalieInLeftCorridor;

      goalieInMiddleCorridor = 
	worldSnap->otherTeam(worldSnap->otherGoalieNo).inCorridor(playerpos, 
								  middleOfGoal, 
								  corridorWidth,
								  PI_6);
      if(goalieInMiddleCorridor)
	{
	  GOALEVALTRACE("baseSearchAndEval -> goalie in middle corridor");
	}
      useMiddle = useMiddle && !goalieInMiddleCorridor;

      
      goalieInRightCorridor = 
	worldSnap->otherTeam(worldSnap->otherGoalieNo).inCorridor(playerpos, 
								  rightEdgeOfGoal, 
								  corridorWidth,
								  PI_6);
      if(goalieInRightCorridor)
	{ 
	  GOALEVALTRACE("baseSearchAndEval -> goalie in right corridor");
	}
      useRight = useRight && !goalieInRightCorridor;
    }



  // deciding which corridor should be used


  // all the corridors are unusable, that means the player is too far 
  // away from the goal and / or the goalie stands in the corridor
  if( !useLeft && !useMiddle && !useRight )
    {
      if( dist(playerpos, target) < anyCaseGoalKickDist )
 	// the dist to the target is so small, so try it on every case
 	{
 	  grade.set_PERFECT();
 	  GOALEVALTRACE("baseSearchAndEval -> dist to target "
 			<< target << " smaller than " << anyCaseGoalKickDist
			<< " ; grade = " 
 			<< grade);
 	} 
      else 
 	{  //if next to goalie, try anyway ...
	  if( dist(playerpos, worldSnap->ourTeam(worldSnap->ourGoalieNo).pos) < goalieNearDist )
	    {
	      grade.set_PERFECT();
	      GOALEVALTRACE("baseSearchAndEval -> all corridors unusable, but goalie near,"
			    << " so try to kick to " << target << " grade = " << grade);
	    } 
	  else 
	    {
	      grade.set_POOR();
	      GOALEVALTRACE("baseSearchAndEval -> all corridors unusable; grade = "
			    << grade);
	    }
 	}
      return grade;
    }


  // now considering the cases when only 1 corridor is usable 

  // only left corridor usable
  if(useLeft && !useMiddle && !useRight)
    {
      target = leftEdgeOfGoal;
      GOALEVALTRACE("baseSearchAndEval -> only left corridor usable");
      grade = evaluateCorridor(worldSnap, playerpos, target, 
			       PI_6, grade, goalieInMiddleCorridor);
      return grade;
    }
  
  // only middle corridor usable
  if(!useLeft && useMiddle && !useRight)
    {
      target = middleOfGoal;
      GOALEVALTRACE("baseSearchAndEval -> only middle corridor usable");
      grade = evaluateCorridor(worldSnap, playerpos, target, 
			       PI_6, grade, goalieInMiddleCorridor);
      return grade;
    }
  
  // only right corridor usable
  if(!useLeft && !useMiddle && useRight)
    {
      target = rightEdgeOfGoal;
      GOALEVALTRACE("baseSearchAndEval -> only right corridor usable");
      grade = evaluateCorridor(worldSnap, playerpos, target,
			       PI_6, grade, goalieInMiddleCorridor);
      return grade;
    }


  // now considering the cases when 2 corridors are usable

  // left & middle corridor usable, right corridor unusable
  if(useLeft && useMiddle && !useRight)
    {
      // which corridor is farer away from the opponents goalie?
      if( (worldSnap->otherGoalieNo != UNKNOWN) 
	  && (worldSnap->otherTeam(worldSnap->otherGoalieNo).age < maxAge) )
	{
	  double distToLeft 
	    = distToLine(worldSnap->otherTeam(worldSnap->otherGoalieNo).pos,
			 playerpos, leftEdgeOfGoal);
	  double distToMiddle
	    = distToLine(worldSnap->otherTeam(worldSnap->otherGoalieNo).pos,
			 playerpos, middleOfGoal);
	  GOALEVALTRACE("baseSearchAndEval -> goalie dist to left =   " << distToLeft);
	  GOALEVALTRACE("                     goalie dist to middle = " << distToMiddle);                      

	  if(distToMiddle > distToLeft)
	    {
	      target = middleOfGoal;
	      GOALEVALTRACE("baseSearchAndEval -> left and middle corridor usable; using middle corridor");
	      grade = evaluateCorridor(worldSnap, playerpos, target, 
				       PI_6, grade, goalieInMiddleCorridor);
	    }
	  else
	    {
	      target = leftEdgeOfGoal;
	      GOALEVALTRACE("baseSearchAndEval -> left and middle corridor usable; using left corridor");
	      grade = evaluateCorridor(worldSnap, playerpos, target,
				       PI_6, grade, goalieInMiddleCorridor);
	    }
	}
      else
	// we do not know the opponents goalie so we need another decision base
	// use the number of the opponents in the corridors
	{
	  GOALEVALTRACE("baseSearchAndEval->goalienumber unknown");
	  
	  int opponentsInLeftCorridor 
	    = worldSnap->otherTeamInCorridor(playerpos, leftEdgeOfGoal, 
					     corridorWidth, PI_6, maxAge);
	  int opponentsInMiddleCorridor
	    = worldSnap->otherTeamInCorridor(playerpos, middleOfGoal, 
					     corridorWidth, PI_6, maxAge);
	  
	  if(opponentsInMiddleCorridor < opponentsInLeftCorridor)
	    // middle corridor includes less opponents
	    {
	      target = middleOfGoal;
	      GOALEVALTRACE("baseSearchAndEval -> left and middle corridor usable; using middle corridor");
	      grade = evaluateCorridor(worldSnap, playerpos, target, 
				       PI_6, grade, goalieInMiddleCorridor);
	    }
	  else
	    // left corridor includes a less or equal number of opponents
	    {
	      target = leftEdgeOfGoal;
	      GOALEVALTRACE("baseSearchAndEval -> left and middle corridor usable; using left corridor");
	      grade = evaluateCorridor(worldSnap, playerpos, target,
				       PI_6, grade, goalieInMiddleCorridor);
	    }
	}
      return grade;
    }


  // left & right corridor usable, middle corridor unusable
  if(useLeft && useRight && !useMiddle)
    {
      if( (worldSnap->otherGoalieNo != UNKNOWN) 
	  && (worldSnap->otherTeam(worldSnap->otherGoalieNo).age < maxAge) )
	{
	  double distToLeft 
	    = distToLine(worldSnap->otherTeam(worldSnap->otherGoalieNo).pos,
			 playerpos, leftEdgeOfGoal);

	  double distToRight
	    = distToLine(worldSnap->otherTeam(worldSnap->otherGoalieNo).pos,
			 playerpos, rightEdgeOfGoal);

	  GOALEVALTRACE("baseSearchAndEval -> goalie dist to left =  " << distToLeft);
	  GOALEVALTRACE("                     goalie dist to right = " << distToRight);  
                    
	  if(distToLeft > distToRight)
	    {
	      target = leftEdgeOfGoal;
	      GOALEVALTRACE("baseSearchAndEval -> left and right corridor usable; using left corridor");
	      grade = evaluateCorridor(worldSnap, playerpos, target, 
				       PI_6, grade, goalieInMiddleCorridor);
	    }
	  else
	    {
	      target = rightEdgeOfGoal;
	      GOALEVALTRACE("baseSearchAndEval -> left and right corridor usable; using right corridor");
	      grade = evaluateCorridor(worldSnap, playerpos, target,
				       PI_6, grade, goalieInMiddleCorridor);
	    }
	}
      else
	// we do not know the opponents goalie so we need another decision base
	// use the number of the opponents in the corridors
	{
	  GOALEVALTRACE("baseSearchAndEval->goalienumber unknown");

	  int opponentsInLeftCorridor 
	    = worldSnap->otherTeamInCorridor(playerpos, leftEdgeOfGoal, 
					     corridorWidth, PI_6, maxAge);
	  int opponentsInRightCorridor 
	    = worldSnap->otherTeamInCorridor(playerpos, rightEdgeOfGoal, 
					     corridorWidth, PI_6, maxAge);

	  
	  if(opponentsInLeftCorridor < opponentsInRightCorridor)
	    // left corridor includes less opponents
	    {
	      target = leftEdgeOfGoal;
	      GOALEVALTRACE("baseSearchAndEval -> left and right corridor usable; using left corridor");
	      grade = evaluateCorridor(worldSnap, playerpos, target, 
				       PI_6, grade, goalieInMiddleCorridor);
	    }
	  else
	    if(opponentsInLeftCorridor == opponentsInRightCorridor)
	      // number of opponents is the same
	      // so take the one with the shorter distance to the goal
	      {
		double distLeft = distToLeftEdge(playerpos);
		double distRight = distToRightEdge(playerpos);
		if (distLeft < distRight)
		  {
		    target = leftEdgeOfGoal;
		    GOALEVALTRACE("baseSearchAndEvalOur -> left & right corridor usable and equal; using left corridor");
		    grade = evaluateCorridor(worldSnap, playerpos, target, 
					     PI_6, grade, goalieInMiddleCorridor);
		  }
		else
		  {
		    target = rightEdgeOfGoal;
		    GOALEVALTRACE("baseSearchAndEvalOur -> left & right corridor usable & equal; using right corridor");
		    grade = evaluateCorridor(worldSnap, playerpos, target, 
					     PI_6, grade, goalieInMiddleCorridor);
		  }
	      }
	    else
	      // right corridor includes less opponents
	      {
		target = rightEdgeOfGoal;
		GOALEVALTRACE("baseSearchAndEvalOur -> left and right corridor usable; using right corridor");
		grade = evaluateCorridor(worldSnap, playerpos, target, 
					 PI_6, grade, goalieInMiddleCorridor);
	      }
	} // goalie not known
      
      return grade;
    }
  
  // middle & right corridor usable, left corridor unusable
  if(useMiddle && useRight && !useLeft)
    {
      // which corridor is farer away from the opponents goalie?
      if( (worldSnap->otherGoalieNo != UNKNOWN) 
	  && (worldSnap->otherTeam(worldSnap->otherGoalieNo).age < maxAge) )
	{
	  double distToMiddle
	    = distToLine(worldSnap->otherTeam(worldSnap->otherGoalieNo).pos,
			 playerpos, middleOfGoal);
	  double distToRight 
	    = distToLine(worldSnap->otherTeam(worldSnap->otherGoalieNo).pos,
			 playerpos, rightEdgeOfGoal);

	  GOALEVALTRACE("baseSearchAndEval -> goalie dist to middle =  " << distToMiddle);
	  GOALEVALTRACE("                     goalie dist to right  = " << distToRight);  
	  
	  if(distToMiddle > distToRight)
	    {
	      target = middleOfGoal;
	      GOALEVALTRACE("baseSearchAndEval -> middle and right corridor usable; using middle corridor");
	      grade = evaluateCorridor(worldSnap, playerpos, target, 
				       PI_6, grade, goalieInMiddleCorridor);
	    }
	  else
	    {
	      target = rightEdgeOfGoal;
	      GOALEVALTRACE("baseSearchAndEval -> middle and right corridor usable; using right corridor");
	      grade = evaluateCorridor(worldSnap, playerpos, target,
				       PI_6, grade, goalieInMiddleCorridor);
	    }
	}
	  
      else
	// we do not know the opponents goalie so we need another decision base
	// use the number of the opponents in the corridors
	{
	  GOALEVALTRACE("baseSearchAndEval->goalienumber unknown");
	  
	  int opponentsInMiddleCorridor
	    = worldSnap->otherTeamInCorridor(playerpos, middleOfGoal, 
					     corridorWidth, PI_6, maxAge);
	  int opponentsInRightCorridor 
	    = worldSnap->otherTeamInCorridor(playerpos, rightEdgeOfGoal, 
					     corridorWidth, PI_6, maxAge);
	

	  if(opponentsInMiddleCorridor < opponentsInRightCorridor)
	    // middle corridor includes less opponents
	    {
	      target = middleOfGoal;
	      GOALEVALTRACE("baseSearchAndEval -> middle and right corridor usable; using middle corridor");
	      grade = evaluateCorridor(worldSnap, playerpos, target, 
				       PI_6, grade, goalieInMiddleCorridor);
	    }
	  else
	    // right corridor includes a less or equal number of opponents
	    {
	      target = rightEdgeOfGoal;
	      GOALEVALTRACE("baseSearchAndEvalOur -> middle and right corridor usable; using right corridor");
	      grade = evaluateCorridor(worldSnap, playerpos, target, 
				       PI_6, grade, goalieInMiddleCorridor);
	    }
	} // goalie not known
      return grade;
    }


  // all the corridors are usable, that means the goalie doesn't stand in any of
  // the corridors and the kicking distance is also ok
  if( useLeft && useMiddle && useRight)
    {
      // which corridor is farer away from the opponents goalie?
      // because all the corridors are usable we only have to consider the right
      // and the left corridor, the goalie can not be in the middle in this case
      if( (worldSnap->otherGoalieNo != UNKNOWN) 
	  && (worldSnap->otherTeam(worldSnap->otherGoalieNo).age < maxAge) )
	{
	  double distToLeft 
	    = distToLine(worldSnap->otherTeam(worldSnap->otherGoalieNo).pos,
			 playerpos, leftEdgeOfGoal);
	  double distToRight 
	    = distToLine(worldSnap->otherTeam(worldSnap->otherGoalieNo).pos,
			 playerpos, rightEdgeOfGoal);
	  
	  GOALEVALTRACE("baseSearchAndEval -> goalie dist to left  = " << distToLeft);
	  GOALEVALTRACE("                     goalie dist to right = " << distToRight);  

	  if (distToLeft > distToRight)
	    {
	      target = leftEdgeOfGoal;
	      GOALEVALTRACE("baseSearchAndEval -> all corridors usable; using left corridor");
	      grade = evaluateCorridor(worldSnap, playerpos, target, 
				       PI_6, grade, goalieInMiddleCorridor);
	    }
	  else
	    {
	      target = rightEdgeOfGoal;
	      GOALEVALTRACE("baseSearchAndEval -> all corridors usable; using right corridor");
	      grade = evaluateCorridor(worldSnap, playerpos, target,
				       PI_6, grade, goalieInMiddleCorridor);
	    }
	}
      else
	// we do not know the opponent's goalie so we need another decision base
	// use the number of the opponents in the corridors
	{
	  GOALEVALTRACE("baseSearchAndEval->goalienumber unknown");
	
	  int opponentsInLeftCorridor 
	    = worldSnap->otherTeamInCorridor(playerpos, leftEdgeOfGoal, 
					     corridorWidth, PI_6, maxAge);
	  int opponentsInMiddleCorridor
	    = worldSnap->otherTeamInCorridor(playerpos, middleOfGoal,
					     corridorWidth, PI_6, maxAge);

	  int opponentsInRightCorridor 
	    = worldSnap->otherTeamInCorridor(playerpos, rightEdgeOfGoal, 
					     corridorWidth, PI_6, maxAge);


	  if( (opponentsInLeftCorridor == opponentsInMiddleCorridor) &&
	      (opponentsInLeftCorridor == opponentsInRightCorridor) )
	    // all the corridors include the same number of opponents
	    // take the one with the shortest distance to the goal that means don't change
	    // the goalTarget
	    {
	      GOALEVALTRACE("baseSearchAndEval -> all corridors usable & equal, using shortest corridor (target = " 
			    << target << " )" );
	      grade = evaluateCorridor(worldSnap, playerpos, target, PI_6,
				       grade, goalieInMiddleCorridor);
	    }
	  else
	    // at most two corridors have an equal number of opponents
	    {
	      if(opponentsInLeftCorridor < opponentsInMiddleCorridor)
		{
		  if(opponentsInRightCorridor < opponentsInLeftCorridor)
		    // right corridor includes the least number of opponents
		    {
		      target = rightEdgeOfGoal;
		      GOALEVALTRACE("baseSearchAndEval -> all corridors usable, right corridor has least number of opponents");
		      grade = evaluateCorridor(worldSnap, playerpos, target, PI_6,
					       grade, goalieInMiddleCorridor);
		    }
		  else
		    if(opponentsInRightCorridor == opponentsInLeftCorridor)
		      // number of opponents in left and right corridor is the same
		      // so take the one with the shorter distance to the goal
		      {
			double distLeft = distToLeftEdge(playerpos);
			double distRight = distToRightEdge(playerpos);
		    
			if (distLeft < distRight)
			  {
			    target = leftEdgeOfGoal;
			    GOALEVALTRACE("baseSearchAndEval -> all corridors usable, left & right corridor equal; using left corridor");
			    grade = evaluateCorridor(worldSnap, playerpos, target, PI_6, 
						     grade, goalieInMiddleCorridor);
			  }
			else
			  {
			    target = rightEdgeOfGoal;
			    GOALEVALTRACE("baseSearchAndEval -> all corridors usabel, left & right corridor equal; using right corridor");
			    grade = evaluateCorridor(worldSnap, playerpos, target, 
						     PI_6, grade, goalieInMiddleCorridor);
			  }
		      }
		    else
		      // left corridor includes the least number of opponents
		      {
			target = leftEdgeOfGoal;
			GOALEVALTRACE("baseSearchAndEval -> all corridors usable, left corridor has least number of opponents");
			grade = evaluateCorridor(worldSnap, playerpos, target, 
						 PI_6, grade, goalieInMiddleCorridor);
		      }
		}
	      else
		// number of opponents in middle corridor <= number of opponents in left corridor
		{
		  if(opponentsInRightCorridor < opponentsInMiddleCorridor)
		    // right corridor includes the least number of opponents
		    {
		      target = rightEdgeOfGoal;
		      GOALEVALTRACE("baseSearchAndEval -> all corridors usable, right corridor has least number of opponents");
		      grade = evaluateCorridor(worldSnap, playerpos, target, 
					       PI_6, grade, goalieInMiddleCorridor);
		    }
		  else
		    if(opponentsInRightCorridor == opponentsInMiddleCorridor)
		      // number of opponents in middle & right corridor is the same and - because
		      // at most two corridors may have the same number of opponents - smaller
		      // than the number of opponents in the left corridor
		      {
			target = rightEdgeOfGoal;
			GOALEVALTRACE("baseSearchAndEval -> all corridors usable, middle & right corridor equal, using right corridor");
			grade = evaluateCorridor(worldSnap, playerpos, target, PI_6,
						 grade, goalieInMiddleCorridor);
		      }
		    else
		      // middle corridor includes less opponents than the right corridor
		      {
			if(opponentsInLeftCorridor == opponentsInMiddleCorridor)
			  {
			    target = leftEdgeOfGoal;
			    GOALEVALTRACE("baseSearchAndEval -> all corridors usable, middle & left corridor equal, using left corridor");
			    grade = evaluateCorridor(worldSnap, playerpos, target, 
						     PI_6, grade, goalieInMiddleCorridor);
			  }
			else
			  // middle corridor has least number of opponents
			  {
			    target = middleOfGoal;
			    GOALEVALTRACE("baseSearchAndEval -> all corridors usable, middle corridor has least number of opponents");
			    grade = evaluateCorridor(worldSnap, playerpos, target,
						     PI_6, grade, goalieInMiddleCorridor);
			  }
		      }
		} // #middle <= #left 
	    } // not all equal
	  return grade;
	} // all corridors usable
    }// goalie not known


  // falls was schiefging
  grade.set_BAD();
  return grade;
}


///////////////////////////////////////////////
//
// GoalKickEval::searchAndOurEvalBestCorridor
//
//////////////////////////////////////////////

Grade GoalKickEval::searchAndEvalOurBestCorridor(const Snapshot *worldSnap, Vector2d playerpos,
						 double posKickLength, Vector2d &target) const
{
  Grade grade;

  // setting the important positions in the goal

  Vector2d leftEdgeOfGoal;
  Vector2d middleOfGoal;
  Vector2d rightEdgeOfGoal;

  leftEdgeOfGoal = Vector2d(FIELD_LENGTH /2., -GOAL_WIDTH /2. + 1);
  middleOfGoal = Vector2d(FIELD_LENGTH / 2., 0);
  rightEdgeOfGoal = Vector2d(FIELD_LENGTH / 2., GOAL_WIDTH / 2. - 1);
      
  // calculate the distances of the player to the left edge, the middle
  // and the right edge of the goal
  
  double distToLeft 
    = distToLeftEdge(playerpos);
  double distToMiddleOfGoal 
    = distToMiddle(playerpos);
  double distToRight 
    = distToRightEdge(playerpos);

  // now the player looks for the best corridor towards the goal
  // by first deciding which corridors(s) are possible by the given posKickLength

  bool useLeftCorridor = true;
  bool useMiddleCorridor = true;
  bool useRightCorridor = true;
  
  if( distToLeft > posKickLength)
    {
      GOALEVALTRACE("searchAndEvalOur->left Edge too far away; left corridor unusable");
      useLeftCorridor = false;
    }
  if( distToMiddleOfGoal > posKickLength)
    {
      GOALEVALTRACE("searchAndEvalOur->middle of goal too far away; middle corridor unusable");
      useMiddleCorridor = false;
    }
  if( distToRight > posKickLength)
    {
      GOALEVALTRACE("searchAndEvalOur->right edge of goal too far away; right corridor unusable");
      useRightCorridor = false;
    }

  grade = baseSearchAndEvalForCorridor(worldSnap, playerpos, target, useLeftCorridor,
				       useMiddleCorridor, useRightCorridor);
  return grade;
}


////////////////////////////////////////////////////
//
// GoalKickEval::searchAndEvalOtherBestCorridor
//
///////////////////////////////////////////////////


Grade GoalKickEval::searchAndEvalOtherBestCorridor(const Snapshot *worldSnap, Vector2d playerpos,
						   double posKickLength, Vector2d &target) const
{
  Grade grade;
  grade.set_PERFECT();

  // setting the important positions in the goal

  Vector2d leftEdgeOfGoal;
  Vector2d middleOfGoal;
  Vector2d rightEdgeOfGoal;

  leftEdgeOfGoal = Vector2d(-FIELD_LENGTH /2., -GOAL_WIDTH /2. + 1);
  middleOfGoal = Vector2d(-FIELD_LENGTH / 2., 0);
  rightEdgeOfGoal = Vector2d(-FIELD_LENGTH / 2., GOAL_WIDTH / 2. - 1);
      
  // calculate the distances of the player to the left edge, the middle
  // and the right edge of the goal
  
  double distToLeft 
    = distToLeftEdge(playerpos, OTHER_TEAM);
  double distToMiddleOfGoal 
    = distToMiddle(playerpos, OTHER_TEAM);
  double distToRight 
    = distToRightEdge(playerpos, OTHER_TEAM);

  // now the player looks for the best corridor towards the goal
  // by first deciding which corridors(s) are possible by the given posKickLength

  bool useLeftCorridor = true;
  bool useMiddleCorridor = true;
  bool useRightCorridor = true;
  
  if( distToLeft > posKickLength)
    {
      GOALEVALTRACE("searchAndEvalOur->left Edge too far away; left corridor unusable");
      useLeftCorridor = false;
    }
  if( distToMiddleOfGoal > posKickLength)
    {
      GOALEVALTRACE("searchAndEvalOur->middle of goal too far away; middle corridor unusable");
      useMiddleCorridor = false;
    }
  if( distToRight > posKickLength)
    {
      GOALEVALTRACE("searchAndEvalOur->right edge of goal too far away; right corridor unusable");
      useRightCorridor = false;
    }


  // in which of the corridors does the goalie stand ? 
  // corridors with goalie in them shall not be used

  bool goalieInMiddleCorridor = false;

  if( (worldSnap->ourGoalieNo != UNKNOWN) && 
      (worldSnap->ourTeam(worldSnap->ourGoalieNo).age < maxAge) )
    {
      bool goalieInCorridor;

      goalieInCorridor = 
	worldSnap->ourTeam(worldSnap->ourGoalieNo).inCorridor(playerpos, 
							      leftEdgeOfGoal, 
							      corridorWidth,
							      PI_6);
      if(goalieInCorridor)
	GOALEVALTRACE("searchAndEvalOther -> goalie in left corridor");
      useLeftCorridor = useLeftCorridor && !goalieInCorridor;

      goalieInCorridor = 
	worldSnap->ourTeam(worldSnap->ourGoalieNo).inCorridor(playerpos, 
							      middleOfGoal, 
							      corridorWidth,
							      PI_6);
      if(goalieInCorridor)
	{
	  GOALEVALTRACE("searchAndEvalOther -> goalie in middle corridor");
	  useMiddleCorridor = useMiddleCorridor && !goalieInCorridor;
	  goalieInMiddleCorridor = true;
	}

      goalieInCorridor = 
	worldSnap->ourTeam(worldSnap->ourGoalieNo).inCorridor(playerpos, 
							      rightEdgeOfGoal, 
							      corridorWidth,
							      PI_6);
      if(goalieInCorridor)
	GOALEVALTRACE("searchAndEvalOther -> goalie in right corridor");
      useRightCorridor = useRightCorridor && !goalieInCorridor;
    }

  // deciding which corridor should be used


  // all the corridors are unusable, that means the player is too far 
  // away from the goal and / or the goalie stands in the corridor
  if( !useLeftCorridor && !useMiddleCorridor && !useRightCorridor )
    {
      grade.set_POOR();
      GOALEVALTRACE("searchAndEvalOther -> all corridors unusable; grade = "
		<< grade);
      return grade;
    }

  // for all other cases we need the number of opponents in each of the 
  // corridors

  // count the opponent players in the left, right and the middle corridor from 
  // the players position to the left /right edge /the middle of the goal
  // corridor has angle PI_6
  int opponentsInLeftCorridor 
    = worldSnap->ourTeamInCorridor(playerpos, 
				   leftEdgeOfGoal, corridorWidth, PI_6, maxAge);
  int opponentsInMiddleCorridor 
    = worldSnap->ourTeamInCorridor(playerpos, 
				   middleOfGoal, corridorWidth, PI_6, maxAge);
  int opponentsInRightCorridor 
    = worldSnap->ourTeamInCorridor(playerpos, 
				   rightEdgeOfGoal, corridorWidth, PI_6, maxAge);

  GOALEVALTRACE("seachAndEvalOther -> opponents in left corridor   " 
		<< opponentsInLeftCorridor );
  GOALEVALTRACE("                                  middle corridor "
		<< opponentsInMiddleCorridor );
  GOALEVALTRACE("                                  right corridor  " 
		<< opponentsInRightCorridor );

  // all the corridors are usable, that means the goalie doesn't stand in any of
  // the corridors and the kicking distance is also ok
  if( useLeftCorridor && useMiddleCorridor && useRightCorridor)
    {
      // search for the corridor with the least number of opponents

      if( (opponentsInLeftCorridor == opponentsInMiddleCorridor) &&
	  (opponentsInLeftCorridor == opponentsInRightCorridor) )
	// all the corridors include the same number of opponents
	// take the one with the shortest distance to the goal that means don't change
	// the goalTarget
	{
	  GOALEVALTRACE("searchAndEvalOther -> all corridors usable & equal, using shortest corridor (target = " 
			<< target << ")");
	  grade = evaluateCorridor(worldSnap, playerpos, target, PI_6,
				   grade, goalieInMiddleCorridor, OTHER_TEAM);
	}
      else
	// at most two corridors have an equal number of opponents
	{
	  if(opponentsInLeftCorridor < opponentsInMiddleCorridor)
	    {
	      if(opponentsInRightCorridor < opponentsInLeftCorridor)
		// right corridor includes the least number of opponents
		{
		  target = rightEdgeOfGoal;
		  GOALEVALTRACE("searchAndEvalOther -> all corridors usable, right corridor has least number of opponents");
		  grade = evaluateCorridor(worldSnap, playerpos, target, PI_6,
					   grade, goalieInMiddleCorridor, OTHER_TEAM);
		}
	      else
		if(opponentsInRightCorridor == opponentsInLeftCorridor)
		  // number of opponents in left and right corridor is the same
		  // so take the one with the shorter distance to the goal
		  {
		    if (distToLeft < distToRight)
		      {
			target = leftEdgeOfGoal;
			GOALEVALTRACE("searchAndEvalOther -> all corridors usable, left & right corridor equal; using left corridor");
			grade = evaluateCorridor(worldSnap, playerpos, target, PI_6, 
						 grade, goalieInMiddleCorridor, OTHER_TEAM);
		      }
		    else
		      {
			target = rightEdgeOfGoal;
			GOALEVALTRACE("searchAndEvalOther -> all corridors usabel, left & right corridor equal; using right corridor");
			grade = evaluateCorridor(worldSnap, playerpos, target, PI_6, 
						 grade, goalieInMiddleCorridor, OTHER_TEAM);
		      }
		  }
		else
		  // left corridor includes the least number of opponents
		  {
		    target = leftEdgeOfGoal;
		    GOALEVALTRACE("searchAndEvalOther -> all corridors usable, left corridor has least number of opponents");
		    grade = evaluateCorridor(worldSnap, playerpos, target, PI_6, 
					     grade, OTHER_TEAM);
		  }
	    }
	  else
	    // number of opponents in middle corridor <= number of opponents in left corridor
	    {
	      if(opponentsInRightCorridor < opponentsInMiddleCorridor)
		// right corridor includes the least number of opponents
		{
		  target = rightEdgeOfGoal;
		  GOALEVALTRACE("searchAndEvalOther -> all corridors usable, right corridor has least number of opponents");
		  grade = evaluateCorridor(worldSnap, playerpos, target, PI_6, 
					   grade, goalieInMiddleCorridor, OTHER_TEAM);
		}
	      else
		if(opponentsInRightCorridor == opponentsInMiddleCorridor)
		  // number of opponents in middle & right corridor is the same and - because
		  // at most two corridors may have the same number of opponents - smaller
		  // than the number of opponents in the left corridor
		  {
		    target = rightEdgeOfGoal;
		    GOALEVALTRACE("searchAndEvalOther -> all corridors usable, middle & right corridor equal, using right corridor");
		    grade = evaluateCorridor(worldSnap, playerpos, target, PI_6,
					    grade, goalieInMiddleCorridor, OTHER_TEAM);
		  }
		else
		// middle corridor includes less opponents than the right corridor
		  {
		    if(opponentsInLeftCorridor == opponentsInMiddleCorridor)
		      {
			target = leftEdgeOfGoal;
			GOALEVALTRACE("searchAndEvalOther -> all corridors usable, middle & left corridor equal, using left corridor");
			grade = evaluateCorridor(worldSnap, playerpos, target, PI_6, 
						 grade, goalieInMiddleCorridor, OTHER_TEAM);
		      }
		    else
		      // middle corridor has least number of opponents
		      {
			target = middleOfGoal;
			GOALEVALTRACE("searchAndEvalOther -> all corridors usable, middle corridor has least number of opponents");
			grade = evaluateCorridor(worldSnap, playerpos, target, PI_6, 
						 grade, goalieInMiddleCorridor, OTHER_TEAM);
		      }
		  }
	    } // #middle <= #left 
	} // not all equal
      return grade;
    } // all corridors usable

  // now considering the cases when only 1 corridor is usable 

  // only left corridor usable
  if(useLeftCorridor && !useMiddleCorridor && !useRightCorridor)
    {
      target = leftEdgeOfGoal;
      GOALEVALTRACE("searchAndEvalOther -> only left corridor usable");
      grade = evaluateCorridor(worldSnap, playerpos, target, PI_6, 
			       grade, goalieInMiddleCorridor, OTHER_TEAM);
      return grade;
    }
  
  // only middle corridor usable
  if(!useLeftCorridor && useMiddleCorridor && !useRightCorridor)
    {
      target = middleOfGoal;
      GOALEVALTRACE("searchAndEvalOther -> only middle corridor usable");
      grade = evaluateCorridor(worldSnap, playerpos, target, PI_6, 
			       grade, goalieInMiddleCorridor, OTHER_TEAM);
      return grade;
    }
  
  // only right corridor usable
  if(!useLeftCorridor && !useMiddleCorridor && useRightCorridor)
    {
      target = rightEdgeOfGoal;
      GOALEVALTRACE("searchAndEvalOther -> only right corridor usable");
      grade = evaluateCorridor(worldSnap, playerpos, target, PI_6, 
			       grade,  goalieInMiddleCorridor, OTHER_TEAM);
      return grade;
    }

  // now considering the case when 2 corridors are usable

  // left & middle corridor usable, right corridor unusable
  if(useLeftCorridor && useMiddleCorridor && !useRightCorridor)
    {
      if(opponentsInMiddleCorridor < opponentsInLeftCorridor)
	// middle corridor includes less opponents
	{
	  target = middleOfGoal;
	  GOALEVALTRACE("searchAndEvalOther -> left and middle corridor usable; using middle corridor");
	  grade = evaluateCorridor(worldSnap, playerpos, target, PI_6, 
				   grade, goalieInMiddleCorridor, OTHER_TEAM);
	}
      else
	// left corridor includes a less or equal number of opponents
	{
	  target = leftEdgeOfGoal;
	  GOALEVALTRACE("searchAndEvalOther -> left and middle corridor usable; using left corridor");
	  grade = evaluateCorridor(worldSnap, playerpos, target, PI_6, 
				   grade, goalieInMiddleCorridor, OTHER_TEAM);
	}
      return grade;
    }

  // left & right corridor usable, middle corridor unusable
  if(useLeftCorridor && useRightCorridor && !useMiddleCorridor)
    {
      if(opponentsInLeftCorridor < opponentsInRightCorridor)
	// left corridor includes less opponents
	{
	  target = leftEdgeOfGoal;
	  GOALEVALTRACE("searchAndEvalOther -> left and right corridor usable; using left corridor");
	  grade = evaluateCorridor(worldSnap, playerpos, target, PI_6, 
				   grade, goalieInMiddleCorridor, OTHER_TEAM);
	}
      else
	if(opponentsInLeftCorridor == opponentsInRightCorridor)
	  // number of opponents is the same
	  // so take the one with the shorter distance to the goal
	  {
	    if (distToLeft < distToRight)
	      {
		target = leftEdgeOfGoal;
		GOALEVALTRACE("searchAndEvalOther -> left & right corridor usable and equal; using left corridor");
		grade = evaluateCorridor(worldSnap, playerpos, target, PI_6, 
					 grade, goalieInMiddleCorridor, OTHER_TEAM);
	      }
	    else
	      {
		target = rightEdgeOfGoal;
		GOALEVALTRACE("searchAndEvalOther -> left & right corridor usable & equal; using right corridor");
		grade = evaluateCorridor(worldSnap, playerpos, target, PI_6, 
					 grade, goalieInMiddleCorridor, OTHER_TEAM);
	      }
	  }
	else
	  // right corridor includes less opponents
	  {
	    target = rightEdgeOfGoal;
	    GOALEVALTRACE("searchAndEvalOther -> left and right corridor usable; using right corridor");
	    grade = evaluateCorridor(worldSnap, playerpos, target, PI_6, 
				     grade, goalieInMiddleCorridor, OTHER_TEAM);
	  }
      return grade;
    }
  
  // middle & right corridor usable, left corridor unusable
  if(useMiddleCorridor && useRightCorridor && !useLeftCorridor)
    {
      if(opponentsInMiddleCorridor < opponentsInRightCorridor)
	// middle corridor includes less opponents
	{
	  target = middleOfGoal;
	  GOALEVALTRACE("searchAndEvalOther -> middle and right corridor usable; using middle corridor");
	  grade = evaluateCorridor(worldSnap, playerpos, target, PI_6, 
				   grade, goalieInMiddleCorridor, OTHER_TEAM);
	}
      else
	// right corridor includes a less or equal number of opponents
	{
	  target = rightEdgeOfGoal;
	  GOALEVALTRACE("searchAndEvalOther -> middle and right corridor usable; using right corridor");
	  grade = evaluateCorridor(worldSnap, playerpos, target, PI_6, 
				   grade, goalieInMiddleCorridor, OTHER_TEAM);
	}
      return grade;
    }

  // if anything went wrong
  grade.set_BAD();
  return grade;
}


