//////////////////////////////////////////////////////////////////////
// Mainz Rolling Brains RoboCup agents developed for the Simulator 
// League of the RoboCup 2000 competition
//
// Copyright (C) 1998, 1999, 2000 Team Members 
// of the Mainz Rolling Brains RoboCup Team 
//
// This program under the terms of the GNU General Public License as
// published by the Free Software Foundation; version 2 of the
// License. See the README file in the root directory of our archive.
//////////////////////////////////////////////////////////////////////

//////////////////////////////////////////////////////////////////////
//
// MRB player class goalkickeval
//
// \$Id: goalkickeval.h,v 1.19 2000/10/19 15:27:52 goetz Exp $
// \$manager: schappel$
//
//////////////////////////////////////////////////////////////////////


#ifndef __GOALKICKEVAL_H__
#define __GOALKICKEVAL_H__

#include "grade.h"
#include "drawMacros.h"
#include "worldmodel.h"
#include "Config.h"


// the GOALEVALTRACE
////////////////////////

#undef GOALEVALTRACE

// use this line if you want prompting of players decision process
//#define GOALEVALTRACE(out) cout << worldSnap->time << " GOAL: " << out << endl 

#ifndef GOALEVALTRACE
#define GOALEVALTRACE(cmd)
#else
#warning GOALEVALTRACE enabled
#endif


#undef GOALEVALTRACESHORT

// use this line if you want prompting of players decision process
//#define GOALEVALTRACESHORT(out) cout << "GOAL: " << out << endl 

#ifndef GOALEVALTRACESHORT
#define GOALEVALTRACESHORT(out)
#else
#warning GOALEVALTRACESHORT enabled
#endif

#undef GOALDRAWTRACE
// comment out the following line to disable trace
//#define GOALDRAWTRACE(output) cout << worldSnap->time << "% " << output << endl

#ifndef GOALDRAWTRACE
#define GOALDRAWTRACE(cmd)
#else
#warning GOALDRAWTRACE enabled
#endif


// enum-type for corridors towards the goal
//////////////////////////////////////////////

enum corridor_type{
  CORRIDOR_LEFT,
  CORRIDOR_MIDDLE,
  CORRIDOR_RIGHT
};


// enum-type for our team / the other team
//////////////////////////////////////////////

enum team_type{
  OUR_TEAM,
  OTHER_TEAM
};


// the class GoalKickEval
///////////////////////////////

class GoalKickEval
{
public:
  // constructor of module GoalKick 
  GoalKickEval();

  ~GoalKickEval();
  
  // do the evaluation for a player of our team
  // For definition of Grade: see grade.h
  Grade ourEval(const Snapshot* worldSnap, Vector2d playerpos) const;

  // do the evaluation for a player of the other team
  // For definition of Grade: see grade.h
  Grade otherEval(const Snapshot* worldSnap, Vector2d playerpos) const;


private:
 
  // some constants this class needs
  //////////////////////////////////////

  // duration a superkick / fixed kick with 1 turn / fixed kick / 
  // kick away needs to be performed
  int superKickDuration;
  int fixedKickWithTurnDuration; // with 1 cycle for turning
  int fixedKickDuration;          // without turning!
  int kickAwayDuration;

  // distance a ball kicked with a superkick / fixed kick / kickaway
  // may run

  double superKickDist;
  double fixedKickDist;
  double kickAwayDist;

  // the lower bound for the distance when a goalkick will get 
  // distancegrade IMPOSSIBLE
  double distGradeIsImpDist;

  // the lower bound for the distance when a goalkick will get 
  // distancegrade BAD
  double distGradeIsBadDist;

  // the lower bound for the distance when a goalkick will get 
  // distancegrade POOR
  double distGradeIsPoorDist;

  // the lower bound for the distance when a goalkick will get 
  // distancegrade AVERAGE
  double distGradeIsAverageDist;

  // the lower bound for the distance when a goalkick will get 
  // distancegrade GOOD
  double distGradeIsGoodDist;

  // the distance when a goalkick will get garde PERFECT in any 
  // case
  double anyCaseGoalKickDist;

  // the distance to the goalie when a player decides to kick 
  // anyway
  double goalieNearDist;

  // the width of the corridor from the player to the selected
  // target where the other's goalie shall not be in
  double corridorWidth;

  // the width the corridor shall have when the corridor with
  // the above width ist not free
  double corridorWidthSmall;
  
  // elder things will not be considered
  int maxAge;   

  // next to the goalline (SECURITY_DIST in goalkickeval.C) a 
  // goalkick is considered impossible, if farer away from goal than
  double goalLineImpDist;

  // some private methods
  /////////////////////////////////////////

  // calculates the distance of the player to the left edge
  // of the opponents goal
  double distToLeftEdge(Vector2d playerpos,
			team_type team = OUR_TEAM) const;

  // calculates the distance of the player to the middle
  // of the opponents goal
  double distToMiddle(Vector2d playerpos,
		      team_type team = OUR_TEAM) const;

  // calculates the distance of the player to the right edge
  // of the opponents goal
  double distToRightEdge(Vector2d playerpos,
			 team_type team = OUR_TEAM) const;

  // calculates the players minimum distance to the opponents goal,
  // considering the left and right edge as well as
  // the middle of the goal; returns the minimum distance
  // and writes the belonging postion in the goal into
  // target
  double minDistToGoal(Vector2d playerpos,
		       Vector2d &target, team_type team = OUR_TEAM) const;

  // answer the question, how much time a player has for doing a kick
  // before an opponent can reach the ball
  // the flag team tells the method if the player is from the own
  // or from the other team
  // if there is enough time for a 
  //    superkick                   return SUPERKICK_DURATION
  //    fixed kick with turning,    return FIXEDKICK_WITHTURN_DURATION  
  //    fixed kick,                 return FIXEDKICK_DURATION
  //    kick away,                  return KICKAWAY_DURATION
  //    else                        return 0
  int availableTimeForKick(const Snapshot *worldSnap, Vector2d playerpos, 
			   team_type team = OUR_TEAM) const;

  // answers the question how long the kick can be with the given
  // available time for preparing the kick
  double possibleKickLength(int time) const;

  // calculates the distance of an object to the straight line between
  // source and target
  // returns if the projection of the object onto the line source->target
  // is not a point on this line the function returns either
  // the dist to the source or the dist to target
  double distToLine(Vector2d object, Vector2d source, Vector2d target) const;  


  // evaluates the given corridor
  Grade evaluateCorridor(const Snapshot *worldSnap, Vector2d playerpos, Vector2d target, 
			 Angle angle, Grade startGrade, 
			 bool goalieInMiddleCorridor,
			 team_type team = OUR_TEAM) const;

  // determines the best corridor for our team and evaluates it; 
  // returns the grade for the best corridor and the belonging target
  Grade baseSearchAndEvalForCorridor(const Snapshot *worldSnap, Vector2d playerpos,
				     Vector2d &target, bool useLeft, bool useMiddle,
				     bool useRight) const;

  // determines the best corridor for our team and evaluates it using 
  // baseSerachAndEvalForCorridor and some more criteria; 
  // returns the grade for the best corridor and the belonging target
  Grade searchAndEvalOurBestCorridor(const Snapshot *worldSnap, Vector2d playerpos,
				     double posKickLength, Vector2d &target) const;

  // determines the best corridor for the other team and evaluates it; 
  // returns the grade for the best corridor
  Grade searchAndEvalOtherBestCorridor(const Snapshot *worldSnap, Vector2d playerpos,
				       double posKickLength, Vector2d &target) const;

  friend class GoalKickModule;    // this module will use some
                                  // of the private methods
 
}; // class GoalKickEval



#endif

