//////////////////////////////////////////////////////////////////////
// Mainz Rolling Brains RoboCup agents developed for the Simulator 
// League of the RoboCup 2000 competition
//
// Copyright (C) 1998, 1999, 2000 Team Members 
// of the Mainz Rolling Brains RoboCup Team 
//
// This program under the terms of the GNU General Public License as
// published by the Free Software Foundation; version 2 of the
// License. See the README file in the root directory of our archive.
//////////////////////////////////////////////////////////////////////

//////////////////////////////////////////////////////////////////////
//
// MRB player PassEval
//
// \$Id: passeval.h,v 1.20 2000/10/19 15:27:53 goetz Exp $
// \$manager: holgi$
//
//////////////////////////////////////////////////////////////////////

#ifndef __PASSEVAL_H__   
#define __PASSEVAL_H__

#include "grade.h"
#include "worldmodel.h"
#include "goalkickeval.h"
#include "AdvSkills.h"

#undef PAEVTRACE
// comment out the following line, if trace shall be disabled
// #define PAEVTRACE(cmd) cmd
#ifndef PAEVTRACE
#define PAEVTRACE(cmd)
#else
#warning PAEVTRACE enabled
#endif

#undef TESTPAEVTRACE
// comment out the following line, if trace shall be disabled
// #define TESTPAEVTRACE(cmd) cmd
#ifndef TESTPAEVTRACE
#define TESTPAEVTRACE(cmd)
#else
#warning TESTPAEVTRACE enabled
#endif

struct PropOf {
  int      number;    // number of partner
  Vector2d pos;       // pos to pass to; not necessary pos of partner! 
  double   dist;      // distance between player and target pos
  double   vel;       // velocity that ball should reach at position
  int      free;      // radius of free circle around pos
  int      corridor;  // free corridor width
  bool     offside;   // partner is not in an offside-positon
  bool     perhaps;   // true, if pass makes sense
  Grade    grade;     // grade of the pass from player to pos
};

// the class PassEval
//////////////////////////

class PassEval
{
public:
  // constructor
  PassEval(const GoalKickEval* _goalkickeval);
  // destructor
  ~PassEval();

  // this methods needs a worldSnap, a playernumber and a partnernumber;
  // it grades the passes from player to partner on the basis of the worldSnap,
  // considering also rangepasses;
  // the position where the pass would be played to is reported in posOfPass;
  Grade ourEval(const Snapshot* worldSnap, int plnumber, int panumber, Vector2d &posOfPass) const;
  
  // just for calling the above method without posOfPass
  Grade ourEval(const Snapshot* worldSnap, int plnumber, int panumber) const
    { Vector2d v; return ourEval(worldSnap, plnumber, panumber, v); };

  // does the same as above but reports PropOf of best pass
  PropOf ourEvalProp(const Snapshot* worldSnap, int plnumber, int panumber) const;

  // this methods does the same, but for players of the other team;
  // it takes no features of the other team into account;
  // does not consider rangepasses
  Grade otherEval(const Snapshot* worldSnap, int plnumber, int panumber) const;

  // get the "better" pass out of two;
  // true, if good is better than bad, else false;
  bool betterPass(const Snapshot *wS, PropOf good, PropOf bad) const;

private:
  // this module is used by the passmodule of a player, too!
  friend class PassModule;

  // some things this module needs:
  /////////////////////////////////////////

  const GoalKickEval* goalkickeval;
 
  // the dangerzone for freestanding-check, probably adaptive?
  double DANGERZONE;

  // to decide, when an player is on the same level with another player
  // I don't want to do it with the exact x-pos
  double SAMELEVEL_LENGTH;
  
  // duration that a superkick/fixed kick/kick away needs to be performed
  int SUPERKICK_DURATION;
  int FIXEDKICK_DURATION;           // without turning!
  int KICKAWAY_DURATION;

  double SUPERKICK_LENGTH;          // some estimated values, communicated by CM
  double FIXEDKICK_LENGTH;
  double KICKAWAY_LENGTH;

  double RPASS_LATTICELENGTH;
  
  // length of a pass that is too short to play it if the sit. is not dangerous
  double SHORTPASS_LENGTH;

  // some private methods
  /////////////////////////////////////////

  // answer the question, if the player can do a safe kick/fixed kick/superkick
  // if there is enough time for a superkick,               return SUPERKICK_DURATION
  //                               fixed kick,              return FIXEDKICK_DURATION
  //                               kick away,               return KICKAWAY_DURATION
  // player is the number of the player that shall pass;
  int check_for_safekick(const Snapshot *wS, int player) const;
 
  // check the passlength that player can manage;
  // the "maximum" length is returned
  double check_passlength(const Snapshot *wS, int player) const;

  // get the PropOf of a pass from player no. "plnumber" to "target"
  // the receiver of the pass is its teammate with no. "panumber"
  // if we == true, the method works for our team, else for the opponent
  bool get_pPass(bool we, const Snapshot *wS, int plnumber,
		 const Vector2d target, int panumber, PropOf &prop) const;

  // get the PropOf of "best" r-pass from player no. "plnumber" to player no. "panumber"
  bool get_rPass(bool we, const Snapshot *wS, int plnumber, int panumber, PropOf &prop) const;

  // calculate the Grade for a passpartner regarding his and the player's position
  // if we == true, look at our team, else at the other team
  Grade calculateGrade(bool we, const Snapshot *wS, const Vector2d &player,
		       const Vector2d &partner, PropOf &prop) const;

  // shout the PropOf
  void coutPropOf(const Snapshot *wS, PropOf prop) const;
};

#endif
