//////////////////////////////////////////////////////////////////////
// Mainz Rolling Brains RoboCup agents developed for the Simulator 
// League of the RoboCup 2000 competition
//
// Copyright (C) 1998, 1999, 2000 Team Members 
// of the Mainz Rolling Brains RoboCup Team 
//
// This program under the terms of the GNU General Public License as
// published by the Free Software Foundation; version 2 of the
// License. See the README file in the root directory of our archive.
//////////////////////////////////////////////////////////////////////

//////////////////////////////////////////////////////////////////////
//
// MRB player module PositionModule
//
// $Id: <add a dollar character here>
// $manager:
//
//////////////////////////////////////////////////////////////////////


#ifndef __POSITIONMODULE_H__
#define __POSITIONMODULE_H__

#include "grade.h"

#include "worldmodel.h"
#include "tacticalmodule.h"
#include "audiomgr.h"
#include "AdvSkills.h"
#include "goalkickeval.h"
#include "passeval.h"
#include "drawMacros.h"
#include "Config.h"

// Trace macro: insert your own short name for your module
// uncomment the line you want.
#undef POSTRACE
//#define POSTRACE(output) cout << wS->time << " PositionModule: " << output << endl

#undef POSTRACESHORT
// use this line if you want prompting of players decision process
//#define POSTRACESHORT(output) cout << "POS: " << output << endl; 

// use this if you want some modern art in the replayer 
//#define POSDRAWTRACE(output) cout << wS->time << "% " << output << endl

#ifndef POSTRACE
#define POSTRACE(output)
#else
#warning POSTRACE enabled
#endif
#ifndef POSTRACESHORT
#define POSTRACESHORT(output)
#else
#warning POSTRACESHORT enabled
#endif
#ifndef POSDRAWTRACE
#define POSDRAWTRACE(output)
#else
#warning POSDRAWTRACE enabled
#endif

//////////////////////////////////////////////////////////////////////
//
// definitions of constants
//
//////////////////////////////////////////////////////////////////////

const double COVER_DISTANCE = 1.5*VISIBLE_DISTANCE; // in this distance an other player is covered

//////////////////////////////////////////////////////////////////////
//
// enum PosBallControlState
//
//////////////////////////////////////////////////////////////////////

enum PosBallControlState
{
  POS_BC_WE,      // we have the ball
  POS_BC_THEY      // they have the ball
};


//Forward declaration
class PositionModuleInfo;


class PositionModule
{
public:
  /** constructor of module PositionModule */
  PositionModule(bool isGoalie, const PassEval *_passEval, 
		 const GoalKickEval *_goalKickEval);

  ~PositionModule();
  
  /** feed information about new world and tactics into module.
   */
  bool feedbackWorld(const Snapshot* wS, TacticalSnapshot* tacticalSnap);

  /** Feed audio information into module. This method is called for
   * each audio message. audioMsg.getSource() is player No or 0 for coach
   * Audiomsg audioInfo is acting like a stream.
   * usage: audioMsg >> auInfo (where PositionModuleInfo auInfo holds the information
   * for module PositionModule in other players)
   */
  bool feedbackAudio(AudioMsg &audioMsg);

  /** Do the evaluation for the next action
   * For definition of Grade: see grade.h
   */
  Grade evaluate(AdvSkills* skills,
                 const Snapshot* wS, TacticalSnapshot* tS);

  /** Perform the action for this module */
  bool act(AdvSkills* skills,
           const Snapshot* wS, TacticalSnapshot* tS);

  /** description of stream AudioMsg: see above. Here you should implement:
   * audioMsg << PositionModuleInfo;
   */
  bool getSayMsg(AudioMsg &audioMsg);

private:
  bool isGoalie;
  // The availableStamina depends on the stamina situation of the player and the 
  // desired priority.
  // stamina > 3100: [0][*]
  // stamina > 2800: [1][*]
  // stamina > 2500: [2][*]
  // else availableStamina = 0
  double posStamina[3][5];
  PosBallControlState posBallControl;
  // Where do I want to go today?!
  Vector2d posGoal;
  // save old posGoal
  Vector2d oldPosGoal;
  // middle of our goal (just for the goalie)
  Vector2d middleOfGoal;
  // to store the messages that shell be send
  PositionModuleInfo *sendInfo;
  // own offside lines
  double ourOffsideLine;
  double otherOffsideLine;
  // otherTeam maximal X position (without goalie)
  double otherTeamMaxX;
  // otherTeam minimal X position (with goalie)
  double otherTeamMinX;
  WorldPlayer me;
  double goToPrecision;
  int priority;
  int nearestFriendNumber;
  int internalTime;
  // the agents modules 
  const PassEval     *passEval;
  const GoalKickEval *goalKickEval;
  // maximal positions
  double myMaxX, myMaxY, myMinX, myMinY;
  int myTactNo;
  int nearestOther;
  // How many points are evaluated.
  int xPoints;
  int yPoints;
  // grid of evaluated positions
  Vector2d **evalGrid;

  // Parameter fr den evalGrid
  int oldX, oldY;
  int lastEvalTime;
  // Evaluation of a position, in the moment simply the minimum
  // over the distances to the lines, our/otherNearestPlayer
  double evalPos(const Snapshot *wS, const Vector2d &evalPos);

  // Am I a possible pass receiver
  bool amIPassReceiver(const Snapshot* wS,const TacticalSnapshot* tS);
  // Am I a pass receiver after the next pass
  bool amI2LevelPassReceiver(const Snapshot* wS,const TacticalSnapshot* tS);
  // Calculate (depending on own tactical role) max. and min. positions
  void calculateMaximalPositions(const Snapshot* wS,
				 const TacticalSnapshot* tS);
  // adapt calculated posGoal to own tactical role,
  // i.e. do not leave relative position to our players
  void truncatePosition();
  // Is this player free ?
  bool otherPlayerFree(const Snapshot *wS,int num, double distance = COVER_DISTANCE);
  bool ourPlayerFree(const Snapshot *wS,int num, double distance = COVER_DISTANCE);
  void respectStandardSituations(const Snapshot* wS,
				 const TacticalSnapshot* tS);
  // not very sophisticate method to avoid collisions with other players
  void avoidPlayerCollision(const Snapshot *wS);
  // check, if player is in my area
  // i.e. in minX-maxX and minY-maxY range
  bool playerInMyArea(const WorldPlayer pl);
  // check, if player is in my sector
  // i.e. behind me and in minY-maxY range
  bool playerInMySector(const WorldPlayer pl);
  // find free position 
  Vector2d calculateFreePosition(const Snapshot *wS,
				 double minX, double maxX, double minY, double maxY);
  Vector2d calculateCoverPosFromPlayerNo(const Snapshot *wS, int plNo);
}; // class PositionModule

//////////////////////////////////////////////////////////////////////
// Class to keep information for modules of other agents.
//////////////////////////////////////////////////////////////////////

class PositionModuleInfo
{
public:

  /** constructor of Info class for module PositionModule.
   * Here you should implement 
   * valid = _valid;
   */
  PositionModuleInfo(bool _valid);
  ~PositionModuleInfo();

  /** flag to determine if instance of PositionModuleInfo is valid. */
  bool valid;
}; // class PositionModuleInfo

/** put information to AudioMsg-Stream.
 * Here you should implement
 * aumsg << info.valid;
 * if (info.valid) aumsg << // further data structures to write ;
 * return aumsg;
 */
class AudioMsg &operator<< (AudioMsg &aumsg, const PositionModuleInfo &info);

/** fill with information from AudioMsg-Stream.
 * Here you should implement
 * aumsg >> info.valid;
 * if (info.valid) aumsg >> // further data structures to read ;
 * return aumsg;
 */
class AudioMsg &operator>> (AudioMsg &aumsg, PositionModuleInfo &info);

#endif

