//////////////////////////////////////////////////////////////////////
// Mainz Rolling Brains RoboCup agents developed for the Simulator 
// League of the RoboCup 2000 competition
//
// Copyright (C) 1998, 1999, 2000 Team Members 
// of the Mainz Rolling Brains RoboCup Team 
//
// This program under the terms of the GNU General Public License as
// published by the Free Software Foundation; version 2 of the
// License. See the README file in the root directory of our archive.
//////////////////////////////////////////////////////////////////////

//////////////////////////////////////////////////////////////////////
//
// MRB player module StandardSituationModule
//
// $Id: <add a dollar character here>
// $manager: gauer$
//
//////////////////////////////////////////////////////////////////////


// better change this to upper case:
#ifndef __STANDARDSITUATIONMODULE_H__
#define __STANDARDSITUATIONMODULE_H__

#include "grade.h"
#include "drawMacros.h"

#include "worldmodel.h"
#include "tacticalmodule.h"
#include "audiomgr.h"
#include "AdvSkills.h"

#include "passmodule.h"
#include "goalkickmodule.h"


const int GOALIE_MOVE_POSITIONS = 15;

const double eps = 1e-9;

#undef DECISIONTRACE
#undef OTHERTRACE
#undef CMDTRACE
#undef ERRORTRACE
#undef PLANTRACE
#undef DRAWTRACE
#undef PUSHTRACE

// use this line if you want prompting of player's decision process
// #define DECISIONTRACE(out) cout << worldSnap->time << " STDSIT[" << worldSnap->my.number << "]: " << out << endl

// use this line if you want prompting text not referring to the player's 
// decision process
// #define OTHERTRACE(out) cout << worldSnap->time << " STDSIT[" << worldSnap->my.number << "]: " << out << endl

// use this line if you want prompting complex texts (e.g. in loops)
// #define CMDTRACE(cmd) cmd

// use this line if you want prompting errors 
// #define ERRORTRACE(out) cerr <<  worldSnap->time << " STDSIT[" << worldSnap->my.number << "]: " << out << endl

// use this if you want prompting plan-messages
// #define PLANTRACE(cmd) cout << worldSnap->time << " STDSIT[" << worldSnap->my.number << "]: " << cmd << endl 

// use this if you want some modern art in the replayer 
// #define DRAWTRACE(output) cout << worldSnap->time << "% " << output << endl

// use this line if you want to prompt pushed players in calcGoalieMoveAndKick
// #define PUSHTRACE(cmd) cmd 



// now define all non-defined TRACE-constructs and prompt warnings for all
// defined ones

#ifndef DECISIONTRACE
#define DECISIONTRACE(cmd)
#else
#warning DECISIONTRACE enabled
#endif

#ifndef OTHERTRACE
#define OTHERTRACE(cmd)
#else 
#warning OTHERTRACE enabled
#endif

#ifndef CMDTRACE
#define CMDTRACE(cmd) 
#else 
#warning CMDTRACE enabled
#endif

#ifndef ERRORTRACE
#define ERRORTRACE(cmd) 
#else
#warning ERRORTRACE enabled
#endif

#ifndef PLANTRACE
#define PLANTRACE(cmd)
#else
#warning PLANTRACE enabled
#endif

#ifndef DRAWTRACE
#define DRAWTRACE(output)
#else
#warning DRAWTRACE enabled
#endif

#ifndef PUSHTRACE
#define PUSHTRACE(output)
#else
#warning PUSHTRACE enabled
#endif


// our memory for a short time after a standardsituation
class AfterStandardSituationPlan
{
public:
  AfterStandardSituationPlan() : 
    thePlan (PLAN_NOTHING),
    maxSuperKickTime (10) {};

  enum plan {
    PLAN_SUPERKICK,
    PLAN_RUN_TO_POSITION,
    PLAN_NO_ACTION,
    PLAN_TURNING,
    PLAN_NOTHING    
  };
  
  // does superkick. execution will be interrupted if not ready after "maxSuperKickTime" cycles
  void setSuperkick(AdvSkills::ExtKickCommand kickCommand_)
    { thePlan = PLAN_SUPERKICK; kickCommand = kickCommand_; 
      time = maxSuperKickTime; };

  // player will run to "pos_" using "stamina_" stamina per cycle
  void setRunToPosition(Vector2d pos_, double stamina_, int time_)
    { thePlan = PLAN_RUN_TO_POSITION; pos = pos_; stamina = stamina_; 
      time = time_; };

  // player will do nothing for "time_" cycles
  void setNoAction(int time_)            
    { thePlan = PLAN_NO_ACTION; time = time_; };

  // player will turn to the (absolute) direction "dir_" and wait there 
  // for "time_" cycles
  void setTurning(Angle dir_, int time_) 
    { thePlan = PLAN_TURNING; dir = dir_; time = time_; };

  // player has no plan
  void setNothing()                      
    { thePlan = PLAN_NOTHING; };

  // is there any plan to be done ?
  bool planExists()                      
    { return (thePlan != PLAN_NOTHING) && (time > 0); };

  // executes the plan, returns false if it fails 
  bool executePlan(AdvSkills *skills, const Snapshot *worlSnap);

  // what plan will be done ?
  plan getPlan()                         
    { return thePlan; }

private:


  plan thePlan;
  Angle dir;
  int time;
  Vector2d pos;
  double stamina;
  int maxSuperKickTime;

  AdvSkills::ExtKickCommand kickCommand;
};


class StandardSituationModuleInfo;



class StandardSituationModule
{
public:
  /** constructor of module StandardSituationModule */
  StandardSituationModule( bool isGoalie,  
			   const PassEval *_passEval, 
			   const GoalKickEval *_goalKickEval);

  ~StandardSituationModule();
  
  /** feed information about new world and tactics into module.
   */
  bool feedbackWorld(const Snapshot* worldSnap, TacticalSnapshot* tacticalSnap);

  /** Feed audio information into module. This method is called for
   * each audio message. audioMsg.getSource() is player No or 0 for coach
   * Audiomsg audioInfo is acting like a stream.
   * usage: audioMsg >> auInfo (where StandardSituationModuleInfo auInfo 
   * holds the information
   * for module StandardSituationModule in other players)
   */
  bool feedbackAudio(AudioMsg &audioMsg);

  /** Do the evaluation for the next action
   * For definition of Grade: see grade.h
   */
  Grade evaluate(AdvSkills* skills,
                 const Snapshot* worldSnap, TacticalSnapshot* tacticalSnap);

  /** Perform the action for this module */
  bool act(AdvSkills* skills,
           const Snapshot* worldSnap, TacticalSnapshot* tacticalSnap);

  /** description of stream AudioMsg: see above. Here you should implement:
   * audioMsg << StandardSituationModuleInfo;
   */
  bool getSayMsg(AudioMsg &audioMsg);


  bool sendSpecialSituation() { return isSpecialSituation; };

private:
  bool weAreKickingOff;
  bool iAmGoalie; 
  bool isSpecialSituation;
  int maxAge;

  int lastTime_PLAY_ON;

  // needed for KickOfThem and KickOfUs
  // to avoid multiple calculation of this information
  int firstPlayer, secondPlayer;
  Vector2d targetPos1, targetPos2;    // the targets where players should walk to:
                                      // firstPlayer will walk to targetPos1, 
                                      // secondPlayer will walk to targetPos2. 
  bool needSecond;                    // Are two players needed for standardsituation ?


  int waitingTime;                    // time for the first player to wait for the second
                                      // (only for active standardsituations where we 
                                      // have the ball)

  double minWallDist;                 // if the ball is closer to our goal than
                                      // do not build a wall
  double maxWallDist;                 // if the ball is closer to our goal than
                                      // build a wall out of two players

  double maxSecondDist;               // max way the second player can have to be needed

  int standardWaitingTime;            // the time we want to wait before doing a standard 
                                      // situation by default
  int waitingTimeAfterGoalieMove;     // the time the goalie waits after a move

  double wallStaminaFactor;           // determines how fast the players are running
  double passiveStaminaFactor;        // when building walls or doing other dash-commands
  double goalieStaminaFactor;         // players should dash (STAMINA_INC_MAX) * ...Factor
  double activeStaminaFactor;         // if they have enough stamina

  double lookAroundAngle;             // angle for lookAround()
  double ballAvoidDist;               // the distance to the ball the players will have in
                                      // standardsituations if they are not responsible
  
  // for communication
  StandardSituationModuleInfo *sendInfo;
  
  // for doing goalie's kick off
  int goalieMovePositions;
  Vector2d goalieMovePos[GOALIE_MOVE_POSITIONS];
  int goalieMovesInCurrentSituation;  // the number of move commands the goalie
                                      // has sent in the current standardsituation
  bool goalieDoesFreeKick;
  bool goalieMayMove;                 // goalie may move in free kick-situation 
                                      // (should be true after a catch-command)
  bool goalieShouldMoveAgain;         // something wrong with the last move command

  bool nextCycleDoNothing;            // next time the goalie does not do anything
                                      // also the evaluation will be bad
                                      // used for the goalie's free-kick only

  bool sendSomething;

  AfterStandardSituationPlan myPlan;  // my plan for the time after standardsituations

  // the agents modules 
  const GoalKickEval *goalKickEval;
  const PassEval     *passEval;


  enum turn_action {TURN_TO_LEFT_SIDE, TURN_TO_RIGHT_SIDE};
  turn_action nextTurnAction;

  // the positions before the kick-off, different systems
  Vector2d kickOffPosWeKickOff_442[11];       
  Vector2d kickOffPosTheyKickOff_442[11];
  Vector2d kickOffPosWeKickOff_343[11];       
  Vector2d kickOffPosTheyKickOff_343[11];
  Vector2d kickOffPosWeKickOff_352[11];       
  Vector2d kickOffPosTheyKickOff_352[11];
  
  // the number and type of players for stdsits
  struct playerTypeCount 
  { int defenders; 
    int midfield;     
    int offenders; } players[2][3][3]; 
  // 2 teams 3 parts of the field 3 tactical systems

  // does what it says
  void readAgent_conf( void );

  // calcs the number of each player type to do a stdsit 
  void calcDesiredPlayerTypes(const Snapshot *ws, const TacticalSnapshot *ts,
			      const Vector2d &target, int &noDef, 
			      int &noMid, int &noOff);

  // pushes one player to a distance of 9.15 around the given "pos".
  void pushPlayer(WorldPlayer &player, Vector2d pos);

  // calculates the positions where the players start before a kick-off
  Vector2d calcKickOffTarget(int playerNumber, const Snapshot *worldSnap, 
			     TacticalSnapshot *tacticalSnap);

  // calculates where the goalie should move and where to kick the ball 
  void calcGoalieMoveAndKick(const Snapshot *worldSnap, 
				  TacticalSnapshot *tacticalSnap,
				  Vector2d &moveTarget,
				  Vector2d &kickTarget,
			          bool goalieMayMoveOnceMore);

  // walks to target using stamina without entering the
  // circle defined by center point m and radius r
  // true if still walking
  // false if target reached and turned to ball
  // if false no action is send (even not skills->noAction()!)
  bool goingTo(Vector2d &target, const double stamina,
				      AdvSkills *skills,
				      const Snapshot *worldSnap, 
				      const Vector2d *m = NULL,
				      const double r = 0);

  // walks to target but walks backwards while looking to the ball
  // don't forget walking backwards needs twice as much stamina than
  // walking forwards
  // true while walking
  // false when target reached
  bool goingBackwardsTo(Vector2d &target, const double stamina,
				      AdvSkills *skills,
				      const Snapshot *worldSnap, 
				      const Vector2d *m = NULL,
				      const double r = 0);

  // does all things neccesary for kick-off situations (US+THEM)
  bool doingKickOff(AdvSkills *skills,
		    const Snapshot *worldSnap,
		    TacticalSnapshot *tacticalSnap);

  // does all things neccesary for Goalie's free-kick situations (THEM)
  bool doingGoalieFreeKickThem(AdvSkills *skills,
			       const Snapshot *worldSnap);

  // does all things neccesary for Goalie's free-kick situations (THEM)
  bool doingFieldPlayerFreeKickThem(AdvSkills *skills,
				    const Snapshot *worldSnap,
				    int firstPlayer, int secondPlayer,
				    bool needSecond, Vector2d &targetPos1,
				    Vector2d &targetPos2);

  bool doingKickInCornerKickThem(AdvSkills *skills,
				 const Snapshot *worldSnap,
				 int firstPlayer, int secondPlayer,
				 bool needSecond, Vector2d &targetPos1,
				 Vector2d &targetPos2);

  bool doingFreeKickUs(AdvSkills *skills,
		       const Snapshot *worldSnap,
		       int firstPlayer, int secondPlayer,
		       bool needSecond,
		       Vector2d &targetPos1, 
		       Vector2d &targetPos2);

  bool doingKickInUs(AdvSkills *skills,
		     const Snapshot *worldSnap,
		     int firstPlayer, int secondPlayer,
		     bool needSecond,
		     Vector2d &targetPos1, 
		     Vector2d &targetPos2);

  bool doingCornerKickUs(AdvSkills *skills,
			 const Snapshot *worldSnap,
			 int firstPlayer, int secondPlayer,
			 bool needSecond,
			 Vector2d &targetPos1, 
			 Vector2d &targetPos2);

  // calc the intersection of the bisector of the angle between the lines 
  // from the ball to the edges of the goal and
  // the line between the edges of the goal
  Vector2d calcPossibleGoalPoint(const Snapshot *worldSnap);

  // calc closest two players to target
  // the players are  taken out of the
  // given number of player types
  // depends on our tactical system
  // and the field part target is in
  // result can be UNKNOWN
  // goalie is never a result
  void calcClosestDesiredPlayersTo(const Vector2d &target, 
				   const Snapshot *worldSnap,
				   const TacticalSnapshot *ts,
				   int &firstPlayer,
				   int &secondPlayer);
  void calcClosestPlayerTo(const Vector2d &target, const int team[], const int count,
			   const Snapshot *ws, int &player);

  void calcClosestTwoPlayersTo(const Vector2d &target, const int team[], const int count,
			       const Snapshot *ws, int &player1, int &player2);

  void calcPlayersAndPos4FreeKickThem(const Snapshot *worldSnap, 
				      const TacticalSnapshot *ts,
				      int &firstPlayer, int &secondPlayer,
				      bool &needSecond, Vector2d &targetPos1,
				      Vector2d &targetPos2) ;

  void calcPlayersAndPos4FreeKickUs(const Snapshot *worldSnap, 
				    const TacticalSnapshot *ts,
				    int &firstPlayer, int &secondPlayer,
				    bool &needSecond, Vector2d &targetPos1,
				    Vector2d &targetPos2) ;

  void calcPlayersAndPos4KickInThem(const Snapshot *worldSnap, 
				    const TacticalSnapshot *ts,
				    int &firstPlayer, int &secondPlayer,
				    bool &needSecond, Vector2d &targetPos1,
				    Vector2d &targetPos2) ;

  void calcPlayersAndPos4KickInUs(const Snapshot *worldSnap, 
				  const TacticalSnapshot *ts,
				  int &firstPlayer, int &secondPlayer,
				  bool &needSecond, Vector2d &targetPos1,
				  Vector2d &targetPos2) ;

  void calcPlayersAndPos4CornerKickThem(const Snapshot *worldSnap, 
					const TacticalSnapshot *ts,
					int &firstPlayer, int &secondPlayer,
					bool &needSecond, Vector2d &targetPos1,
					Vector2d &targetPos2);

  void calcPos4CornerKickUs(const Snapshot *worldSnap, 
			    Vector2d &targetPos1,
			    Vector2d &targetPos2);

  void calcPlayersAndPos4CornerKickUs(const Snapshot *worldSnap, 
				      const TacticalSnapshot *ts,
				      int &firstPlayer,
				      int &secondPlayer,
				      bool &needSecond,
				      Vector2d &targetPos1,
				      Vector2d &targetPos2) ;

  // calcs the point where to go next if "to" is not reachebale without entering 
  // the circle
  // defined by "m" and "r"
  // the result will be returned in "to"
  void calcWayTo(const Vector2d &from, Vector2d &to, const double stamina,
		 const Vector2d &middle, const double &radius, 
		 const Snapshot *worldSnap);

  // if there are players in my way from my.pos to target
  // target is changed to a position to walk to instead
  // m is the middle of a circle with radius r the players can not walk in
  // target and *m have to be in absolute coordinates
  void walkAroundPlayers(const Snapshot *worldSnap,  Vector2d &target,
			 const double stamina = STAMINA_INC_MAX,
			 const Vector2d *m = NULL, const double r = 0);

  // turns the players head to the left and the right of the ball
  void lookAround(AdvSkills *skills,
		  const Snapshot *ws);

  // check if players of the other team are in a cricle with radius r around pos
  bool posIsFree(const Vector2d &pos, const Snapshot *worldSnap, const double r = 3.);

  // goalies goal-kick-situation
  bool doingGoalieFreeKickUs(AdvSkills *skills,
			     const Snapshot *worldSnap,
			     TacticalSnapshot *tacticalSnap);

  // if all passes are bad, this should be used to make an emergency kick-off.
  // routine scans our and other player in the left and the right third and returns an
  // "optimal" move-position
  Vector2d calcEmergencyMovePosition(const Snapshot *worldSnap);


  // calculates where to pass the ball when we have kick-off
  Vector2d calcKickOffKickTarget(const Snapshot *worldSnap, 
				 TacticalSnapshot *tacticalSnap,
				 int &passReceiver);

  // changes the players so that the ways to walk are shorter
  void calcShortWays(const Snapshot *worldSnap, 
		     int &firstPlayer, int &secondPlayer,  
		     const Vector2d &targetPos1, const Vector2d &targetPos2);

  // gets our tactical system and returns an int (442, 343, 352, ...)
  int getOurTacticalSystem(const TacticalSnapshot *tacticalSnap);  

  // calculates the intersection of two evens 
  Vector2d calcIntersection(Vector2d A, Vector2d B, 
			    Vector2d C, Vector2d D);

  // calculates the shortest distance from A to B around the circle (M,r)
  // worldsnap is only needed for drawing nice pics
  // note: this only works if AB and (M,r) intersect
  double calcDistAroundCircle(Vector2d A, Vector2d B, Vector2d M, 
			      double r, const Snapshot *worldSnap);

  // does player have to avoid the ball ?
  bool mustAvoidBall(const Snapshot *worldSnap);

  // avoids the ball: if player is near to the ball, run away
  void avoidBall(AdvSkills *skills, const Snapshot *worldSnap,
		 TacticalSnapshot *tacticalSnap);

  // does the evaluation for PLAY_MODE_FREE_KICK_US
  Grade evalFreeKickUs(AdvSkills* skills,
		       const Snapshot* worldSnap, 
		       TacticalSnapshot* tacticalSnap);
  
}; // class StandardSituationModule







//////////////////////////////////////////////////////////////////////
// Class to keep information for modules of other agents.
//////////////////////////////////////////////////////////////////////

class StandardSituationModuleInfo
{
public:

  /** constructor of Info class for module StandardSituationModule.
   */
  StandardSituationModuleInfo();
  ~StandardSituationModuleInfo();

  //double value1;
  //bool value2;

}; // class StandardSituationModuleInfo






/** put information to AudioMsg-Stream.
 * Here you should implement
 * aumsg << info.valid;
 * if (info.valid) aumsg << // further data structures to write ;
 * return aumsg;
 */
class AudioMsg &operator<< (AudioMsg &aumsg, const StandardSituationModuleInfo &info);

/** fill with information from AudioMsg-Stream.
 * Here you should implement
 * aumsg >> info.valid;
 * if (info.valid) aumsg >> // further data structures to read ;
 * return aumsg;
 */
class AudioMsg &operator>> (AudioMsg &aumsg, StandardSituationModuleInfo &info);

#endif

