//////////////////////////////////////////////////////////////////////
// Mainz Rolling Brains RoboCup agents developed for the Simulator 
// League of the RoboCup 2000 competition
//
// Copyright (C) 1998, 1999, 2000 Team Members 
// of the Mainz Rolling Brains RoboCup Team 
//
// This program under the terms of the GNU General Public License as
// published by the Free Software Foundation; version 2 of the
// License. See the README file in the root directory of our archive.
//////////////////////////////////////////////////////////////////////

// -*- c++ -*- ///////////////////////////////////
//
// $Id: playerlogReader.C,v 1.10 2000/10/19 15:27:57 goetz Exp $
//
// playerlogReader.C
//
//////////////////////////////////////////////////

#include "playerlogReader.h"

const int MAX_LINE_CHAR = 4096;

PlayerlogReader::PlayerlogReader(const char* filename) {

  // initialize

  for (int i=0; i<MAX_OUTPUT_LINE; i++) {
    lines[i] = NULL;
  }
  maxLine = -1;

  // read logfile

  FILE* playerlogfile = fopen (filename, "r");
  
  if (playerlogfile) {
  
    char line [MAX_LINE_CHAR];
    
    while (fscanf (playerlogfile, "%[^\n]\n", line) != EOF) {
      
      int time;
      int scanok = sscanf (line, "%d", &time);
      if ((scanok == 1) && (time >= 0)) {
	
	maxLine++;
	if (maxLine < MAX_OUTPUT_LINE) {
	  lines[maxLine] = new char [strlen(line)+1];
	  strcpy(lines[maxLine], line);
	  
	} else cerr << "PlayerLog: MAX_OUTPUT_LINE = " << MAX_OUTPUT_LINE 
		    << " too small to read whole file !" << endl;
	
      } // if valid line	
      
    } // while !EOF

    // clean up afterwards:
    fclose (playerlogfile);
    
    if (maxLine < MAX_OUTPUT_LINE) 
      cout << "read " << maxLine+1 << " valid lines from " << filename << endl;
    
  } // file open ok
  
  else {
    cerr << "PlayerLog: file '" << filename << "' does not exist!"<< endl;
  }
  
}


PlayerlogReader::~PlayerlogReader() {

  for (int i=0; i<MAX_OUTPUT_LINE; i++) {
    delete(lines[i]);
    lines[i] = NULL;
  }
  
}


void
PlayerlogReader::printOutput(int time) {

  for (int i=0; i<maxLine; i++) {
 
    int linetime;
    char c;
    int scanok = sscanf (lines[i], "%d%c", &linetime, &c);
    if ((scanok == 2) && (time == linetime) &&
	( ( c == '>' ) || ( c == ' ' ) ) ) {
      cout << lines[i] << endl;
    }
    
  }

}

void 
PlayerlogReader::printCommands(int time) {

  for (int i=0; i<maxLine; i++) {
 
    int linetime;
    char c;
    int scanok = sscanf (lines[i], "%d%c", &linetime, &c);
    if ((scanok == 2) && (c == '>') &&
	(linetime >= time-NUM_OLD_COMMANDS) &&
	(linetime < time) ) {
      cout << lines[i] << endl;
    }
    
  }

}

bool eatChar( const char* cmd, int &readpos, const char c)
{
  int len= strlen(cmd);
  while (cmd[readpos] != c && readpos < len) readpos++;
  readpos++;
  if (readpos >= len) return false;
  else return true;
}

void parseCmd( const char* cmd, DrawablesList *drawables)    
{
  int readpos= 0;
  while (int len= strlen(&cmd[readpos]) != 0 ) {
    double x1, y1, x2, y2, rad;
    char c1(0), c2(0), s[1024];
    if (!eatChar(cmd, readpos, '(')) return;
    switch (cmd[readpos]) {
    case 'c':
       if (sscanf(&cmd[readpos+1], " ( %lf , %lf ) %lf %c %c", 
       		 &x1, &y1, &rad, &c1, &c2) >= 3 ) {
	 drawables->add(new DCircle(x1, y1, rad, 
 				   c1 == 'g' || c2 == 'g',
 				   c1 == 'l' || c2 == 'l' ));
 	if (!eatChar(cmd, readpos, ')') ||
 	    !eatChar(cmd, readpos, ')')) return;
       }
      break;
    case 'l':
      if (sscanf(&cmd[readpos+1], " ( %lf , %lf ) ( %lf , %lf ) %c",
		 &x1, &y1, &x2, &y2, &c1) >= 4) {
	drawables->add(new DLine(x1, y1, x2, y2, c1 == 'l'));     
	if (!eatChar(cmd, readpos, ')') ||
	    !eatChar(cmd, readpos, ')') ||
	    !eatChar(cmd, readpos, ')')) return;
      }
      break;
    case 'p':
      if (sscanf(&cmd[readpos+1], " ( %lf , %lf ) %c", &x1, &y1, &c1) >= 2) {
	drawables->add(new DPosition(x1, y1, c1 == 'l'));
	if (!eatChar(cmd, readpos, ')') ||
	    !eatChar(cmd, readpos, ')')) return;
      }
      break;
    case 't':
      if (sscanf(&cmd[readpos+1], " ( %lf , %lf ) %[^)]", &x1, &y1, s) == 3) {
	drawables->add(new DText(x1, y1, s));
	if (!eatChar(cmd, readpos, ')') ||
	    !eatChar(cmd, readpos, ')')) return;
      }
      break;
    case 'n':
      drawables->clear();
      break;
    case 'u':
      drawables->remove();
      break;
    default:
      return;
    }
  }
}

void
PlayerlogReader::drawOutput(int time, DrawablesList *drawables) {

  for (int i=0; i<maxLine; i++) {
 
    int linetime;
    char c;
    int scanok = sscanf (lines[i], "%d%c", &linetime, &c);
    if ((scanok == 2) && (c == '%') && (time == linetime)) {
      parseCmd(lines[i], drawables);
    }
  }
}
