//////////////////////////////////////////////////////////////////////
// Mainz Rolling Brains RoboCup agents developed for the Simulator 
// League of the RoboCup 2000 competition
//
// Copyright (C) 1998, 1999, 2000 Team Members 
// of the Mainz Rolling Brains RoboCup Team 
//
// This program under the terms of the GNU General Public License as
// published by the Free Software Foundation; version 2 of the
// License. See the README file in the root directory of our archive.
//////////////////////////////////////////////////////////////////////

//// -*- c++ -*- ///////////////////////////////////////////////////
//
// $Id: replayer.h,v 1.14 2000/10/19 15:27:57 goetz Exp $
// $Manager: eri$
//
//  replayer.h - replay snaps.log and player.log files
//
/////////////////////////////////////////////////////////////////////



#ifndef _REPLAYER_H_
#define _REPLAYER_H_

#include <unistd.h>

#include "worldmodel.h"
#include "worldmap_r.h"
#include "worldmap_r_two_snap.h"
#include "snaplogReader.h"
#include "serverlogReader.h"
#include "playerlogReader.h"

#include "drawables.h"

// constants
/////////////////////////////////////////////////////////////////////

const char*  PLAYER_LOG      = "player_"; // filenames
const char*  SNAP_LOG        = "snaps_";
const char*  EXTENSION       = ".log";

const double MAPWINDOW_SIZE  = 5.0;


// RePlayer class
/////////////////////////////////////////////////////////////////////



class RePlayer {

public:

  
  RePlayer(const char* _teamName, int _playerNumber);
  
  ~RePlayer();        
  
  void replay();       // user interactive replaying
  void evaluate();     // any numerical evaluation, f.ex. calculation of errors
  
private:

  WorldMap_r *mapwindow;
  WorldMap_r_two_snap *mapwindow_two;

  DrawablesList *drawables;

  SnaplogReader*   snaplogReader;   // reader for snapshots saved to file
  SnaplogReader*   coachlogReader;  // reader for snapshots from coach
  PlayerlogReader* playerlogReader; // reader for player's ascii output file

  /*
    // no longer used, coach is more exact and has more info.
    // and serverlogreader would have to be adapted to all sserver-changes !
  ServerlogReader* serverlogReader; // reader for sserver logfiles
  */

  int   maxTime;  // maximum recorded timestamp
  bool  twoSnap;  // flag to determine display mode
  char* team;     // player's team name
  int   number;   // player's number

};


#endif

