//////////////////////////////////////////////////////////////////////
// Mainz Rolling Brains RoboCup agents developed for the Simulator 
// League of the RoboCup 2000 competition
//
// Copyright (C) 1998, 1999, 2000 Team Members 
// of the Mainz Rolling Brains RoboCup Team 
//
// This program under the terms of the GNU General Public License as
// published by the Free Software Foundation; version 2 of the
// License. See the README file in the root directory of our archive.
//////////////////////////////////////////////////////////////////////

// -*- c++ -*- ///////////////////////////////////
//
// $Id: snaplogReader.C,v 1.11 2000/10/19 15:27:58 goetz Exp $
//
// snaplogReader.C
//
//////////////////////////////////////////////////

#include "snaplogReader.h"

SnaplogReader::SnaplogReader(const char* filename) {

  // initialize

  for (int i=0; i<MAX_TIME; i++) {
    index[i] = -1;
  }
  maxTime = -1;
  
  snap = new Snapshot();
  
  // read logfile
  
  in.open (filename, ios::in);

  if(! in.is_open()) {

    cerr << "SnaplogReader: file '" << filename << "' does not exist." << endl;
    maxTime = -1;

  } else {
    
    int  i=0;
    int  lastTime = -1;
    bool readOk = true;
  
    while (!in.eof() && readOk && (lastTime < MAX_TIME-1) ) {
      
      readOk &= snap->binaryReadFromStream(in);
      
      if ( readOk ) {
	
	if (snap->time != lastTime) {
	  if (snap->time > lastTime) {
	    index[snap->time] = i;
	    lastTime = snap->time;
	  } else {
	    cerr << "SnaplogReader: snaps were not saved in ascending order " 
		 << endl;
	    in.close();
	  }
	}
	
	i++;
	
      }
      
    } //while reading

    if (lastTime == MAX_TIME-1) {
      cerr << "SnaplogReader (" << filename << ") : MAX_TIME=" << MAX_TIME 
	   << " too small to read all Snapshots from file !" << endl;
      maxTime = lastTime;
    } else if (in) { // stream still ok, but some error was reached
      cerr << "SnaplogReader (" << filename 
	   << ") : some error occurred during reading" << endl;
      in.close();
    } else { // eof reached, reading finished
      cout << "read " << i << " snapshots from " << filename << endl;
      maxTime = lastTime;
    }
  }
  
}


SnaplogReader::~SnaplogReader() {

  delete (snap);
  in.close();

}


Snapshot*
SnaplogReader::getSnapshot(int time) {

  if ( ( index[time] >= 0 ) && ( time <= maxTime ) ) {
    in.clear();
    in.seekg(index[time] * (sizeof(Snapshot)+sizeof(int)+1), ios::beg);
    if (in.good()) {
      snap->binaryReadFromStream(in);
    }
    if (snap->time != time) {
      cerr << "SnaplogReader: fuck! snap->time does not match indexed time!" 
	   << endl;
    }
  }
  
  return snap;
  
}
