//////////////////////////////////////////////////////////////////////
// Mainz Rolling Brains RoboCup agents developed for the Simulator 
// League of the RoboCup 2000 competition
//
// Copyright (C) 1998, 1999, 2000 Team Members 
// of the Mainz Rolling Brains RoboCup Team 
//
// This program under the terms of the GNU General Public License as
// published by the Free Software Foundation; version 2 of the
// License. See the README file in the root directory of our archive.
//////////////////////////////////////////////////////////////////////

//////////////////////////////////////////////////////////////////////
//
// $Id: worldmap_r_two_snap.C,v 1.12 2000/10/19 15:27:58 goetz Exp $	
// $Manager: goetz$
// 
// 
//  worldmap_r_two_snap.C - World model display
// 
//  ( Christian Bauer <cebix> )
//    Goetz Schwandtner
//
//
//////////////////////////////////////////////////////////////////////

#include <stdio.h>
#include <string.h>

#include "worldmap_r_two_snap.h"

// change this when possible !!
#include "constants.h"
#include "coordinates.h"

// Internal constants

static char *num2str[] = {"? ", "1 ", "2 ", "3 ", "4 ", "5 ",
			 "6 ", "7 ", "8 ", "9 ", "10", "11"};

/* 
 * Constructor
 */

WorldMap_r_two_snap::WorldMap_r_two_snap(const char *windowtitle, double mag, double relmag) :
  WorldMap_r(windowtitle, mag, relmag)
{
}

/*
 *  Open map window
 */

bool WorldMap_r_two_snap::Open()
{
  // Open absolute map window and prepare X connection ...
  if (!WorldMap_r::Open())
    return false;
  
  // just in case something went terribly wrong ...
  if (!the_disp)
    return false;

  // Get colors
  XParseColor(the_disp, the_cmap, "rgb:a/0/0", &dred_color);
  XAllocColor(the_disp, the_cmap, &dred_color);
  dred = dred_color.pixel;
  XParseColor(the_disp, the_cmap, "rgb:0/6/0", &dgreen_color);
  XAllocColor(the_disp, the_cmap, &dgreen_color);
  dgreen = dgreen_color.pixel;
  XParseColor(the_disp, the_cmap, "rgb:6/6/d", &dblue_color);
  XAllocColor(the_disp, the_cmap, &dblue_color);
  dblue = dblue_color.pixel;
  XParseColor(the_disp, the_cmap, "rgb:a/a/0", &dyellow_color);
  XAllocColor(the_disp, the_cmap, &dyellow_color);
  dyellow = dyellow_color.pixel;
  
  // Wait until window is ready
  XSync(the_disp, false);
  return(true);
}

/*
 *  Draw snapshot
 */

void WorldMap_r_two_snap::Map(Snapshot *snap_primary, 
			      Snapshot *snap_secondary, 
			      int max_age)
{
  if (the_disp == NULL)
    return;
  
  if (snap_primary == NULL)
    return;
  
  
  // Draw absolute and relative WorldMap
  WorldMap_r::Map(snap_primary, max_age);


  // set player position for relative display (secondary)
  setownpos(snap_secondary->my);

  // Draw myself
  drawownplayer(snap_secondary->my);

  // Draw ball
  drawball(snap_secondary->ball);
	
  // Draw my team
  drawmyteam(snap_secondary);
	
  // Draw other team
  drawotherteam(snap_secondary);
  
  // Draw pixmap
  XClearWindow(the_disp, the_win);
  XClearWindow(the_disp, rel_win); 
  XFlush(the_disp);
}

void WorldMap_r_two_snap::setownpos(WorldPlayer ownplayer)
{
  sec_ownx = ownplayer.pos.getX();
  sec_owny = ownplayer.pos.getY();
}

void WorldMap_r_two_snap::drawownplayer(WorldPlayer ownplayer)
{
  XSetForeground(the_disp, map_gc, dblue);  
  XSetForeground(the_disp, rel_map_gc, dblue);
  
  // First absolute map:
  // My Coordinates 

  int x = WorldMap::mx(sec_ownx);
  int y = WorldMap::my(sec_owny);
    
  XDrawArc(the_disp, the_map, map_gc,		// Position
	   x - PSIZE, y - PSIZE,
	   PSIZE*2, PSIZE*2,
	   0, 360 * 64);

  XFillArc(the_disp, the_map, map_gc,		// Body direction
	   x - PSIZE, y - PSIZE,
	   PSIZE*2, PSIZE*2,
	   -int(RAD_TO_DEG(ownplayer.bodyDir + PI_4) * 64), 
	   90 * 64);
  XDrawString(the_disp, the_map, map_gc,	// Number
	      x - PSIZE, y + PSIZE + 8,
	      num2str[ownplayer.number], 2);

  // Then relative map:

  // My Coordinates 

  x = WorldMap_r::mx(0);
  y = WorldMap_r::my(0);
    
  XDrawArc(the_disp, rel_map, rel_map_gc,		// Position
	   x - PSIZE, y - PSIZE,
	   PSIZE*2, PSIZE*2,
	   0, 360 * 64);

  XFillArc(the_disp, rel_map, rel_map_gc,		// Body direction
	   x - PSIZE, y - PSIZE,
	   PSIZE*2, PSIZE*2,
	   -int(RAD_TO_DEG(ownplayer.bodyDir + PI_4) * 64), 
	   90 * 64);

  XDrawString(the_disp, rel_map, rel_map_gc,	// Number
	      x - PSIZE, y + PSIZE + 8,
	      num2str[ownplayer.number], 2);
}

void WorldMap_r_two_snap::drawball(WorldBall ball)
{

  XSetForeground(the_disp, rel_map_gc, dyellow);
  XSetForeground(the_disp, map_gc, dyellow);
  
  // First absolute map:
  
  int x = WorldMap::mx(ball.pos.getX());
  int y = WorldMap::my(ball.pos.getY());

  XDrawArc(the_disp, the_map, map_gc,	// Position
	   x - PSIZE, y - PSIZE,
	   PSIZE*2, PSIZE*2,
	   0, 360 * 64);

  XDrawLine(the_disp, the_map, map_gc,	// Velocity
	    x, y,
	    x + WorldMap::mv(ball.v.getX()), y + WorldMap::mv(ball.v.getY()));

  // Now relative map:
  
  x = WorldMap_r::mx(ball.pos.getX() - sec_ownx);
  y = WorldMap_r::my(ball.pos.getY() - sec_owny);
  
  XDrawArc(the_disp, rel_map, rel_map_gc,	// Position
	   x - PSIZE, y - PSIZE,
	   PSIZE*2, PSIZE*2,
	   0, 360 * 64);

  XDrawLine(the_disp, rel_map, rel_map_gc,	// Velocity
	    x, y,
	    x + WorldMap_r::mv(ball.v.getX()), y + WorldMap_r::mv(ball.v.getY()));

}

void WorldMap_r_two_snap::drawmyteam(Snapshot *snap)
{
  XSetForeground(the_disp, map_gc, dgreen);
  XSetForeground(the_disp, rel_map_gc, dgreen);
  for (int i=1; i<=11; i++) {
    if (i == snap->my.number)
      continue;
    drawplayer(snap->ourTeam(i));
  }
}

void WorldMap_r_two_snap::drawotherteam(Snapshot *snap)
{
  XSetForeground(the_disp, map_gc, dred);
  XSetForeground(the_disp, rel_map_gc, dred);
  for (int i=1; i<=11; i++) {
    drawplayer(snap->otherTeam(i));
  }
}

void WorldMap_r_two_snap::drawplayer(WorldPlayer player)
{
  // First absolute map

  int x = WorldMap::mx(player.pos.getX());
  int y = WorldMap::my(player.pos.getY());
  int num = player.number;

  XDrawArc(the_disp, the_map, map_gc,		// Position
	   x - PSIZE, y - PSIZE,
	   PSIZE*2, PSIZE*2,
	   0, 360 * 64);

  XFillArc(the_disp, the_map, map_gc,		// Body direction
	   x - PSIZE, y - PSIZE,
	   PSIZE*2, PSIZE*2,
	   -int(RAD_TO_DEG(player.bodyDir + PI_4) * 64), 
	   90 * 64);
  if (num <= 0) num = 0;
  XDrawString(the_disp, the_map, map_gc,	// Number
	      x - PSIZE, y + PSIZE + 8,
	      num2str[num], 2);

  // Now relative map

  x = WorldMap_r::mx(player.pos.getX() - sec_ownx);
  y = WorldMap_r::my(player.pos.getY() - sec_owny);

  XDrawArc(the_disp, rel_map, rel_map_gc,		// Position
	   x - PSIZE, y - PSIZE,
	   PSIZE*2, PSIZE*2,
	   0, 360 * 64);
  XFillArc(the_disp, rel_map, rel_map_gc,		// face direction
	   x - PSIZE, y - PSIZE,
	   PSIZE*2, PSIZE*2,
	   -int(RAD_TO_DEG(player.bodyDir + PI_4) * 64), 
	   90 * 64);
  if (num <= 0) num = 0;
  XDrawString(the_disp, rel_map, rel_map_gc,	// Number
	      x - PSIZE, y + PSIZE + 8,
	      num2str[num], 2);
}

