//////////////////////////////////////////////////////////////////////
// Mainz Rolling Brains RoboCup agents developed for the Simulator 
// League of the RoboCup 2000 competition
//
// Copyright (C) 1998, 1999, 2000 Team Members 
// of the Mainz Rolling Brains RoboCup Team 
//
// This program under the terms of the GNU General Public License as
// published by the Free Software Foundation; version 2 of the
// License. See the README file in the root directory of our archive.
//////////////////////////////////////////////////////////////////////

// -*- c++ -*- ///////////////////////////////////////////////////
//
// $Id: scanpos.C,v 1.3 2000/10/19 15:27:58 goetz Exp $
// $Manager: holgi$
//
//  scanpos.C
//
//  author: <holgi>
//
/////////////////////////////////////////////////////////////////////

#include "scanpos.h"

///////////////////////////////////////////////////////////////////////////////////////////
//
// functions to open and close the logfiles
//
///////////////////////////////////////////////////////////////////////////////////////////

/////////////
//
// int ScanPos::OpenFile(const char *logfilename)
//
//////////////

int ScanPos::OpenFile(const char *logfilename)
{
  if (logfileopen == true) {
    cerr << "error - there's already a logfile open"
	 << endl;
    return 0;
  }
  
  logfile.open(logfilename, ios::in); // assign file logfilename with logfile
                                      // for reading

  if (!logfile) {                        // file doesn't exists
    cerr << "error while opening file "
         << logfilename
         << endl;
    return 0;
  }
  
  char testen[4];                          // read header of file
  int i = 0;
  
  while (i <= 2) {                         // error while reading header
    if (!(logfile >> testen[i])) {
      cerr << "error while reading header - can't read header at position "
	   << i+1
           << endl;
      return 0;
    }
    i++;
  }

  if ((testen[0] != 'U') &&              // wrong header
      (testen[1] != 'L') && 
      (testen[2] != 'G')) {   
    cerr << "error while reading header - wrong header"
	 << endl;
    return 0;
  }
  
  if (!(logfile >> testen[3])) {          // reading version-number
    cerr << "error while reading version"
         << endl; 
    return 0;
  }

  logfileopen = true; // set flag to true, because:

  for (int x = 0; x < 11; x++) {
    char name[32];
    sprintf(name, "object%d.out", x+1);
    posobject[x].open(name, ios::out);
  }
  return 1;           // logfile successfully read
}

/////////////
//
// int ScanPos::CloseFile()
//
//////////////

int ScanPos::CloseFile()
{
  if (logfileopen == true) {
    logfile.close();
    for (int i=0; i < 11; i++) {
      posobject[i].close();
    }
    return 1;
  }
  else {
    cerr << "error - there's no logfile to close"
	 << endl;
    return 0;
  } 
}

///////////////////////////////////////////////////////////////////////////////////////////

//////////
//
// void ScanPos::writePos()
//
/////////

int ScanPos::writePos(bool side)
{
  showinfo_t show;
  int offset;

  if (readnextShow(show)) {
    if (side) {         // left
      offset = 0;
    }
    else {              // right
      offset = 11;
    }
    for (int i = 0; i < 11; i++) {
      if (show.pmode != PM_PlayOn)
	posobject[i] << endl;
      posobject[i] << show.pos[i+1+offset].x/SHOWINFO_SCALE
		   << " "
		   << show.pos[i+1+offset].y/SHOWINFO_SCALE
		   << endl;
    }
    return 1;
  }
  return 0;
}

//////////
//
// int ScanPos::readnextShow()
//
/////////////

int ScanPos::readnextShow(showinfo_t &show)
{
  int mode = 0;

  do {
    mode = readMode();  // read next mode
    if (mode == -1)     // error occured while reading mode
      return 0;
    else {
      switch (mode) {   // mode should be  0, 1, 2 or 3
      case 0:                          
	break;                            // no info, do nothing
      case 1:
	show = readShow();                // read showinfo_t to show

	// test, if show contains a "bad" playmode = PM_AfterGoal_Left/Right
	// in this case, set mode to -1 and continue with next entry
        if ((show.pmode == PM_AfterGoal_Left) || (show.pmode == PM_AfterGoal_Right)) 
	  mode = -1;

	break;
      case 2:
	msginfo_t msginfo = readMsg();    // read msginfo_t for nothing
	break;
      case 3:
	drawinfo_t drawinfo = readDraw(); // read drawinfo_t for nothing
	break;
      default:
	cerr << "error in logfile: "
	     << mode
	     << " is no valid mode"
	     << endl;
	return 0;
	break;
      }
    }
  }
  while (mode != 1); // while no showinfo_t found in logfile

  return 1; // successfully read showinfo_t from logfile
}

///////////////////////////////////////////////////////////////////////////////////////////
//
// helping-functions for reading the logfile
//
///////////////////////////////////////////////////////////////////////////////////////////

///////////////
//
// int ScanPos::readMode()
//
///////////////

int ScanPos::readMode()
{
  char dummy;
  short mode = 0;
  
  if (!(logfile >> dummy))  // try do read character
    return -1;
  else {
    mode = short(dummy);
    return int(mode);       // and return the int-value
  }
}

/////////////
//
// showinfo_t ScanPos::readShow()
//
////////////

showinfo_t ScanPos::readShow()
{
  showinfo_t wanted;  
  char *buffer = (char*)&wanted;
  for (unsigned int i = 0; i < sizeof(wanted); ++i)
    logfile.get(buffer[i]);
  return wanted;
}

//////////
//
// msginfo_t ScanPos::readMsg()
//
/////////////

msginfo_t ScanPos::readMsg()
{
  msginfo_t wanted;
  char *buffer = (char*)&wanted;
  for (unsigned int i = 0; i < sizeof(wanted); ++i)
    logfile.get(buffer[i]);
  return wanted;
}

//////////
//
// drawinfo_t ScanPos::readDraw()
//
/////////////

drawinfo_t ScanPos::readDraw()
{
  drawinfo_t wanted;
  char *buffer = (char*)&wanted;
  for (unsigned int i = 0; i < sizeof(wanted); ++i)
    logfile.get(buffer[i]);
  return wanted;
}

//////////////////////////////////////////////////////////////////////////////////////////

////////////////
//
// int ScanPos::doField()
//
//////////////

int ScanPos::doField()
{
  ofstream field;
  field.open("field.out", ios::out);

  // borderlines of the field
  field << -FIELD_LENGTH/2 << " " << -FIELD_WIDTH/2 << endl; 
  field << -FIELD_LENGTH/2 << " " <<  FIELD_WIDTH/2 << endl;
  field <<  FIELD_LENGTH/2 << " " <<  FIELD_WIDTH/2 << endl;
  field <<  FIELD_LENGTH/2 << " " << -FIELD_WIDTH/2 << endl;
  field << -FIELD_LENGTH/2 << " " << -FIELD_WIDTH/2 << endl;
  field << endl;

  // middleline
  field << 0 << " " << -FIELD_WIDTH/2 << endl; 
  field << 0 << " " <<  FIELD_WIDTH/2 << endl;
  field << endl;
  
  // penaltyarea left
  field << -FIELD_LENGTH/2                       << " " <<  PENALTY_AREA_WIDTH/2 << endl;
  field << -FIELD_LENGTH/2 + PENALTY_AREA_LENGTH << " " <<  PENALTY_AREA_WIDTH/2 << endl;
  field << -FIELD_LENGTH/2 + PENALTY_AREA_LENGTH << " " << -PENALTY_AREA_WIDTH/2 << endl;
  field << -FIELD_LENGTH/2                       << " " << -PENALTY_AREA_WIDTH/2 << endl;
  field << endl;
  
  // goalarea left
  field << -FIELD_LENGTH/2                    << " " <<  GOAL_AREA_WIDTH/2 << endl;
  field << -FIELD_LENGTH/2 + GOAL_AREA_LENGTH << " " <<  GOAL_AREA_WIDTH/2 << endl;
  field << -FIELD_LENGTH/2 + GOAL_AREA_LENGTH << " " << -GOAL_AREA_WIDTH/2 << endl;
  field << -FIELD_LENGTH/2                    << " " << -GOAL_AREA_WIDTH/2 << endl;
  field << endl;

  // penaltyarea right
  field << FIELD_LENGTH/2                       << " " <<  PENALTY_AREA_WIDTH/2 << endl;
  field << FIELD_LENGTH/2 - PENALTY_AREA_LENGTH << " " <<  PENALTY_AREA_WIDTH/2 << endl;
  field << FIELD_LENGTH/2 - PENALTY_AREA_LENGTH << " " << -PENALTY_AREA_WIDTH/2 << endl;
  field << FIELD_LENGTH/2                       << " " << -PENALTY_AREA_WIDTH/2 << endl;
  field << endl;
  
  // goalarea right
  field << FIELD_LENGTH/2                    << " " <<  GOAL_AREA_WIDTH/2 << endl;
  field << FIELD_LENGTH/2 - GOAL_AREA_LENGTH << " " <<  GOAL_AREA_WIDTH/2 << endl;
  field << FIELD_LENGTH/2 - GOAL_AREA_LENGTH << " " << -GOAL_AREA_WIDTH/2 << endl;
  field << FIELD_LENGTH/2                    << " " << -GOAL_AREA_WIDTH/2 << endl;
  field << endl;

  field.close();
  return 1;
}

////////
//
// int main(int argc, char **argv)
//
//////////

int main(int argc, char **argv) {
  if (argc == 3) {
    char name[32];
    strcpy(name, argv[1]);
    char side;
    side = argv[2][0];

    cout << endl
	 << "MRB scanpos"
	 << endl
	 << "scanning " << side << "-side of file " << name
	 << endl;

    bool b;
    if (side == 'l')
      b = true;
    else
      b = false;
	 
    cout << "..."
	 << endl;
     
    ScanPos cp;

    if (cp.OpenFile(name)) {
      while (cp.writePos(b));
    }

    cp.doField();

    cout << "scanned logfile, output written to object<i>.out"
	 << endl
	 << endl;

    return 1;
  }
  cerr << "Syntax: scanpos <soccerlog> <l/r>"
       << endl;
  return 0;
}
